/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Portable Parallel Library.                                              */
/* User main program.                                                      */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 29.03.1995  Droux       Created                                         */
/* 08.05.1995  Droux       dds_init replaced by dds_set                    */
/* 08.05.1995  Droux       Inverted parameters of dds_copy                 */
/* 08.05.1995  Droux       dds_store replaced by dds_get                   */
/*-------------------------------------------------------------------------*/
/* filename: ddsmain.c                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel                                                   */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <ppc/parlib.h>
#include <ppc/matrixutil.h>
#include <parlib/ddstest.h>

#define VALUE_INT 12345
#define M_NB_ROW 5
#define M_NB_COL 5
#define FILE_NAME "/tmp/ddsfile"

static void print_directory (void);
static void init_matrix (double **m, int nb_row, int nb_col);

/*-------------------------------------------------------------------------*/
/* ppc_user_main:                                                          */
/* User program entry point.                                               */
/*-------------------------------------------------------------------------*/
void ppc_user_main (void)
{
  double **m;
  double **n;
  double **p;
  double *v;
  int i;
  FILE *file;
  
  printf("DDS test program started\n");
  
  /* object creation */
  printf("-> creating int scalar ... "); fflush(stdout);
  dds_create("a", INT, DDS_REPLICATED, 1, END);
  printf("OK\n");
  print_directory();
  
  printf("-> creating unsigned short matrix ... "); fflush(stdout);
  dds_create("matrix", UNSIGNED_SHORT, DDS_REPLICATED, 2, 3, END);
  printf("OK\n");
  print_directory();
  
  printf("-> creating double vector ... "); fflush(stdout);
  dds_create("vector", DOUBLE, DDS_REPLICATED, 10, END);
  printf("OK\n");
  print_directory();
  
  /* recreating existing object */
  printf("-> recreating existing object ... "); fflush(stdout);
  dds_create("a", INT, DDS_REPLICATED, 1, END);
  printf("OK\n");
  print_directory();

  /* object replacement */
  printf("-> replacing object by another ... "); fflush(stdout);
  dds_create("a", DOUBLE, DDS_REPLICATED, 5, 2, END);
  printf("OK\n");
  print_directory();

  /* object deletion */  
  printf("-> deleting matrix ... "); fflush(stdout);
  dds_delete("matrix");
  printf("OK\n");
  print_directory();

  /* clearing the objects database */
  printf("-> clearing database ... "); fflush(stdout);
  dds_clear_database();
  printf("OK\n");  
  print_directory();

  /* creation after deletion */
  printf("-> creating double matrix ... "); fflush(stdout);
  dds_create("m", DOUBLE, DDS_REPLICATED, M_NB_ROW, M_NB_COL, END);
  printf("OK\n");
  print_directory();
  
  /* object initialization */
  printf("-> original double matrix m:\n");
  m = dmatalloc(M_NB_ROW, M_NB_COL);
  init_matrix(m, M_NB_ROW, M_NB_COL);
  dmatprint(m, M_NB_ROW, M_NB_COL, "%5.2f  ");

  printf("-> initializing double matrix ... "); fflush(stdout);
  dds_set("m", m, END);
  printf("OK\n");

  /* copying objects */ 
  printf("-> copying m into n ... "); fflush(stdout);
  dds_copy("m", "n");
  printf("OK\n");
  print_directory();  

  /* object storage in local data */
  n = dmatalloc(M_NB_ROW, M_NB_COL);
  dmatzero(n, M_NB_ROW, M_NB_COL);
  printf("-> getting n double matrix value ... "); fflush(stdout);
  dds_get("n", n, END);
  printf("OK\n");
  printf("-> double matrix copy n:\n");
  dmatprint(n, M_NB_ROW, M_NB_COL, "%5.2f  ");

  /* access to DDS object from within a parallel routine */
  printf("-> calling test routine ... "); fflush(stdout);
  ddstest("n", M_NB_ROW, M_NB_COL);
  printf("OK\n");
  
  /* partial object initialization */
  /* replace the third row of the matrix "m" with a given vector */
  printf("-> partial m matrix initialization ... "); fflush(stdout);
  v = dvecalloc(M_NB_COL);
  for (i = 0; i < M_NB_COL; i++)
    v[i] = -1.0*(double)(M_NB_COL-i);
  dds_set("m", v, 2, END);
  dds_get("m", m, END);
  printf("OK\n");
  printf("-> double matrix m:\n");
  dmatprint(m, M_NB_ROW, M_NB_COL, "%5.2f  ");
  
  /* partial object storage in local data */
  /* store the second row of the matrix "m" within a given vector */
  printf("-> partial m matrix storage in local data ... "); fflush(stdout);
  dveczero(v, M_NB_ROW);
  dds_get("m", v, 1, END);
  printf("OK\n");
  printf("-> vector v:\n");
  dvecprint(v, M_NB_COL, "%5.2f  ");
  
  { /* print the properties of the matrix m */
    dds_object_entry prop;
    int c;
    
    printf("-> properties of matrix m:\n");
    dds_object_prop("m", &prop);
    printf("   name         : \"%s\"\n", prop.name);
    printf("   type         : %s\n", types_names_str[prop.type]);
    printf("   distribution : %s\n", 
      dds_distribution_str[prop.distribution]);
    printf("   dimensions   : ");
    for (c = 0; c < prop.nb_dim; c++)
    {
      if (c > 0)
        printf("x");
      printf("%d", prop.dim[c]);
    }
    printf("\n");
  }
  
  /* store m into a file */
  printf("-> store m into " FILE_NAME " ... "); fflush(stdout);
  file = fopen(FILE_NAME, "w");
  dds_get_to_stream("m", file, END);
  fclose(file);
  printf("OK\n");
  
  /* read file and init new object */
  printf("-> creating matrix p ... "); fflush(stdout);
  dds_create("p", DOUBLE, DDS_REPLICATED, M_NB_ROW, M_NB_COL, END);
  printf("OK\n");
  print_directory();
  
  printf("-> initializing p from " FILE_NAME " ... "); fflush(stdout);
  file = fopen(FILE_NAME, "r");
  dds_set_from_stream("p", file, END);
  fclose(file);
  printf("OK\n");
  
  p = dmatalloc(M_NB_ROW, M_NB_COL);
  dds_get("p", p, END);
  printf("-> matrix p:\n");
  dmatprint(p, M_NB_ROW, M_NB_COL, "%5.2f  ");
}

/*-------------------------------------------------------------------------*/
/* print_directory:                                                        */
/* print a description of the DDS objects.                                 */
/*-------------------------------------------------------------------------*/
static void print_directory (void)
{
  dds_object_entry entry;
  int eo_entries;
  int c;
  
  printf("   name           ");
  printf(" type           ");
  printf(" distribution    ");
  printf(" nb_dim  dim\n");

  dds_directory_init();
  eo_entries = dds_directory_get_entry(&entry);
  while (!eo_entries)
  {
    printf("   %-15s %-15s %-15s  %-6d  ", 
      entry.name, types_names_str[entry.type], 
      dds_distribution_str[entry.distribution], entry.nb_dim);
    for (c = 0; c < entry.nb_dim; c++)
    {
      if (c > 0)
        printf("x");
      printf("%d", entry.dim[c]);
    }
    printf("\n");
    eo_entries = dds_directory_get_entry(&entry);
  }
}

/*-------------------------------------------------------------------------*/
/* init_matrix:                                                            */
/* Initialize the given matrix.                                            */
/*-------------------------------------------------------------------------*/
static void init_matrix (double **m, int nb_row, int nb_col)
{
  int i;
  int j;
  
  for (i = 0; i < nb_row; i++)
    for (j = 0; j < nb_col; j++)
      m[i][j] = (double)(i+1)*(j+1);
}

