/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* CHOLESKY parallel library routine.                                      */
/* implementation for BANDED definite positive matrices.                   */
/* Master.                                                                 */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.03.1995  Chevalier   Created                                         */
/* 25.04.1995  Droux       Cleanup, math.h inclusion                       */
/*-------------------------------------------------------------------------*/
/* filename: master.c                                                      */
/* Tested on :                                                             */
/*   0.9.6                                                                 */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <math.h>
#include <stdlib.h>
#include <ppc/csp.h>
#include <ppc/parlib.h>
#include <ppc/matrixutil.h>

/*-------------------------------------------------------------------------*/
/* ppc_main:                                                               */
/* Master task entry point.                                                */
/*-------------------------------------------------------------------------*/
int ppr_main(
  channel from_host,
  channel to_host,
  channel from_slaves [], 
  channel to_slaves [],
  int nb_row,
  int nb_col,
  int NL,
  int bandwidth,
  double **matA0,
  double *vecB0,
  double *vecX0 
)
{  
  int i, j, k, n ;
  int nb_slaves = nb_row * nb_col;
  int nb_col_per_slave, proc_number ;
  int N, line_number, changed, component ;
  double sum ;
  double *vecB, *vecX, *vecY ;
  double **matA ;
  double elmt ;
   
       
  /***********************************************************/     

  N = ((NL-1) / nb_slaves + 1)
* nb_slaves ;
  
  matA = dmatalloc(N, bandwidth+1) ;
  
  vecB = dvecalloc(N) ;
  vecX = dvecalloc(N) ;
  vecY = dvecalloc(N) ;
  dveczero (vecB, N) ;
  dveczero (vecY, N) ;
  dveczero (vecX, N) ;
  
  for (i=1 ; i<= NL ; i++)
    for (j=1; j<=bandwidth+1; j++)
      matA[i-1][j-1] = matA0[i-1][j-1] ;
  
  for (i=NL+1 ; i<= N ; i++)
      matA[i-1][bandwidth] = 1.0 ;
  
  for (i=1 ; i<= NL ; i++)
      vecB[i-1] = vecB0[i-1] ;

  /****************************************************************/ 
  /******** CHOLESKY - FACTORIZATION ******************************/
  /****************************************************************/ 

  nb_col_per_slave = N / nb_slaves ;

  /*  Sending data to slaves:       */
  
  /*  Sending to slaves size N of the matrix and number of columns 
      it must receive:       */
  for (n = 0; n < nb_slaves; n++) 
    send_channel (to_slaves[n],
      INT, N, 
      INT, nb_col_per_slave,
      INT, bandwidth,
      END) ;
    
  /*  distribute the processors their own columns : */	     
  for (j = 1; j <= N; j++) 
  {
    proc_number = (j-1) % nb_slaves ; 
    send_channel (to_slaves[proc_number], 
      INT, j,
      END) ; 

    send_channel (to_slaves[proc_number], 
      DOUBLE_ARRAY, bandwidth+1, matA[j-1], 
      END) ; 
  }

  /* Beginning of the CHOLESKY-Decomposition */

  for (j=1; j <= N; j++) 
  {
    sum = 0.0 ;
    for (k=1; k<= bandwidth; k++) 
      sum = sum + matA[j-1][k-1] * matA[j-1][k-1] ;

    matA[j-1][bandwidth] = sqrt(matA[j-1][bandwidth] - sum) ;
     
    for (n = 0; n < nb_slaves; n++) 
      send_channel (to_slaves[n], 
	INT, j,
	DOUBLE_ARRAY, bandwidth+1, matA[j-1], 
	END) ;

    for (k = 1; k <= bandwidth; k++)
      for (n = 0; n < nb_slaves; n++) 
      {
        receive_channel (from_slaves[n],
	  INT, &changed, 
	  INT, &line_number,
	  INT, &component,
	  DOUBLE, &elmt,
	  END) ;
          
        if (changed > 0)
          matA[line_number-1][component] = elmt ;
      }
  }

  /****************************************************************/ 
  /******** END OF CHOLESKY - FACTORIZATION ***********************/
  /****************************************************************/ 
 
  /******** Solution of Triangular system :************************/

  /* vector vecY is already initialized ! */
  for (k = 1; k <= N; k++)
  {
    sum = 0.0;
    for (j = 1; j <= bandwidth; j++)
      sum = sum + matA[k-1][j-1] * vecY[k-bandwidth+j-2] ;

    vecY[k-1] = (vecB[k-1] - sum) / matA[k-1][bandwidth] ;
  }

  for (i = N; i >= 1 ; i--)
  {
    sum = 0.0;
    for (k = 1; k <= bandwidth; k++)
      if (i+k <=N)
        sum = sum + matA[i+k-1][bandwidth-k] * vecX[i+k-1] ;

    vecX[i-1] = (vecY[i-1] - sum) / matA[i-1][bandwidth] ;
  }

  for (i=1 ; i <= NL ; i++)
    vecX0[i-1] = vecX[i-1] ;

  dvecfree(vecB) ;
  dvecfree(vecX) ;
  dvecfree(vecY) ;
  dmatfree(matA, N) ;
  
  return PPR_OK ;
}
