/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* LEAST-SQUARES CONJUGATE-GRADIENT parallel library routine.              */
/* implementation for sparse definite positive matrices.                   */
/* Master.                                                                 */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.03.1995  Chevalier   Created                                         */
/* 25.04.1995  Droux       Cleanup, math.h inclusion                       */
/*-------------------------------------------------------------------------*/
/* filename: master.c                                                      */
/* Tested on :                                                             */
/*   0.9.6                                                                 */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <math.h>
#include <stdlib.h>
#include <ppc/csp.h>
#include <ppc/parlib.h>
#include <ppc/matrixutil.h>

static double dotprod(int n, double *x, double *y) 
{
  double s = 0.0 ;
  int i ;
  
  for(i=1; i<=n; i++)
      s += x[i-1]*y[i-1] ;
  
  return s;
}

int ppr_main (
  channel from_host,
  channel to_host,
  channel from_slaves[],
  channel to_slaves[],
  int nb_row,
  int nb_col,
  int nonZeroNumber,
  int N,
  double *valvector,
  int *linvector,
  int *colvector,
  double *vecB,
  double *vecX 
)
{  
  int nb_slaves = nb_row * nb_col;
  int i, n ;
  int nc, length, upperbound, k, terminate ;
  int *colslice, *linslice ;
  double rho0, rhok, beta, dotp, alpha, epsilon ;
  double *valslice ; 
  double *vecR, *vecP, *vecW ;
   
       
  nc = nonZeroNumber / nb_slaves ;

  valslice = dvecalloc(nc+nb_slaves) ;
  linslice = ivecalloc(nc+nb_slaves) ;
  colslice = ivecalloc(nc+nb_slaves) ;

  for(n=0;n<nb_slaves-1;n++) 
  {
    for(i=0;i<nc;i++)
    {
      valslice[i] = valvector[n*nc+i] ;
      linslice[i] = linvector[n*nc+i] ;
      colslice[i] = colvector[n*nc+i] ;
    }

    send_channel (to_slaves[n],
      INT, nc,
      INT, N,
      END) ; 

    send_channel (to_slaves[n], 
      DOUBLE_ARRAY, nc, valslice, 
      INT_ARRAY, nc, linslice, 
      INT_ARRAY, nc, colslice, 
      END) ; 
  }

  upperbound = nc+nonZeroNumber%nb_slaves ;
  for(i=0;i<upperbound;i++)
  {
    valslice[i] = valvector[(nb_slaves-1)*nc+i] ;
    linslice[i] = linvector[(nb_slaves-1)*nc+i] ;
    colslice[i] = colvector[(nb_slaves-1)*nc+i] ;
  }

  length = nc+nonZeroNumber%nb_slaves ;
  send_channel (to_slaves[nb_slaves-1],
    INT, length,
    INT, N,
    END) ; 

  send_channel (to_slaves[nb_slaves-1], 
    DOUBLE_ARRAY, length, valslice, 
    INT_ARRAY, length, linslice, 
    INT_ARRAY, length, colslice, 
    END) ; 
  
  /***********************************************************/     

  vecR = dvecalloc(N) ;
  vecP = dvecalloc(N) ;
  vecW = dvecalloc(N) ;

  for(i=1;i<=N;i++)
  {
    vecR[i-1] = vecB[i-1] ;
    vecP[i-1] = vecR[i-1] ;
  }

  rho0 = dotprod(N, vecR, vecR) ; 
  k = 1 ;

  dotp = 0.0 ; 
  for(n=0;n<nb_slaves;n++) 
  send_channel (to_slaves[n], 
    DOUBLE_ARRAY, N, vecP, 
    END) ; 
 
  dveczero (vecW, N) ;
 
  send_channel (to_slaves[0], 
    DOUBLE_ARRAY, N, vecW, 
    DOUBLE, dotp,
    END) ; 
  
  receive_channel (from_slaves[nb_slaves-1], 
    DOUBLE_ARRAY, N, vecW, 
    DOUBLE, &dotp,
    END) ; 
  
  alpha = rho0/dotp ;

  for(i=1;i<=N;i++)
  {
    vecX[i-1] = vecX[i-1] + alpha*vecP[i-1] ;
    vecR[i-1] = vecR[i-1] - alpha*vecW[i-1] ;
  }

  rhok = dotprod(N, vecR, vecR) ;
  k++ ;

  epsilon = 10.0e-10 ;
  terminate = 0 ;    
  do 
  {
    beta = rhok/rho0 ;
    for(i=1;i<=N;i++)
    {
      vecP[i-1] = vecR[i-1] + beta*vecP[i-1] ;
      vecW[i-1] = 0.0 ;
    }

    dotp = 0.0 ; 
    
    for(n=0;n<nb_slaves;n++) 
      send_channel (to_slaves[n], 
	DOUBLE_ARRAY, N, vecP, 
	END) ; 
    
    send_channel (to_slaves[0], 
      DOUBLE_ARRAY, N, vecW, 
      DOUBLE, dotp,
      END) ; 
    
    receive_channel (from_slaves[nb_slaves-1], 
      DOUBLE_ARRAY, N, vecW, 
      DOUBLE, &dotp,
      END) ; 
  
    alpha = rhok/dotp ;

    for(i=1;i<=N;i++)
    {
      vecX[i-1] = vecX[i-1] + alpha*vecP[i-1] ;
      vecR[i-1] = vecR[i-1] - alpha*vecW[i-1] ;
    }
    
    rho0 = rhok ; 
    rhok = dotprod(N, vecR, vecR) ;
    
    if(rhok <= epsilon*epsilon*dotprod(N, vecB, vecB))
      terminate = 1 ;

    for(n=0;n<nb_slaves;n++) 
      send_channel (to_slaves[n], 
	INT, terminate, 
	END) ; 

     k++ ;
  } while(terminate < 1) ;

  return PPR_OK;
}

