/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* The traveling salesman problem using genetic algorithms.                */
/* Utilities.                                                              */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 01.03.1995  R. Cattin   Sequential version                              */
/* 01.03.1995  Droux       Parallel version                                */
/* 01.03.1995  Droux       Added support for dynamic allocation            */
/* 01.03.1995  Droux       Optimized sequential parts for large problems   */
/* 01.03.1995  Droux       Suppressed Town structure                       */
/*-------------------------------------------------------------------------*/
/* filename: genutil.c                                                     */
/* Tested on :                                                             */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "genutil.h"

#define NrOfGenes NrOfCities

static int NrOfIndiv;         /* number of individuals in local population */
static int NrOfCities;                       /* initialized by InitTowns() */
static double *townsX;                              /* towns to be visited */
static double *townsY;
static Gene *tempFather, *tempMother;
static double tempFatherFitness, tempMotherFitness;
static int *used;

/*-------------------------------------------------------------------------*/
/* InitGenUtil:                                                            */
/* Initialize the utilities module.                                        */
/*-------------------------------------------------------------------------*/
void InitGenUtil (int nrTowns, int nrIndiv)
{
  NrOfCities = nrTowns;
  NrOfIndiv = nrIndiv;
  AllocIndividual(&tempFather);
  AllocIndividual(&tempMother);
  used = (int *)malloc(sizeof(int)*NrOfCities);
}

/*-------------------------------------------------------------------------*/
/* InitPopulation:                                                         */
/* initializes each individual in a population of NrOfIndiv individuals    */
/* with randomly selected tours.                                           */
/*-------------------------------------------------------------------------*/
void InitPopulation(Gene **population, double *fitness)
{
  int indiv, i, r;

  for(indiv=0; indiv<NrOfIndiv; indiv++)
  {
    for(i=0; i<NrOfGenes; i++)
      used[i] = FALSE;
    for(i=0; i<NrOfGenes; i++)
    {
      do
      {
        r = RandomOf(NrOfGenes);
      } while ( used[r] );
      population[indiv][i] = r;
      used[r] = TRUE;
    }
    fitness[indiv] = Fitness(population[indiv]);
  }
}

/*-------------------------------------------------------------------------*/
/* InitTowns:                                                              */
/* Initializes towns and their coordinates                                 */
/*-------------------------------------------------------------------------*/
void InitTowns (double *theTownsX, double *theTownsY, int initialize)
{
  int i;

  townsX = theTownsX;
  townsY = theTownsY;

  if (initialize)
  {
    for(i=0; i<NrOfCities; i++)
    {
      townsX[i] = Random();
      townsY[i] = Random();
    }
  }
}

/*-------------------------------------------------------------------------*/
/* Fitness:                                                                */
/* Returns the fitness of an individual. The fitness is the reciprocal     */
/* value of the tour distance.                                             */
/*-------------------------------------------------------------------------*/
double Fitness(Gene *individual)
{
  int i;
  double x1, y1, x2, y2, distance;

  x1 = townsX[individual[0]];
  y1 = townsY[individual[0]];
  distance = 0.0;
  for(i=1; i<NrOfGenes; i++)
  {
    x2 = townsX[individual[i]];
    y2 = townsY[individual[i]];
    distance += sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) );
    x1=x2; y1=y2;
  }
  
  return 1.0/distance;
}

/*-------------------------------------------------------------------------*/
/* UpdateFitness:                                                          */
/* Updates the fitness of the individuals in population and returns the    */
/* sum. The individuals are sorted with decreasing fitness.                */
/*-------------------------------------------------------------------------*/
double UpdateFitness(Gene **population, double *fitness)
{
  int indiv, i, j;
  Gene *temp;
  double tempFitness;
  double sumfitness;

  /* compute fitness of each individual */
  sumfitness = 0.0;
  for(indiv=0; indiv<NrOfIndiv; indiv++)
    sumfitness += fitness[indiv] = Fitness(population[indiv]);
    
  /* sort with increased fitness (simple bubble sort ) */
  for(i=NrOfIndiv; i>0; i--)
  {
    for(j=1; j<i; j++)
    {
      if( fitness[j-1] < fitness[j])
      {
	temp = population[j-1];
	population[j-1] = population[j];
	population[j] = temp;
	tempFitness = fitness[j-1];
	fitness[j-1] = fitness[j];
	fitness[j] = tempFitness;
      }
    }
  }
  
  return sumfitness;
}

/*-------------------------------------------------------------------------*/
/* Select:                                                                 */
/* Selects randomly an individual from population and returns the index of */
/* this individual in population. 'sumfitness' is the fitness sum of all   */
/* individuals. Uses Random wheel selection.                               */
/*-------------------------------------------------------------------------*/
int Select(double *fitness, double sumfitness)
{
  int indiv;
  double partsum, rand;

  /* roulette wheel selection */
  rand = Random() * sumfitness;
  indiv = 0; partsum = 0.0;
  do
    partsum += fitness[indiv++];
  while( (partsum < rand) && (indiv < NrOfIndiv) );
  
  return --indiv;
}

/*-------------------------------------------------------------------------*/
/* SwapGenes:                                                              */
/* Exchanges gene1 and gene2 in the chromosome of an individual.           */
/*-------------------------------------------------------------------------*/
void SwapGenes(int gene1, int gene2, Gene *individual)
{
  int temp;

  temp = individual[gene1];
  individual[gene1] = individual[gene2];
  individual[gene2] = temp;
}

/*-------------------------------------------------------------------------*/
/* FindGene:                                                               */
/* Finds the position of gene in the chromosome in an individual.          */
/*-------------------------------------------------------------------------*/
int FindGene(int gene, Gene *individual)
{
  int j;

  j = 0;
  while( individual[j] != gene) j++;
  
  return j;
}

/*-------------------------------------------------------------------------*/
/* Crossover:                                                              */
/* Performs a partially matched crossover scheme with two parent           */
/* individuals and returns the fitter of the two resulting offsprings. The */
/* crossover is executed with a probability of 'CrossProb' %.              */
/*-------------------------------------------------------------------------*/
void Crossover(int father, int mother, Gene **population, double *fitness,
  Gene *ret, double *retFitness)
{
  int hi_test, lo_cross, hi_cross, j;

  if( HitProbability(CrossProb) )
  {
    lo_cross = RandomOf( NrOfGenes-1);
    while ( (hi_cross = RandomOf( NrOfGenes )) <= lo_cross);
    hi_test = (hi_cross+1) % NrOfGenes;
    CopyIndividual(tempFather, population[father]);
    tempFatherFitness = fitness[father];
    CopyIndividual(tempMother, population[mother]);
    tempMotherFitness = fitness[mother];

    /* partially matched crossover (PMX) */
    j = lo_cross;
    if( (lo_cross != hi_cross) && (lo_cross != hi_test) )
    while( j != hi_test )
    {
      SwapGenes( j, FindGene(tempFather[j], tempMother), tempMother);
      SwapGenes( j, FindGene(tempMother[j], tempFather), tempFather);
      j = (j+1) % NrOfGenes;
    }
    if( Fitness(tempMother) > Fitness(tempFather) )
    {
      CopyIndividual(ret, tempMother);
      *retFitness = tempMotherFitness;
    }
    else
    {
      CopyIndividual(ret, tempFather);
      *retFitness = tempFatherFitness;
    }
  }
  else
  {
    CopyIndividual(ret, population[father]);
    *retFitness = fitness[father];
  }
}

/*-------------------------------------------------------------------------*/
/* EventuallyMutate:                                                       */
/* Performs a mutation with a probability 'MutationProb' %. The mutation   */
/* exchanges two randomly selected genes.                                  */
/*-------------------------------------------------------------------------*/
void EventuallyMutate(Gene *individual)
{
  int hi, lo;
  Gene tmp;

  if( HitProbability(MutationProb) )
  {
    lo = RandomOf(NrOfGenes-1);
    while ( (hi = RandomOf( NrOfGenes )) <= lo);

    tmp = individual[lo];
    individual[lo] = individual[hi];
    individual[hi] = tmp;
  }
}

/*-------------------------------------------------------------------------*/
/* RandomOf:                                                               */
/* Returns a random integer in the range 0 .. (range-1)                    */
/*-------------------------------------------------------------------------*/
int RandomOf(int range)
{
  return (rand() % (range));
}

/*-------------------------------------------------------------------------*/
/* Random:                                                                 */
/* Returns a random double in the range 0.0 .. 1.0                         */
/*-------------------------------------------------------------------------*/
double Random(void)
{
  return ((double)rand() / (double)RAND_MAX);
}

/*-------------------------------------------------------------------------*/
/* HitProbability:                                                         */
/* Returns TRUE with a probability of 'percent' %                          */
/*-------------------------------------------------------------------------*/
int HitProbability(unsigned int percent)
{
  return (RandomOf(100) <= percent);
}

/*-------------------------------------------------------------------------*/
/* DisplayTour:                                                            */
/* Display the tour corresponding to the given individual.                 */
/*-------------------------------------------------------------------------*/
void DisplayTour(Gene *individual, double fitness)
{
  int i, town;
  double x,y;

  for(i=0; i<NrOfGenes; i++)
  {
    town = individual[i];
    x = townsX[town];
    y = townsY[town];
    printf("Town %d\t%f\t%f\n", town, x, y);
  }
  printf("Distance: %f\n", 1.0 / fitness);
}

/*-------------------------------------------------------------------------*/
/* NumberOfSpecies:                                                        */
/* Returns the number of species in the specified population.              */
/*-------------------------------------------------------------------------*/
int NumberOfSpecies(double *fitness)
{
  int species, indiv;
  double actualFitness;

  species = 1;
  actualFitness = fitness[0];

  /* assumes that the same fitness means the same tour; this may
      be not correct ! */
  for(indiv=0; indiv<NrOfIndiv; indiv++)
    if(actualFitness != fitness[indiv])
    {
      species++;
      actualFitness = fitness[indiv];
    }
    
  return species;
}

/*-------------------------------------------------------------------------*/
/* DisplayStatistics:                                                      */
/* Print the current state of the population.                              */
/*-------------------------------------------------------------------------*/
void DisplayStatistics(Gene **population, double *fitness, int generation)
{
  printf("%03d\t%5.2f\t%5.2f\t%5.2f\t%d\n", generation, 
    1.0/MeanFitness(fitness),
    1.0/Fitness(population[0]), 1.0/Fitness(population[NrOfIndiv-1]),
    NumberOfSpecies(fitness));
}

/*-------------------------------------------------------------------------*/
/* MeanFitness:                                                            */
/* Returns the mean of fitness of all individuals in the population        */
/*-------------------------------------------------------------------------*/
double MeanFitness(double *fitness)
{
  int i;
  double sum;

  sum = 0.0;
  for(i=0; i<NrOfIndiv; i++)
    sum += fitness[i];
  
  return sum / (double) NrOfIndiv;
}

/*-------------------------------------------------------------------------*/
/* CopyIndividual:                                                         */
/* Copy an individual into an other one.                                   */
/*-------------------------------------------------------------------------*/
void CopyIndividual (Gene *dest, Gene *src)
{
  int c;
  
  for (c = 0; c < NrOfCities; c++)
    dest[c] = src[c];
}

/*-------------------------------------------------------------------------*/
/* AllocIndividual:                                                        */
/* Allocate a new individual.                                              */
/*-------------------------------------------------------------------------*/
void AllocIndividual (Gene **ind)
{
  *ind = (Gene *)malloc(sizeof(Gene)*NrOfCities);
}

/*-------------------------------------------------------------------------*/
/* FreeIndividual:                                                         */
/* Free an individual.                                                     */
/*-------------------------------------------------------------------------*/
void FreeIndividual (Gene *ind)
{
  free(ind);
}

/*-------------------------------------------------------------------------*/
/* PrintIndividual:                                                        */
/* Print the given individual.                                             */
/*-------------------------------------------------------------------------*/
void PrintIndividual (Gene *ind)
{
  int c;
  
  for (c = 0; c < NrOfCities; c++)
  {
    printf("%d ", ind[c]);
  }
  printf("\n");
}

/*-------------------------------------------------------------------------*/
/* SaveTour:                                                               */
/* Save the given tour in a file.                                          */
/*-------------------------------------------------------------------------*/
void SaveTour(char *fileName, Gene *individual)
{
  int i, town;
  double x,y;
  FILE *out;
  
  out = fopen(fileName, "w");

  for(i=0; i<NrOfGenes; i++)
  {
    town = individual[i];
    x = townsX[town];
    y = townsY[town];
    fprintf(out, "%f %f\n", x, y);
  }
}
