/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* The traveling salesman problem using genetic algorithms.                */
/* Utilities.                                                              */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 01.03.1995  R. Cattin   Sequential version                              */
/* 01.03.1995  Droux       Parallel version                                */
/* 01.03.1995  Droux       Added support for dynamic allocation            */
/* 01.03.1995  Droux       Optimized sequential parts for large problems   */
/* 01.03.1995  Droux       Suppressed Town structure                       */
/*-------------------------------------------------------------------------*/
/* filename: genutil.h                                                     */
/* Tested on :                                                             */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#ifndef _genutil_h
#define _genutil_h

#define CrossProb 60                /* probability of crossover in percent */
#define MutationProb 4            /* probability of a mutation in percent */

#define TRUE  1
#define FALSE 0

typedef int Gene;

/*-------------------------------------------------------------------------*/
/* InitGenUtil:                                                            */
/* Initialize the utilities module.                                        */
/*-------------------------------------------------------------------------*/
void InitGenUtil (int nrTowns, int nrIndiv);

/*-------------------------------------------------------------------------*/
/* InitPopulation:                                                         */
/* initializes each individual in a population of NrOfIndiv individuals    */
/* with randomly selected tours.                                           */
/*-------------------------------------------------------------------------*/
void InitPopulation(Gene **population, double *fitness);

/*-------------------------------------------------------------------------*/
/* InitTowns:                                                              */
/* Initializes towns and their coordinates                                 */
/*-------------------------------------------------------------------------*/
void InitTowns (double *theTownsX, double *theTownsY, int initialize);

/*-------------------------------------------------------------------------*/
/* Fitness:                                                                */
/* Returns the fitness of an individual. The fitness is the reciprocal     */
/* value of the tour distance.                                             */
/*-------------------------------------------------------------------------*/
double Fitness(Gene *individual);

/*-------------------------------------------------------------------------*/
/* UpdateFitness:                                                          */
/* Updates the fitness of the individuals in population and returns the    */
/* sum. The individuals are sorted with decreasing fitness.                */
/*-------------------------------------------------------------------------*/
double UpdateFitness(Gene **population, double *fitness);

/*-------------------------------------------------------------------------*/
/* Select:                                                                 */
/* Selects randomly an individual from population and returns the index of */
/* this individual in population. 'sumfitness' is the fitness sum of all   */
/* individuals. Uses Random wheel selection.                               */
/*-------------------------------------------------------------------------*/
int Select(double *fitness, double sumfitness);

/*-------------------------------------------------------------------------*/
/* SwapGenes:                                                              */
/* Exchanges gene1 and gene2 in the chromosome of an individual.           */
/*-------------------------------------------------------------------------*/
void SwapGenes(int gene1, int gene2, Gene *individual);

/*-------------------------------------------------------------------------*/
/* FindGene:                                                               */
/* Finds the position of gene in the chromosome in an individual.          */
/*-------------------------------------------------------------------------*/
int FindGene(int gene, Gene *individual);

/*-------------------------------------------------------------------------*/
/* Crossover:                                                              */
/* Performs a partially matched crossover scheme with two parent           */
/* individuals and returns the fitter of the two resulting offsprings. The */
/* crossover is executed with a probability of 'CrossProb' %.              */
/*-------------------------------------------------------------------------*/
void Crossover(int father, int mother, Gene **population, double *fitness,
  Gene *ret, double *retFitness);

/*-------------------------------------------------------------------------*/
/* EventuallyMutate:                                                       */
/* Performs a mutation with a probability 'MutationProb' %. The mutation   */
/* exchanges two randomly selected genes.                                  */
/*-------------------------------------------------------------------------*/
void EventuallyMutate(Gene *individual);

/*-------------------------------------------------------------------------*/
/* RandomOf:                                                               */
/* Returns a random integer in the range 0 .. (range-1)                    */
/*-------------------------------------------------------------------------*/
int RandomOf(int range);

/*-------------------------------------------------------------------------*/
/* Random:                                                                 */
/* Returns a random double in the range 0.0 .. 1.0                         */
/*-------------------------------------------------------------------------*/
double Random(void);

/*-------------------------------------------------------------------------*/
/* HitProbability:                                                         */
/* Returns TRUE with a probability of 'percent' %                          */
/*-------------------------------------------------------------------------*/
int HitProbability(unsigned int percent);

/*-------------------------------------------------------------------------*/
/* DisplayTour:                                                            */
/* Display the tour corresponding to the given individual.                 */
/*-------------------------------------------------------------------------*/
void DisplayTour(Gene *individual, double fitness);

/*-------------------------------------------------------------------------*/
/* NumberOfSpecies:                                                        */
/* Returns the number of species in the specified population.              */
/*-------------------------------------------------------------------------*/
int NumberOfSpecies(double *fitness);

/*-------------------------------------------------------------------------*/
/* DisplayStatistics:                                                      */
/* Print the current state of the population.                              */
/*-------------------------------------------------------------------------*/
void DisplayStatistics(Gene **population, double *fitness, int generation);

/*-------------------------------------------------------------------------*/
/* MeanFitness:                                                            */
/* Returns the mean of fitness of all individuals in the population        */
/*-------------------------------------------------------------------------*/
double MeanFitness(double *fitness);

/*-------------------------------------------------------------------------*/
/* AllocIndividual:                                                        */
/* Allocate a new individual.                                              */
/*-------------------------------------------------------------------------*/
void AllocIndividual (Gene **ind);

/*-------------------------------------------------------------------------*/
/* CopyIndividual:                                                         */
/* Copy an individual into an other one.                                   */
/*-------------------------------------------------------------------------*/
void CopyIndividual (Gene *dest, Gene *src);

/*-------------------------------------------------------------------------*/
/* FreeIndividual:                                                         */
/* Free an individual.                                                     */
/*-------------------------------------------------------------------------*/
void FreeIndividual (Gene *ind);

/*-------------------------------------------------------------------------*/
/* PrintIndividual:                                                        */
/* Print the given individual.                                             */
/*-------------------------------------------------------------------------*/
void PrintIndividual (Gene *ind);

/*-------------------------------------------------------------------------*/
/* SaveTour:                                                               */
/* Save the given tour in a file.                                          */
/*-------------------------------------------------------------------------*/
void SaveTour(char *fileName, Gene *individual);

#endif
