/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* The traveling salesman problem using genetic algorithms.                */
/* Master task.                                                            */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 01.03.1995  R. Cattin   Sequential version                              */
/* 01.03.1995  Droux       Parallel version                                */
/* 01.03.1995  Droux       Suppressed name of town for very large problems */
/* 01.03.1995  Droux       Added support for dynamic allocation            */
/* 01.03.1995  Droux       Optimized sequential parts for large problems   */
/* 01.03.1995  Droux       Suppressed Town structure                       */
/*-------------------------------------------------------------------------*/
/* filename: master.c                                                      */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3                                                          */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <ppc/csp.h>
#include "genutil.h"

static int NrOfCities;                   /* number of cities to be visited */
static int NrOfIndiv;        /* number of individuals in global population */
static int MaxGenerations;                        /* number of generations */

static double *townsX;         /* x coordinates of the towns to be visited */
static double *townsY;         /* y coordinates of the towns to be visited */
static Gene **population;         /* population chromosomes of individuals */
static double *fitness;

/*-------------------------------------------------------------------------*/
/* Allocate:                                                               */
/* Allocate dynamic objects.                                               */
/*-------------------------------------------------------------------------*/
void Allocate (int nb_slaves)
{
  int c;

  townsX = (double *)malloc(NrOfCities*sizeof(double));
  townsY = (double *)malloc(NrOfCities*sizeof(double));
  population = (Gene **)malloc(NrOfIndiv*sizeof(Gene *));
  fitness = (double *)malloc(NrOfIndiv*sizeof(double));

  for (c = 0; c < NrOfIndiv; c++)
    AllocIndividual(&population[c]);
}


/*-------------------------------------------------------------------------*/
/* ppc_main:                                                               */
/* Master task entry point.                                                */
/*-------------------------------------------------------------------------*/
void ppc_main (
  channel from_host,
  channel to_host,
  channel from_slaves [],
  channel to_slaves [],
  int nb_row,
  int nb_col
)
{
  int s;
  int p;
  int nb_slaves = nb_row*nb_col;
  time_t startTime, endTime;
  int partSize;

  /* misc inputs */
  printf("number of cities: "); fflush(stdout); 
  scanf("%d", &NrOfCities);
  NrOfIndiv = 10*NrOfCities;
  NrOfIndiv = NrOfIndiv + nb_slaves - (NrOfIndiv % nb_slaves);
  MaxGenerations = 10*NrOfCities;
  
  /* initialize genetics module and allocate local data */
  InitGenUtil(NrOfCities, NrOfIndiv);
  Allocate(nb_slaves);
  
  /* partition the total population and send initial parameters */
  partSize = NrOfIndiv / nb_slaves;
  for (s = 0; s < nb_slaves; s++)
  {
    send_channel(to_slaves[s],
      INT, NrOfCities,
      INT, MaxGenerations,
      INT, partSize,
      END);
  }
  
  InitTowns(townsX, townsY, TRUE);
  InitPopulation(population, fitness);
  SaveTour("before.tour", population[0]);
  
  time(&startTime);
  for (s = 0; s < nb_slaves; s++)
  {
    int to;

    send_channel(to_slaves[s],
      DOUBLE_ARRAY, NrOfCities, townsX,
      DOUBLE_ARRAY, NrOfCities, townsY,
      DOUBLE_ARRAY, partSize, &fitness[s*partSize],
      END);
      
    to = (s+1)*partSize;
    for (p = s*partSize; p < to; p++)
      send_channel(to_slaves[s],
	INT_ARRAY, NrOfCities, population[p],
	END);
  }
  
  #ifdef STATISTICS
  {
    int generation;

    for(generation=0; generation < MaxGenerations; generation++)
    {
      static double actFitness;
      static double bestFitness;
      
      bestFitness = 0.0;
      for (s = 0; s < nb_slaves; s++)
      {
	receive_channel(from_slaves[s],
	  DOUBLE, &actFitness,
	  END);
	if (actFitness > bestFitness) 
	  bestFitness = actFitness;
      }
      
      printf("%5.2f\n", 1.0/bestFitness);  
    }
  }
  #endif

  for (s = 0; s < nb_slaves; s++)
  {
    int to;

    receive_channel(from_slaves[s],
      DOUBLE_ARRAY, partSize, &fitness[s*partSize],
      END);

    to = (s+1)*partSize;
    for (p = s*partSize; p < to; p++)
        receive_channel(from_slaves[s],
	  INT_ARRAY, NrOfCities, population[p],
	  END);
  }
  time(&endTime);

  UpdateFitness(population, fitness);
  DisplayTour(population[0], fitness[0]);
  printf("Time to solve: %lds\n", (long)(endTime-startTime));
  SaveTour("after.tour", population[0]);
}

