/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* The traveling salesman problem using genetic algorithms.                */
/* Slave task.                                                             */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 01.03.1995  R. Cattin   Sequential version                              */
/* 01.03.1995  Droux       Parallel version                                */
/* 01.03.1995  Droux       Added support for dynamic allocation            */
/* 01.03.1995  Droux       Optimized sequential parts for large problems   */
/*-------------------------------------------------------------------------*/
/* filename: slave.c                                                       */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3                                                          */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <ppc/csp.h>
#include "genutil.h"

static int NrOfCities;                   /* number of cities to be visited */
static int NrOfIndiv;         /* number of individuals in local population */
static int MaxGenerations;                        /* number of generations */

static double *townsX;         /* x coordinates of the towns to be visited */
static double *townsY;         /* y coordinates of the towns to be visited */
static Gene **population;         /* population chromosomes of individuals */
static double *fitness;
static Gene **newPopulation;                   /* new generated population */
static double *newFitness;
static Gene *child;
static double childFitness;
static int my_number;
static double sumfitness;

/*-------------------------------------------------------------------------*/
/* AllocateSlave:                                                          */
/* Allocate dynamic objects.                                               */
/*-------------------------------------------------------------------------*/
void AllocateSlave (void)
{
  int c;

  townsX = (double *)malloc(NrOfCities*sizeof(double));
  townsY = (double *)malloc(NrOfCities*sizeof(double));
  population = (Gene **)malloc(NrOfIndiv*sizeof(Gene *));
  fitness = (double *)malloc(NrOfIndiv*sizeof(double));
  newPopulation = (Gene **)malloc(NrOfIndiv*sizeof(Gene *));
  newFitness = (double *)malloc(NrOfIndiv*sizeof(double));

  for (c = 0; c < NrOfIndiv; c++)
  {
    AllocIndividual(&population[c]);
    AllocIndividual(&newPopulation[c]);
  }

  AllocIndividual(&child);
}

/*-------------------------------------------------------------------------*/
/* Generation:                                                             */
/* Runs one reproduction cycle. Crossover and Mutations are done until a   */
/* new population is filled up. The parent are always different            */
/* individuals and are divorced after reproduction. Any individual may be  */
/* parent (no sex). The fittest individual of the parent generation is     */
/* preserved.                                                              */
/*-------------------------------------------------------------------------*/
void Generation(void)
{
  int father, mother, newindiv;
  
  sumfitness = UpdateFitness(population, fitness);
  
  /* preserve best individual */
  CopyIndividual(newPopulation[0], population[0]);
  newFitness[0] = fitness[0];
  
  for (newindiv = 1; newindiv < NrOfIndiv; newindiv++)
  {
    father = Select(fitness, sumfitness);
    while ( (mother = Select(fitness, sumfitness)) == father );
    Crossover(father, mother, population, fitness, child, &childFitness);
    EventuallyMutate(child);
    CopyIndividual(newPopulation[newindiv], child);
    newFitness[newindiv] = childFitness;
  }

  /* copy the new population into the current population */
  for (newindiv = 0; newindiv < NrOfIndiv; newindiv++)
  {
    CopyIndividual(population[newindiv], newPopulation[newindiv]);
    fitness[newindiv] = newFitness[newindiv];
  }
  UpdateFitness(population, fitness);
}

/*-------------------------------------------------------------------------*/
/* ppc_main:                                                               */
/* Slave task entry point.                                                 */
/*-------------------------------------------------------------------------*/
void ppc_main (
  channel from_master,
  channel to_master,
  channel from_north,
  channel to_north,
  channel from_south,
  channel to_south,
  channel from_west,
  channel to_west,
  channel from_east,
  channel to_east,
  int nb_row,
  int nb_col,
  int my_row,
  int my_col
)
{
  int p;
  int generation;
  int c;
  channel dest;
  channel src;
  
  my_number = my_row*nb_col+my_col;
  srand(my_number);

  receive_channel(from_master,
    INT, &NrOfCities,
    INT, &MaxGenerations,
    INT, &NrOfIndiv,
    END);
    
  InitGenUtil(NrOfCities, NrOfIndiv);
  AllocateSlave();
  InitTowns(townsX, townsY, FALSE);
  
  receive_channel(from_master,
    DOUBLE_ARRAY, NrOfCities, townsX,
    DOUBLE_ARRAY, NrOfCities, townsY,
    DOUBLE_ARRAY, NrOfIndiv, fitness,
    END);
    
  for (p = 0; p < NrOfIndiv; p++)
    receive_channel(from_master,
      INT_ARRAY, NrOfCities, population[p],
      END);

  c = 0;
  for(generation=0; generation < MaxGenerations; generation++)
  {
    Generation();
    
    #ifdef STATISTICS
    send_channel(to_master,
      DOUBLE, fitness[0],
      END);
    #endif
        
    switch(c)
    {
      case 0: src = from_south; dest = to_north; break;
      case 1: src = from_east; dest = to_west; break;
      case 2: src = from_north; dest = to_south; break;
      case 3: src = from_west; dest = to_east; break;
    }
    
    if (! (generation % 10))
    {                               /* migrate individuals between islands */
      if (((src == from_east || src == from_west) && my_col % 2) ||
	  ((src == from_south || src == from_north) && my_row % 2))
      {
	receive_channel(src,
	  DOUBLE, &fitness[NrOfIndiv-1],
	  INT_ARRAY, NrOfCities, population[NrOfIndiv-1],
	  END);
	  
	send_channel(dest,
	  DOUBLE, fitness[0],
	  INT_ARRAY, NrOfCities, population[0],
	  END);
      }
      else
      {
	send_channel(dest,
	  DOUBLE, fitness[0],
	  INT_ARRAY, NrOfCities, population[0],
	  END);
	  
	receive_channel(src,
	  DOUBLE, &fitness[NrOfIndiv-1],
	  INT_ARRAY, NrOfCities, population[NrOfIndiv-1],
	  END);
      }
    }

    c = ((c+1) % 4);
  }

  send_channel(to_master,
    DOUBLE_ARRAY, NrOfIndiv, fitness,
    END);
    
  for (p = 0; p < NrOfIndiv; p++)
    send_channel(to_master,
      INT_ARRAY, NrOfCities, population[p],
      END);
}

