/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* DDS user functions.                                                     */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.03.1995  Droux       Created                                         */
/* 03.05.1995  Droux       Added dds_create_nv()                           */
/* 07.05.1995  Droux       Added dds_store_to_channel_nv()                 */
/* 08.05.1995  Droux       dds_init replaced by dds_set                    */
/* 08.05.1995  Droux       Inverted parameters of dds_copy                 */
/* 08.05.1995  Droux       dds_store replaced by dds_get                   */
/* 16.05.1995  Droux       Ported to the SP-2                              */
/*-------------------------------------------------------------------------*/
/* filename: dds.h                                                         */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#ifndef PPC_DDS_H
#define PPC_DDS_H

#include <stdio.h>
#include <ppc/csp/types.h>

#ifndef DDS_MGMT_IMPL
#ifdef DDS_IMPL
typedef int channel;
#define CHANNEL_TYPE_DEFINED
#endif
#endif

/*-------------------------------------------------------------------------*/
/* String equivalent for common definitions (types, distribution, ...)     */
/*-------------------------------------------------------------------------*/
#ifndef DDS_DIR_IMPL
extern char *types_names_str[];
extern char *dds_distribution_str[];
#endif

/*-------------------------------------------------------------------------*/
/* DDS_DISTRIBUTION:                                                       */
/* Allowed data distribution schemes.                                      */
/*-------------------------------------------------------------------------*/
typedef enum 
{
  DDS_REPLICATED,
  DDS_NB_DISTRIBUTIONS,
} DDS_DISTRIBUTION;

/*-------------------------------------------------------------------------*/
/* dds_object_entry:                                                       */
/* Used by the directory functions.                                        */
/*-------------------------------------------------------------------------*/

#define DDS_MAX_DIM 4                      /* maximum number of dimensions */
#define DDS_MAX_NAME_LEN 20                  /* maximum object name length */

typedef struct _dds_object_entry
{
  char name[DDS_MAX_NAME_LEN];
  types_names type;
  DDS_DISTRIBUTION distribution;
  int nb_dim;
  int dim[DDS_MAX_DIM];
} dds_object_entry;

/*-------------------------------------------------------------------------*/
/* dds_clear_database:                                                     */
/* Clear the DDS object database. The memory used by the objects is        */
/* released.                                                               */
/*-------------------------------------------------------------------------*/
int dds_clear_database (void);

/*-------------------------------------------------------------------------*/
/* dds_create:                                                             */
/* Create a new DDS object.                                                */
/*-------------------------------------------------------------------------*/
#ifndef DDS_MGMT_IMPL
#ifndef SP2
int dds_create (
  char *name,
  types_names type,
  DDS_DISTRIBUTION distribution,
  int dim, 
  ...
);
#endif
#endif

/*-------------------------------------------------------------------------*/
/* dds_create_nv:                                                          */
/* The same dds_create() but without variable parameters.                  */
/*-------------------------------------------------------------------------*/
int dds_create_nv (
  char *name,
  types_names type,
  DDS_DISTRIBUTION distribution,
  int nb_dim,                                /* number of elements in dims */
  int *dims                                                  /* dimensions */
);

/*-------------------------------------------------------------------------*/
/* dds_delete:                                                             */
/* Delete a DDS object. The memory used by the object is released.         */
/*-------------------------------------------------------------------------*/
int dds_delete (
  char *name
);

/*-------------------------------------------------------------------------*/
/* dds_copy:                                                               */
/* Copy a DDS object into another one. The new object is created. If it    */
/* already exist, it is re-allocated if the type or the dimensions are not */
/* compatible with the source.                                             */
/*-------------------------------------------------------------------------*/
int dds_copy (
  char *src,
  char *dest
);

/*-------------------------------------------------------------------------*/
/* dds_set_from_stream:                                                    */
/* Initialize the specified object from the given stream.                  */
/*-------------------------------------------------------------------------*/
#ifndef DDS_MGMT_IMPL
#ifndef SP2
int dds_set_from_stream (
  char *name,
  FILE *stream,
  int idx,
  ...
);
#endif
#endif

/*-------------------------------------------------------------------------*/
/* dds_set_from_channel:                                                   */
/* Initialize the specified object from the given channel.                 */
/*-------------------------------------------------------------------------*/
#ifndef DDS_MGMT_IMPL
#ifndef SP2
int dds_set_from_channel (
  char *name,
  channel chan,
  int idx,
  ...
);
#endif
#endif

/*-------------------------------------------------------------------------*/
/* dds_set:                                                                */
/* Initialize a DDS object with a specified value.                         */
/*-------------------------------------------------------------------------*/
#ifndef DDS_MGMT_IMPL
#ifndef SP2
int dds_set (
  char *name,
  void *data,
  int idx,
  ...
);
#endif
#endif

/*-------------------------------------------------------------------------*/
/* dds_get_to_stream:                                                      */
/* Fed the value of a DDS object into a stream.                            */
/*-------------------------------------------------------------------------*/
#ifndef DDS_MGMT_IMPL
#ifndef SP2
int dds_get_to_stream (
  char *name,
  FILE *stream,
  int idx,
  ...
);
#endif
#endif

/*-------------------------------------------------------------------------*/
/* dds_get_to_channel:                                                     */
/* Fed the value of a DDS object into a channel.                           */
/*-------------------------------------------------------------------------*/
#ifndef DDS_MGMT_IMPL
#ifndef SP2
int dds_get_to_channel (
  char *name,
  channel chan,
  int idx,
  ...
);
#endif
#endif

/*-------------------------------------------------------------------------*/
/* dds_get_to_channel_nv:                                                  */
/* Same as dds_get_to_channel() but without variable parameters. In        */
/* addition, this version of the function allows to define functions which */
/* have to be called before and after a data transmission.                 */
/* Such functions must be defined as follows:                              */
/*   void before_or_after (                                                */
/*     types_names data_type, int size, void *custom_data);                */
/*-------------------------------------------------------------------------*/
#ifndef SP2
int dds_get_to_channel_nv (
  char *name,                                               /* object name */
  channel chan,                                  /* channel to destination */
  int nb_idx,                                 /* number of elements in idx */
  int *idx,                                 /* indices to access sub-array */
  void (* before_func) (types_names, int, void *),       /* to call before */
  void *before_data,                       /* value to pass to before_func */
  void (* after_func) (types_names, int, void *),         /* to call after */
  void *after_data                          /* value to pass to after_func */
);
#endif

/*-------------------------------------------------------------------------*/
/* dds_get:                                                                */
/* Store the value of a DDS object in the given data location.             */
/*-------------------------------------------------------------------------*/
#ifndef DDS_MGMT_IMPL
#ifndef SP2
int dds_get (
  char *name,
  void *data,
  int idx,
  ...
);
#endif
#endif

/*-------------------------------------------------------------------------*/
/* dds_directory_init:                                                     */
/* Must be called before reading the DDS objects directory.                */
/*-------------------------------------------------------------------------*/
void dds_directory_init (void);

/*-------------------------------------------------------------------------*/
/* dds_directory_get_entry:                                                */
/* Get the next DDS object entry. Returns non-zero if no more objects are  */
/* available, 0 otherwise.                                                 */
/*-------------------------------------------------------------------------*/
int dds_directory_get_entry (
  dds_object_entry *entry
);

/*-------------------------------------------------------------------------*/
/* dds_object_prop:                                                        */
/* Return the properties of the specified object. Returns -1 if the        */
/* object cannot be found, 0 otherwise.                                    */
/*-------------------------------------------------------------------------*/
int dds_object_prop (
  char *name,
  dds_object_entry *entry
);

/*-------------------------------------------------------------------------*/
/* dds_object_ptr:                                                         */
/* Return a pointer to the specified object.                               */
/* For master and slaves.                                                  */
/*-------------------------------------------------------------------------*/
void *dds_object_ptr (
  char *name
);

/*-------------------------------------------------------------------------*/
/* dds_receive_name:                                                       */
/* Get an object name from the master. Each slave must call this function  */
/* for each dds_send_name master call.                                     */
/* The returned pointer is valid until the next call to this function.     */
/* For slaves only.                                                        */
/*-------------------------------------------------------------------------*/
char *dds_receive_name (void);

/*-------------------------------------------------------------------------*/
/* dds_send_name:                                                          */
/* Send object name to slaves. Each slave must call the function           */
/* dds_receive_name for each call to this function.                        */
/* For master only.                                                        */
/*-------------------------------------------------------------------------*/
void dds_send_name (
  char *name
);

#endif
