/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Client-Server Library.                                                  */
/* This module implements the UNIX server-side connection functions of the */
/* PPC client-server library.                                              */
/* The Functions implemented here are declared in the header file cslib.h  */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 09.02.1994  Droux       Created                                         */
/* 08.07.1994  Droux       New structure of source directory               */
/* 25.07.1994  Droux       to_client_fd & to_client_fd static for PACT-C   */
/* 25.07.1994  Droux       new_cslib_channel not called under PACT-C       */
/* 10.09.1994  Droux       Added get_sockets()                             */
/* 10.09.1994  Droux       Added unixconn.h inclusion                      */
/* 16.03.1995  Droux       Server sends its address to the client          */
/* 01.05.1995  Droux       Local functions are declared as static          */
/* 22.05.1995  Droux       Repeat accept() if EINTR for the SP2            */
/*-------------------------------------------------------------------------*/
/* filename: unixconn.c                                                    */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   SunOS Release 4.1.3, SPARC                                            */
/*   SunOS Release 5.3 (Solaris 2), SPARC                                  */
/*   Linux 0.99.pl9, Intel Computers                                       */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#ifdef RS6K
  #define _BSD 44
#endif

#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <netdb.h>
#include <ppc/cslib.h>
#include "unixconn.h"

#ifdef PVM_CSLIB
  #include <ppc/csp/chanpvmuh.h>
#elif defined(PGON_CSLIB)
  #include <ppc/csp/chanparagon.h>
#else
  #include "chanunixuh.h"
#endif

#ifdef LINUX
  #define SOMAXCONN 5
#endif

static int to_client_fd;
static int from_client_fd;

/*-------------------------------------------------------------------------*/
/* my_bzero:                                                         LOCAL */
/* Fill the given buffer with the specified number of 0 bytes.             */
/*-------------------------------------------------------------------------*/
static void my_bzero (void *buffer, int length)
{
  char *b = (char *)buffer;
  int curr_char;
  
  for (curr_char = 0; curr_char < length; curr_char++)
    b[curr_char] = 0;
}

/*--------------------------------------------------------------------------*/
/* copy_struct:                                                       LOCAL */
/* copy the <len> first bytes of <source> to <dest>.                        */
/*--------------------------------------------------------------------------*/
static void copy_struct (void *source, void *dest, int len)
{
  char *s = (char *)source;
  char *d = (char *)dest;

  while (len)
  {
    *d = *s;
    d++;
    s++;
    len--;
  }
}

/*-------------------------------------------------------------------------*/
/* create_socket:                                                    LOCAL */
/* Creates a TCP socket based connection between the client and the        */
/* server. Returns a file descriptor corresponding to the socket or -1     */
/* if an error occured.                                                    */
/*-------------------------------------------------------------------------*/
static int create_socket (void)
{
  int socket_fd;                                      /* socket descriptor */
  int fd;                                   /* effective socket descriptor */
  struct sockaddr_in sock_addr;                          /* socket address */
  int servlen;
  int port_num;
  char port_name[PORT_NAME_LEN];

  if ((socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) /* create sockets */
    return -1;
  
  my_bzero((char *) &sock_addr,sizeof(sock_addr));
  sock_addr.sin_family = AF_INET;          /* initializes socket addresses */ 
  sock_addr.sin_port = 0;
  sock_addr.sin_addr.s_addr = INADDR_ANY;
  servlen = sizeof(struct sockaddr_in);
  
  if (bind(socket_fd, (struct sockaddr *)&sock_addr, servlen))
    return -1;
    
  if (getsockname(socket_fd, (struct sockaddr *)&sock_addr, &servlen))
    return -1;
  port_num = ntohs(sock_addr.sin_port);      /* send port number to client */
  sprintf(port_name, "%d", port_num);
  writen(1, port_name, PORT_NAME_LEN);

  if (listen(socket_fd, SOMAXCONN) < 0)      /* wait for client connection */
    return -1;

  do {
    fd = accept(socket_fd,(struct sockaddr *)&sock_addr, &servlen);
  } while (fd == -1 && errno == EINTR);
    
  return fd;
}

/*-------------------------------------------------------------------------*/
/* send_server_addr:                                                       */
/* Send the server address to the client.                                  */
/*-------------------------------------------------------------------------*/
void send_server_addr (void)
{
  char to_be_sent [MAX_SERVER_ADDR_LEN];
  char host_name [MAXHOSTNAMELEN];
  struct hostent *hinfo;
  struct in_addr hnet_addr;
  char host_addr;
  
  gethostname(host_name, MAXHOSTNAMELEN);
  hinfo = gethostbyname(host_name);
  if (hinfo == NULL)
    syslog(LOG_ERR, "error in gethostbyname()");
  my_bzero (&hnet_addr, hinfo->h_length);
  copy_struct (*hinfo->h_addr_list, &hnet_addr, hinfo->h_length); 
  strcpy(to_be_sent, inet_ntoa(hnet_addr));
  
  writen(1, to_be_sent, MAX_SERVER_ADDR_LEN);
}

/*-------------------------------------------------------------------------*/
/* connect_to_client:                                                      */
/* This function must be the first function called when a PPC program is   */
/* used as a CS-Lib server. It returns two channels that must be used      */
/* to send and receive messages to and from the client.                    */
/* Returns -1 if the connection failed, 0 otherwise.                       */
/*-------------------------------------------------------------------------*/
int connect_to_client (
  channel *to_client,
  channel *from_client
)
{
  if (server_check_version(0, 1) == -1)    /* check protocol compatibility */
    return -1;
    
  send_server_addr();
                                    /* create the server to client channel */
  if ((to_client_fd = create_socket()) == -1)  
    return -1;   
                                    /* create the client to server channel */
  if ((from_client_fd = create_socket()) == -1)  
    return -1;
  
  #ifndef PACTC_CSLIB
  new_cslib_channel(from_client_fd, to_client_fd,       /* update channels */
    from_client, to_client);
  #endif

  return 0;
}

/*-------------------------------------------------------------------------*/
/* disconnect_from_client:                                                 */
/* Must be called before terminating the program when the PPC program is   */
/* used as a server application.                                           */
/*-------------------------------------------------------------------------*/
void disconnect_from_client (void)
{
}

/*-------------------------------------------------------------------------*/
/* get_sockets:                                                            */
/* Returns the TCP sockets descriptors corresponding to the channels       */
/* to and from the client.                                                 */
/*-------------------------------------------------------------------------*/
void get_sockets (
  int *to_client_socket,
  int *from_client_socket
)
{
  *to_client_socket = to_client_fd;
  *from_client_socket = from_client_fd;
}
