/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Paragon channels implementation.                                        */
/* This file implements the functions that are defined in the header file  */
/* chan.h and chanpriv.h.                                                  */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 06.01.1995  Droux       Created                                         */
/* 01.05.1995  Droux       Added complex type                              */
/* 01.05.1995  Droux       Local functions are declared as static          */
/* 03.05.1995  Droux       Added boolean type                              */
/*-------------------------------------------------------------------------*/
/* filename: chanparagon.c                                                 */
/* Tested on :                                                             */
/*   Paragon XP/S OSF/1 1.0.4 R1.2.7                                       */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define chan_impl

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <varargs.h>
#include <syslog.h>

#include <nx.h>

#include <ppc/csp/complex.h>
#include <ppc/csp/chanpriv.h>
#include <ppc/csp/chanparagon.h>
#include "chanservunix.h"

/* Two kind of channels are accessed through this module: PGON channels    */
/* and CS-Lib channels. When an operation on a channel is requested by the */
/* PPC program, this module must be able to determinate if this channel    */
/* is a PGON or a CS-Lib channel. In order to allows this distinction,     */
/* we define the CS-Lib channel ids as constant after the                  */
/* PGON channel possible values. This is possible since only two           */
/* two CS-Lib channels can be opened by a server program.                  */
#define FROM_CLIENT_CHAN_ID PARAGON_CHANNELS
#define TO_CLIENT_CHAN_ID   (FROM_CLIENT_CHAN_ID+1)

typedef struct            /* structure containing the Paragon channels ... */
{                                                       /* ... properties. */
  int free;                                              /* 1 if free slot */
  long type;                                /* nxlib type, here channel id */
  long dest_node;    /* dest. node, only for output channels, here task id */
} uh_pgon_channel_descr;

static uh_pgon_channel_descr uh_pgon_channels [PARAGON_CHANNELS];
static long ppc_ptype;                                /* application ptype */

/*-------------------------------------------------------------------------*/
/* chan_init:                                                              */
/* Initializes the channels implementation. This function is never called  */
/* directly by the programmer.                                             */
/*-------------------------------------------------------------------------*/
void chan_init (void)
{
  int c;                                                        /* counter */
  
  for (c = 0; c < PARAGON_CHANNELS; c++)
    uh_pgon_channels[c].free = 1;                    /* all slots are free */

  ppc_ptype = myptype();
}

/*-------------------------------------------------------------------------*/
/* par_channel:                                                            */
/* Creates a new channel to communicate between two parallel threads.      */
/*-------------------------------------------------------------------------*/
channel par_channel (void)
{
  return -1;
}

/*-------------------------------------------------------------------------*/
/* chan_close:                                                             */
/* Close the given channel.                                                */
/*-------------------------------------------------------------------------*/
void chan_close (channel id)
{
}

/*-------------------------------------------------------------------------*/
/* pgon_chan_write:                                                  LOCAL */
/* Low-Level Paragon channel read.                                         */
/*-------------------------------------------------------------------------*/
static int pgon_chan_write (channel id, void *vbuf, int nitems, int type)
{
  csend(uh_pgon_channels[id].type, (char *)vbuf, nitems*type_size[type],
    uh_pgon_channels[id].dest_node, ppc_ptype);
}

/*-------------------------------------------------------------------------*/
/* pgon_chan_read:                                                   LOCAL */
/* Low-Level Paragon channel read.                                         */
/*-------------------------------------------------------------------------*/
static int pgon_chan_read (channel id, void *vbuf, int nitems, int type)
{
  crecv(uh_pgon_channels[id].type, (char *)vbuf,
    nitems*type_size[type]);
}

/*-------------------------------------------------------------------------*/
/* send_channel:                                                           */
/* Send the given data on the given channel.                               */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value>                             */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    send_channel(c, INT, a, FLOAT_ARRAY, 100, b, END);                   */
/*-------------------------------------------------------------------------*/
void send_channel (va_alist)
va_dcl
{
  va_list pvar;                  /* needed for variable arguements parsing */
  int act_type;                                        /* actual item type */
  int chan_id;                                   
  int len;
  int nitems;
  static int (* writef) (channel, void *, int, int);
  
  va_start(pvar);
  chan_id = va_arg(pvar, channel);
  
  if (chan_id == TO_CLIENT_CHAN_ID)
  {
    chan_id = TO_CLIENT;
    writef = (int (*)(channel, void *, int, int)) servunix_chan_write;
  }
  else
    writef = (int (*)(channel, void *, int, int)) pgon_chan_write;

  act_type = va_arg(pvar, int);
  while(act_type != END)
  {
    switch(act_type)
    {
      case BOOLEAN:
      case INT:
      {
	int val = va_arg(pvar, int);
	writef(chan_id, &val, 1, INT);
	break;
      }
      case SHORT:
      {
	short val = va_arg(pvar, int);
	writef(chan_id, &val, 1, SHORT);
	break;
      }
      case UNSIGNED_SHORT:
      {
	unsigned short val = va_arg(pvar, unsigned int);
	writef(chan_id, &val, 1, UNSIGNED_SHORT);
	break;
      }
      case LONG:
      {
	long val = va_arg(pvar, long);
	writef(chan_id, &val, 1, LONG);
	break;
      }
      case UNSIGNED_LONG:
      {
	unsigned long val = va_arg(pvar, unsigned long);
	writef(chan_id, &val, 1, UNSIGNED_LONG);
	break;
      }
      case FLOAT:
      {
	float val = va_arg(pvar, double);
	writef(chan_id, &val, 1, FLOAT);
	break;
      }
      case DOUBLE:
      {
	double val = va_arg(pvar, double);
	writef(chan_id, &val, 1, DOUBLE);
	break;
      }
      case CHAR:
      {
	char val = va_arg(pvar, int);
	writef(chan_id, &val, 1, CHAR);
	break;
      }
      case COMPLEX:
      {
	ppc_complex val;
	val = va_arg(pvar, ppc_complex);
	writef(chan_id, &val, 1, COMPLEX);
	break;
      }
      case INT_ARRAY:
      case SHORT_ARRAY:
      case UNSIGNED_SHORT_ARRAY:
      case LONG_ARRAY:
      case UNSIGNED_LONG_ARRAY:
      case FLOAT_ARRAY:
      case DOUBLE_ARRAY:
      case CHAR_ARRAY:
      case COMPLEX_ARRAY:
      case BOOLEAN_ARRAY:
	nitems = va_arg(pvar, int);
	writef(chan_id, va_arg(pvar, void *), nitems, act_type);
	break; 
      default: 
        fprintf(stderr, "error: unknown type identifier\n");
	exit(1);
    }
    act_type = va_arg(pvar, int);
  }
}

/*-------------------------------------------------------------------------*/
/* receive_channel:                                                        */
/* Get the given data from the specified channel.                          */
/* The parameters list is a list of tuples. A tuple can be:                */
/*    1. atomic value:    <type_spec>, <value address>                     */
/*    2. array:           <type_spec>, <nitems>, <value address>           */
/*    3. termination:     END                                              */
/* For example:                                                            */
/*    int a;                                                               */
/*    float b [100];                                                       */
/*    channel c;                                                           */
/*    receive_channel(c, INT, &a, FLOAT_ARRAY, 100, b, END);               */
/*-------------------------------------------------------------------------*/
void receive_channel (va_alist)
va_dcl
{
  va_list pvar;                  /* needed for variable arguements parsing */
  int act_type;                                        /* actual item type */
  int chan_id;                                    
  int nitems;
  static int (* readf) (channel, void *, int, int);      
  
  va_start(pvar);
  chan_id = va_arg(pvar, channel);
  
  if (chan_id == FROM_CLIENT_CHAN_ID)
  {
    chan_id = FROM_CLIENT;
    readf = (int (*)(channel, void *, int, int)) servunix_chan_read;
  }
  else
    readf = (int (*)(channel, void *, int, int)) pgon_chan_read;

  act_type = va_arg(pvar, int);
  while(act_type != END)
  {
    switch(act_type)
    {
      case INT:
      case SHORT:
      case UNSIGNED_SHORT:
      case LONG:
      case UNSIGNED_LONG:
      case FLOAT:
      case DOUBLE:
      case CHAR:
      case COMPLEX:
      case BOOLEAN:
        nitems = 1;
	break;
      case INT_ARRAY:
      case SHORT_ARRAY:
      case UNSIGNED_SHORT_ARRAY:
      case LONG_ARRAY:
      case UNSIGNED_LONG_ARRAY:
      case FLOAT_ARRAY:
      case DOUBLE_ARRAY:
      case CHAR_ARRAY:
      case COMPLEX_ARRAY:
      case BOOLEAN_ARRAY:
        nitems = va_arg(pvar, int);
	break;
      default: 
        fprintf(stderr, "error: unknown type identifier\n");
	exit(1);
    }
    readf(chan_id, va_arg(pvar, void *), nitems, act_type);
    act_type = va_arg(pvar, int);
  }
}

/*-------------------------------------------------------------------------*/
/* new_pgon_channel:                                                       */
/* Declares a new Paragon channel.                                         */
/*-------------------------------------------------------------------------*/
channel new_pgon_channel (long dest_node, long type)
{
  int c;                                                        /* counter */
  int chan_id;                                    
  
  for (c = 0; (c < PARAGON_CHANNELS) && (!uh_pgon_channels[c].free); c++);
  chan_id = c;
  
  uh_pgon_channels[chan_id].free = 0;
  uh_pgon_channels[chan_id].type = type;
  uh_pgon_channels[chan_id].dest_node = dest_node;
  
  return chan_id;
}

/*-------------------------------------------------------------------------*/
/* new_cslib_channel:                                                      */
/* Return the two channels from and to the client application and          */
/* initialize them with the given values.                                  */
/* This function is called by the unixconn module.                         */
/*-------------------------------------------------------------------------*/
void new_cslib_channel(
  int from_client,
  int to_client,
  channel *from_client_chan,
  channel *to_client_chan
)
{
  servunix_init_channels(from_client, to_client);
  *from_client_chan = FROM_CLIENT_CHAN_ID;
  *to_client_chan = TO_CLIENT_CHAN_ID;
}

/*-------------------------------------------------------------------------*/
/* pgon_correspondent_node:                                                */
/* Return the correspondent node of the specified channel.                 */
/*-------------------------------------------------------------------------*/
int pgon_correspondent_node (channel chan)
{
  return uh_pgon_channels[chan].dest_node;
}

