/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* School of Engineering of Biel/Bienne, Switzerland.                      */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* Matrix utilities.                                                       */
/* Matrix allocation routines. These routines can be used to allocate and  */
/* free matrices local to a processor. They are used when one of the tasks */
/* of the network needs to allocate temporary data.                        */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 21.10.1994  Droux       Created                                         */
/* 12.01.1995  Droux       Defines NULL if unknown (useful for SUN4)       */
/*-------------------------------------------------------------------------*/
/* filename: matrixalloc.c                                                 */
/* Tested on :                                                             */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 School of Engineering of Biel/Bienne, Switzerland.   */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <ppc/matrixutil/matrixalloc.h>

#ifndef NULL
#define NULL ((void *)0)
#endif

/*-------------------------------------------------------------------------*/
/* MATALLOC_BODY:                                                    MACRO */
/* Structure of a matrix allocation routine.                               */
/*-------------------------------------------------------------------------*/
#define MATALLOC_BODY(type) { \
  int i; \
  type **mat; \
  int failed; \
  \
  mat = (type **)malloc(n * sizeof(type *)); \
  if (mat == NULL) \
    return NULL; \
  \
  for (i = 0, failed = 0; i < n && !failed; i++) \
  { \
    mat[i] = (type *)malloc(m * sizeof(type)); \
    failed = mat[i] == NULL; \
  } \
  \
  if (failed) \
  { \
    int row; \
    \
    for (row = 0; row < i; row++) \
      free(mat[row]); \
    free(mat); \
    \
    return NULL; \
  } \
 \
  return mat; \
}

/*-------------------------------------------------------------------------*/
/* MATFREE_BODY:                                                     MACRO */
/* Structure of a matrix free routine.                                     */
/*-------------------------------------------------------------------------*/
#define MATFREE_BODY(type) { \
  int i; \
  \
  for (i = 0; i < n; i++) \
    free(mat[i]); \
  \
  free(mat); \
}

/*-------------------------------------------------------------------------*/
/* dmatalloc:                                                              */
/* Allocates a n x m matrix of doubles.                                    */
/*-------------------------------------------------------------------------*/
double **dmatalloc (
  int n,
  int m
)
MATALLOC_BODY(double)

/*-------------------------------------------------------------------------*/
/* fmatalloc:                                                              */
/* Allocates a n x m matrix of floats.                                     */
/*-------------------------------------------------------------------------*/
float **fmatalloc (
  int n,
  int m
)
MATALLOC_BODY(float)

/*-------------------------------------------------------------------------*/
/* imatalloc:                                                              */
/* Allocates a n x m matrix of integers.                                   */
/*-------------------------------------------------------------------------*/
int **imatalloc (
  int n,
  int m
)
MATALLOC_BODY(int)

/*-------------------------------------------------------------------------*/
/* dmatfree:                                                               */
/* Free the memory allocated for a double matrix.                          */
/*-------------------------------------------------------------------------*/
void dmatfree (
  double **mat,
  int n
)
MATFREE_BODY(double)

/*-------------------------------------------------------------------------*/
/* fmatfree:                                                               */
/* Free the memory allocated for a float matrix.                           */
/*-------------------------------------------------------------------------*/
void fmatfree (
  float **mat,
  int n
)
MATFREE_BODY(float)

/*-------------------------------------------------------------------------*/
/* imatfree:                                                               */
/* Free the memory allocated for an int matrix.                            */
/*-------------------------------------------------------------------------*/
void imatfree (
  int **mat,
  int n
)
MATFREE_BODY(int)
