/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* DDS objects initialization.                                             */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.03.1995  Droux       Created                                         */
/* 04.05.1995  Droux       Fixed data distribution problem with            */
/*                         dds_init_from_channel() and init_from_stream()  */
/* 07.05.1995  Droux       Added dds_store_to_channel_nv()                 */
/* 08.05.1995  Droux       dds_init replaced by dds_set                    */
/*-------------------------------------------------------------------------*/
/* filename: ddsinit.c                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define DDS_IMPL
#define DDS_MGMT_IMPL
#define MKERNEL_IMPL

#include <stdlib.h>
#include <ppc/portable.h>
#include <ppc/parlib/mkernel.h>
#include <ppc/parlib/dds.h>
#include <ppc/parlib/ddspriv.h>
#include <ppc/csp/chan.h>
#include "ddsutil.h"

/*-------------------------------------------------------------------------*/
/* ddsi_init_scan_data:                                              LOCAL */
/* Scan the given data from the specified stream.                          */
/*-------------------------------------------------------------------------*/
static void ddsi_init_scan_data (
  void *value,
  int type,
  int act_dim,
  int nb_dim,
  int *dim,
  int nb_idx,
  int *idx,
  FILE *stream
)
{
  int c;

  if (act_dim == nb_dim-1)
  {
    ddsi_scan_vector(stream, type, dim[act_dim], value);
    for (c = 0; c < nb_slaves; c++)
      send_channel(to_slaves[c],
        type+INT_ARRAY, dim[act_dim], value,
	END);
    return;
  }
    
  {
    void **ptr = value;
    
    if (act_dim < nb_idx)
      ddsi_init_scan_data(ptr[idx[act_dim]], type, act_dim+1, 
        nb_dim, dim, nb_idx, idx, stream);
    else
    for (c = 0; c < dim[act_dim]; c++)
      ddsi_init_scan_data(ptr[c], type, act_dim+1, nb_dim, dim,
        nb_idx, idx, stream);
  }
}

/*-------------------------------------------------------------------------*/
/* dds_set_from_stream:                                                    */
/* Initialize the specified object from the given stream.                  */
/*-------------------------------------------------------------------------*/
#ifdef PACTC
int dds_set_from_stream (
  char *name,
  FILE *stream,
  int act_idx,
  ...
)
#else
int dds_set_from_stream (va_alist)
va_dcl
#endif
{
  #ifndef PACTC
  char *name;
  FILE *stream;
  int act_idx;
  #endif
  dds_object *object;
  int idx[DDS_MAX_DIM];
  int nb_idx;
  va_list pvar;
  int c;
  
  #ifdef PACTC
  va_start(pvar, act_idx);
  #else
  va_start(pvar);
  name = va_arg(pvar, char *);
  stream = va_arg(pvar, FILE *);
  act_idx = va_arg(pvar, int);
  #endif

  if ((object = ddsi_find_object(ddsi_master_get_first_object(), name)) 
      == NULL)
    return -1;                                         /* object not found */

  ddsi_master_send_req(DDSI_REQ_INIT);
  
  for (nb_idx = 0; act_idx != END; nb_idx++)                /* get indices */
  {
    idx[nb_idx] = act_idx;
    act_idx = va_arg(pvar, int);
  }

  for (c = 0; c < nb_slaves; c++)
    send_channel(to_slaves[c],
      STRING, DDS_MAX_NAME_LEN, name,
      INT_ARRAY, DDS_MAX_DIM, idx,
      INT, nb_idx,
      END);
  
  ddsi_init_scan_data(object->value, object->type, 0, 
    object->nb_dim, object->dim, nb_idx, idx, stream);
  
  DDSI_GET_ACK;

  return 0;
}

/*-------------------------------------------------------------------------*/
/* ddsi_init_receive_data:                                           LOCAL */
/* Initialize object using data from the specified channel.                */
/*-------------------------------------------------------------------------*/
static void ddsi_init_receive_data (
  void *value,
  int type,
  int act_dim,
  int nb_dim,
  int *dim,
  int nb_idx,
  int *idx,
  channel chan
)
{
  int c;

  if (act_dim == nb_dim-1)
  {
    receive_channel(chan,
      type+INT_ARRAY, dim[act_dim], value,
      END);
    for (c = 0; c < nb_slaves; c++)
      send_channel(to_slaves[c],
        type+INT_ARRAY, dim[act_dim], value,
	END);
    return;
  }
    
  {
    void **ptr = value;
    
    if (act_dim < nb_idx)
      ddsi_init_receive_data(ptr[idx[act_dim]], type, act_dim+1, 
        nb_dim, dim, nb_idx, idx, chan);
    else
      for (c = 0; c < dim[act_dim]; c++)
	ddsi_init_receive_data(ptr[c], type, act_dim+1, nb_dim, dim,
	  nb_idx, idx, chan);
  }
}

/*-------------------------------------------------------------------------*/
/* dds_set_from_channel:                                                   */
/* Initialize the specified object from the given channel.                 */
/*-------------------------------------------------------------------------*/
#ifdef PACTC
int dds_set_from_channel (
  char *name,
  channel chan,
  int act_idx,
  ...
)
#else
int dds_set_from_channel (va_alist)
va_dcl
#endif
{
  #ifndef PACTC
  char *name;
  channel chan;
  int act_idx;
  #endif
  dds_object *object;
  int idx[DDS_MAX_DIM];
  int nb_idx;
  va_list pvar;
  int c;
  
  #ifdef PACTC
  va_start(pvar, act_idx);
  #else
  va_start(pvar);
  name = va_arg(pvar, char *);
  chan = va_arg(pvar, channel);
  act_idx = va_arg(pvar, int);
  #endif
  
  if ((object = ddsi_find_object(ddsi_master_get_first_object(), name)) 
      == NULL)
    return -1;                                         /* object not found */

  ddsi_master_send_req(DDSI_REQ_INIT);
  
  for (nb_idx = 0; act_idx != END; nb_idx++)                /* get indices */
  {
    idx[nb_idx] = act_idx;
    act_idx = va_arg(pvar, int);
  }

  for (c = 0; c < nb_slaves; c++)
    send_channel(to_slaves[c],
      STRING, DDS_MAX_NAME_LEN, name,
      INT_ARRAY, DDS_MAX_DIM, idx,
      INT, nb_idx,
      END);
  
  ddsi_init_receive_data(object->value, object->type, 0, 
    object->nb_dim, object->dim, nb_idx, idx, chan);
  
  DDSI_GET_ACK;

  return 0;
}

/*-------------------------------------------------------------------------*/
/* ddsi_init_send_data:                                              LOCAL */
/* Send the given data to the slaves.                                      */
/*-------------------------------------------------------------------------*/
static void ddsi_init_send_data (
  void *value,
  int type,
  int act_dim,
  int nb_dim,
  int *dim
)
{
  int c;

  if (act_dim == nb_dim-1)
  {
    for (c = 0; c < nb_slaves; c++)
      send_channel(to_slaves[c],
        type+INT_ARRAY, dim[act_dim], value,
	END);
    return;
  }
    
  {
    void **ptr = value;
    
    for (c = 0; c < dim[act_dim]; c++)
      ddsi_init_send_data(ptr[c], type, act_dim+1, nb_dim, dim);
  }
}

/*-------------------------------------------------------------------------*/
/* dds_set:                                                                */
/* Initialize a DDS object with a specified value.                         */
/*-------------------------------------------------------------------------*/
#ifdef PACTC
int dds_set (
  char *name,
  void *data,
  int act_idx,
  ...
)
#else
int dds_set (va_alist)
va_dcl
#endif
{
  #ifndef PACTC
  char *name;
  void *data;
  int act_idx;
  #endif
  dds_object *object;
  int c;
  int idx[DDS_MAX_DIM];
  int nb_idx;
  va_list pvar;
  
  #ifdef PACTC
  va_start(pvar, act_idx);
  #else
  va_start(pvar);
  name = va_arg(pvar, char *);
  data = va_arg(pvar, void *);
  act_idx = va_arg(pvar, int);
  #endif

  if ((object = ddsi_find_object(ddsi_master_get_first_object(), name)) 
      == NULL)
    return -1;                                         /* object not found */

  ddsi_master_send_req(DDSI_REQ_INIT);
  
  for (nb_idx = 0; act_idx != END; nb_idx++)                /* get indices */
  {
    idx[nb_idx] = act_idx;
    act_idx = va_arg(pvar, int);
  }

  for (c = 0; c < nb_slaves; c++)
    send_channel(to_slaves[c],
      STRING, DDS_MAX_NAME_LEN, name,
      INT_ARRAY, DDS_MAX_DIM, idx,
      INT, nb_idx,
      END);
  
  ddsi_init_send_data(data, object->type, 0, 
    object->nb_dim-nb_idx, object->dim+nb_idx);
  
  ddsi_value_copy(                      /* keep a local copy of the object */
    data,
    object->value,
    object->type,
    0,
    object->nb_dim,
    object->dim,
    nb_idx,
    idx,
    DDSI_COPY_IDX_DEST);

  DDSI_GET_ACK;

  return 0;
}

