/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* DDS database management.                                                */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.03.1995  Droux       Created                                         */
/* 03.05.1995  Droux       Added dds_create_nv()                           */
/* 03.05.1995  Droux       Do not create new object if already exist       */
/* 06.05.1995  Droux       Added complex and boolean types                 */
/* 08.05.1995  Droux       Inverted parameters of dds_copy                 */
/*-------------------------------------------------------------------------*/
/* filename: ddsmgmt.c                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define DDS_IMPL
#define DDS_MGMT_IMPL
#define MKERNEL_IMPL

#include <stdlib.h>
#include <ppc/portable.h>
#include <ppc/parlib/mkernel.h>
#include <ppc/parlib/dds.h>
#include <ppc/parlib/ddspriv.h>
#include <ppc/csp/chan.h>
#include "ddsutil.h"

/*-------------------------------------------------------------------------*/
/* dds_clear_database:                                                     */
/* Clear the DDS object database. The memory used by the objects is        */
/* released.                                                               */
/*-------------------------------------------------------------------------*/
int dds_clear_database (void)
{
  ddsi_master_send_req(DDSI_REQ_CLEAR);
  ddsi_master_clear_database();
  DDSI_GET_ACK;
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* dds_create_nv:                                                          */
/* The same dds_create() but without variable parameters.                  */
/*-------------------------------------------------------------------------*/
int dds_create_nv (
  char *name,
  types_names type,
  DDS_DISTRIBUTION distribution,
  int nb_dim,                                /* number of elements in dims */
  int *dims                                                  /* dimensions */
)
{
  dds_object *new;
  int act_dim;

  if ((type != INT) &&                         /* check data type validity */
      (type != SHORT) &&
      (type != UNSIGNED_SHORT) &&
      (type != LONG) &&
      (type != UNSIGNED_LONG) &&
      (type != FLOAT) &&
      (type != DOUBLE) &&
      (type != CHAR) &&
      (type != COMPLEX) &&
      (type != BOOLEAN))
  {
    fprintf(stderr, "dds_create: unknown data type\n");
    return -1;
  }
  
  {                                      /* check if object already exists */
    dds_object *old;
    
    if ((old = ddsi_find_object(ddsi_master_get_first_object(), name)) 
	!= NULL)
    {                         /* object already exist, check if compatible */
      int compatible;
      int c;
      
      compatible = (old->type == type) &&
	(old->distribution == distribution) &&
	(old->nb_dim == nb_dim);
      for (c = 0; c < old->nb_dim; c++)
	compatible = compatible && (old->dim[c] == dims[c]);
  
      if (!compatible)
      {  /* objects are not compatible, destination must be re-initialized */
	dds_delete(name);
      }
      else
      {                          /* object already exist and is compatible */
        return 0;
      }
    }
  }
  
  ddsi_master_send_req(DDSI_REQ_CREATE);

  new = (dds_object *)malloc(sizeof(dds_object));

  strcpy(new->name, name);
  new->type = type;
  new->distribution = distribution;
  new->nb_dim = 0;
  
  for (act_dim = 0; act_dim < nb_dim; act_dim++)
    new->dim[act_dim] = dims[act_dim];
  new->nb_dim = nb_dim;
  
  new->value = ddsi_value_allocate(
    new->type,
    0,
    new->nb_dim,
    new->dim);

  ddsi_master_send_object_prop(new);   /* send object properties to slaves */
  ddsi_master_add_object(new);               /* add object to objects list */
  
  DDSI_GET_ACK;

  return 0;
}

/*-------------------------------------------------------------------------*/
/* dds_create:                                                             */
/* Create a new DDS object.                                                */
/*-------------------------------------------------------------------------*/
#ifdef PACTC
int dds_create (
  char *name,
  types_names type,
  DDS_DISTRIBUTION distribution,
  int dim, 
  ...
)
#else
int dds_create (va_alist)
va_dcl
#endif
{
  va_list pvar;
  #ifndef PACTC
  char *name;
  types_names type;
  DDS_DISTRIBUTION distribution;
  int dim;
  #endif
  
  int nb_dims;
  int dims[DDS_MAX_DIM];
  int act_dim;

  #ifdef PACTC
  va_start(pvar, dim);
  #else
  va_start(pvar);
  name = va_arg(pvar, char *);
  type = va_arg(pvar, types_names);
  distribution = va_arg(pvar, DDS_DISTRIBUTION);
  dim = va_arg(pvar, int);
  #endif
  
  nb_dims = 0;
  act_dim = dim;
  do
  {
    dims[nb_dims] = act_dim;
    nb_dims++;
    act_dim = va_arg(pvar, int);
  } while (act_dim != END);
  
  return dds_create_nv(name, type, distribution, nb_dims, dims);
}

/*-------------------------------------------------------------------------*/
/* dds_delete:                                                             */
/* Delete a DDS object. The memory used by the object is released.         */
/*-------------------------------------------------------------------------*/
int dds_delete (
  char *name
)
{
  dds_object *object;
  int c;
  
  if ((object = ddsi_find_object(ddsi_master_get_first_object(), name)) == 
      NULL)
    return -1;                                         /* object not found */
  
  ddsi_master_send_req(DDSI_REQ_DELETE);
  for (c = 0; c < nb_slaves; c++)
    send_channel(to_slaves[c],
      STRING, DDS_MAX_NAME_LEN, object->name,
      END);
  
  ddsi_master_remove_object(object);
  
  DDSI_GET_ACK;
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* ddsi_duplicate:                                                   LOCAL */
/* Duplicate an object. The destination is allocated.                      */
/*-------------------------------------------------------------------------*/
static dds_object *ddsi_duplicate (
  dds_object *src_object,                                 /* source object */
  char *dest                                           /* destination name */
)
{
  dds_object *dest_object;
  int c;

  ddsi_master_send_req(DDSI_REQ_CREATE);
  
  dest_object = (dds_object *)malloc(sizeof(dds_object));
  strcpy(dest_object->name, dest);
  dest_object->type = src_object->type;
  dest_object->distribution = src_object->distribution;
  dest_object->nb_dim = src_object->nb_dim;
  for (c = 0; c < src_object->nb_dim; c++)
    dest_object->dim[c] = src_object->dim[c];
    
  ddsi_master_send_object_prop(dest_object);
  ddsi_master_add_object(dest_object);
  
  dest_object->value = ddsi_value_allocate(
    src_object->type,
    0,
    src_object->nb_dim,
    src_object->dim);

  ddsi_value_copy(
    src_object->value, 
    dest_object->value,
    src_object->type,
    0,
    src_object->nb_dim,
    src_object->dim,
    0, NULL,
    DDSI_COPY_IDX_NONE);

  DDSI_GET_ACK;
  
  return dest_object;
}

/*-------------------------------------------------------------------------*/
/* dds_copy:                                                               */
/* Copy a DDS object into another one. The new object is created. If it    */
/* already exist, it is re-allocated if the type or the dimensions are not */
/* compatible with the source.                                             */
/*-------------------------------------------------------------------------*/
int dds_copy (
  char *src,
  char *dest
)
{
  dds_object *src_object;
  dds_object *dest_object;
  int c;
  
  if ((src_object = ddsi_find_object(ddsi_master_get_first_object(), src)) 
      == NULL)
    return -1;                                     /* src object not found */

  if ((dest_object = ddsi_find_object(ddsi_master_get_first_object(), dest)) 
      == NULL)
  {                               /* destination does not exist, create it */
    dest_object = ddsi_duplicate(src_object, dest);
  }
  else
  {                 /* destination exists, check if compatible with source */
    int compatible;
    
    compatible = (dest_object->type == src_object->type) &&
      (dest_object->distribution == src_object->distribution) &&
      (dest_object->nb_dim == src_object->nb_dim);
    for (c = 0; c < src_object->nb_dim; c++)
      compatible = compatible && (dest_object->dim[c] == src_object->dim[c]);

    if (!compatible)
    {    /* objects are not compatible, destination must be re-initialized */
      dds_delete(dest);
      dest_object = ddsi_duplicate(src_object, dest);
    }
  }
  
  ddsi_master_send_req(DDSI_REQ_COPY);
  for (c = 0; c < nb_slaves; c++)
  {
    send_channel(to_slaves[c],
      STRING, DDS_MAX_NAME_LEN, src,
      STRING, DDS_MAX_NAME_LEN, dest,
      END);
  }
  
  DDSI_GET_ACK;

  return 0;
}

