/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* DDS slave routines.                                                     */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.03.1995  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: ddsslave.c                                                    */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define DDS_IMPL
#define DDS_SLAVE_IMPL

#include <stdlib.h>
#include <ppc/portable.h>
#include <ppc/parlib/dds.h>
#include <ppc/parlib/ddspriv.h>
#include <ppc/parlib/skernel.h>
#include <ppc/csp/chan.h>
#include "ddsutil.h"

static dds_object *dds_slave_objects_first;
static dds_object *dds_slave_objects_last;

/*-------------------------------------------------------------------------*/
/* ddsi_slave_init:                                                        */
/* Initialize the DDS. Called by the slave.                                */
/*-------------------------------------------------------------------------*/
void ddsi_slave_init (void)
{
  dds_slave_objects_first = NULL;
  dds_slave_objects_last = NULL;
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_receive_object_prop:                                         */
/* Receive an object property from the master.                             */
/* Allocates the object property fields if necessary.                      */
/*-------------------------------------------------------------------------*/
void ddsi_slave_receive_object_prop (
  dds_object *object
)
{
  receive_channel(from_master,
    INT, &(object->type),
    INT, &(object->distribution),
    INT, &(object->nb_dim),
    STRING, DDS_MAX_NAME_LEN, object->name,
    INT_ARRAY, DDS_MAX_DIM, object->dim,
    END);
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_create:                                                LOCAL */
/* Slave part of dds_create().                                             */
/* Create a new object.                                                    */
/*-------------------------------------------------------------------------*/
static void ddsi_slave_create (void)
{
  dds_object *new;
  
  new = (dds_object *)malloc(sizeof(dds_object));
  
  ddsi_slave_receive_object_prop(new);
  ddsi_add_object(
    &dds_slave_objects_first, 
    &dds_slave_objects_last, 
    new);

  new->value = ddsi_value_allocate(
    new->type,
    0,
    new->nb_dim,
    new->dim);
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_terminate:                                                   */
/* Terminate the DDS. Called by the slave.                                 */
/*-------------------------------------------------------------------------*/
void ddsi_slave_terminate (void)
{
  ddsi_clear_database(&dds_slave_objects_first, &dds_slave_objects_last);
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_delete:                                                LOCAL */
/* Slave part of dds_delete().                                             */
/* Delete an object.                                                       */
/*-------------------------------------------------------------------------*/
static void ddsi_slave_delete (void)
{
  char name[DDS_MAX_NAME_LEN];
  dds_object *object;
  
  receive_channel(from_master,
    STRING, DDS_MAX_NAME_LEN, name,
    END);
    
  if ((object = ddsi_find_object(dds_slave_objects_first, name)) == NULL)
    return;                                            /* object not found */

  ddsi_remove_object(
    &dds_slave_objects_first, 
    &dds_slave_objects_last, 
    object);
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_copy:                                                  LOCAL */
/* Slave part of dds_copy().                                               */
/* Duplicate two objects.                                                  */
/*-------------------------------------------------------------------------*/
static void ddsi_slave_copy (void)
{
  char src[DDS_MAX_NAME_LEN];
  char dest[DDS_MAX_NAME_LEN];
  dds_object *src_object;
  dds_object *dest_object;
  
  receive_channel(from_master,
    STRING, DDS_MAX_NAME_LEN, src,
    STRING, DDS_MAX_NAME_LEN, dest,
    END);

  if ((src_object = ddsi_find_object(dds_slave_objects_first,  src)) 
      == NULL)
    return;                                            /* object not found */
  if ((dest_object = ddsi_find_object(dds_slave_objects_first, dest)) 
      == NULL)
    return;                                            /* object not found */

  ddsi_value_copy(
    src_object->value, 
    dest_object->value,
    src_object->type,
    0,
    src_object->nb_dim,
    src_object->dim,
    0, NULL,
    DDSI_COPY_IDX_NONE);
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_receive_value:                                         LOCAL */
/* Receive object value from master.                                       */
/*-------------------------------------------------------------------------*/
static void ddsi_slave_receive_value (
  void *value,
  int type,
  int act_dim,
  int nb_dim,
  int *dim,
  int nb_idx,
  int *idx
)
{
  int c;

  if (act_dim == nb_dim-1)
  {
    receive_channel(from_master,
      type+INT_ARRAY, dim[act_dim], value,
      END);
    return;
  }
    
  {
    void **ptr = value;
    
    if (act_dim < nb_idx)
      ddsi_slave_receive_value(ptr[idx[act_dim]], type, act_dim+1, 
        nb_dim, dim, nb_idx, idx);
    else
      for (c = 0; c < dim[act_dim]; c++)
	ddsi_slave_receive_value(ptr[c], type, act_dim+1, 
	  nb_dim, dim, nb_idx, idx);
  }
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_send_value:                                            LOCAL */
/* Send object value to master.                                            */
/*-------------------------------------------------------------------------*/
static void ddsi_slave_send_value (
  void *value,
  int type,
  int act_dim,
  int nb_dim,
  int *dim
)
{
  int c;

  if (act_dim == nb_dim-1)
  {
    send_channel(to_master,
      type+INT_ARRAY, dim[act_dim], value,
      END);
    return;
  }
    
  {
    void **ptr = value;
    
    for (c = 0; c < dim[act_dim]; c++)
      ddsi_slave_send_value(ptr[c], type, act_dim+1, nb_dim, dim);
  }
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_object_init:                                           LOCAL */
/* Slave part of the objects initialization functions.                     */
/*-------------------------------------------------------------------------*/
static void ddsi_slave_object_init (void)
{
  char name[DDS_MAX_NAME_LEN];
  dds_object *object;
  int idx[DDS_MAX_DIM];
  int nb_idx;

  receive_channel(from_master,
    STRING, DDS_MAX_NAME_LEN, name,
    INT_ARRAY, DDS_MAX_DIM, idx,
    INT, &nb_idx,
    END);
    
  if ((object = ddsi_find_object(dds_slave_objects_first, name)) == NULL)
    return;                                            /* object not found */

  ddsi_slave_receive_value(object->value, object->type,
    0, object->nb_dim, object->dim, nb_idx, idx);
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_object_store:                                          LOCAL */
/* Slave part of ddsi_store                                                */
/* Return the value of an object to the master.                            */
/*-------------------------------------------------------------------------*/
static void ddsi_slave_object_store (void)
{
  char name[DDS_MAX_NAME_LEN];
  dds_object *object;

  receive_channel(from_master,
    STRING, DDS_MAX_NAME_LEN, name,
    END);

  if ((object = ddsi_find_object(dds_slave_objects_first, name)) == NULL)
    return;                                            /* object not found */

  ddsi_slave_send_value(object->value, object->type,
    0, object->nb_dim, object->dim);
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_get_first_object:                                            */
/* Return the first object in the slave's objects list.                    */
/* ddsslave.c                                                              */
/*-------------------------------------------------------------------------*/
dds_object *ddsi_slave_get_first_object (void)
{
  return dds_slave_objects_first;
}

/*-------------------------------------------------------------------------*/
/* dds_slave_object_ptr:                                                   */
/* Return a pointer to the specified object.                               */
/*-------------------------------------------------------------------------*/
void *dds_slave_object_ptr (
  char *name
)
{
  dds_object *object;
  
  object = ddsi_find_object(ddsi_slave_get_first_object(), name);
  if (object == NULL)
    return object;
  
  return object->value;
}

/*-------------------------------------------------------------------------*/
/* dds_receive_name:                                                       */
/* Get an object name from the master. Each slave must call this function  */
/* for each dds_send_name master call.                                     */
/* The returned pointer is valid until the next call to this function.     */
/* For slaves only.                                                        */
/*-------------------------------------------------------------------------*/
char *dds_receive_name (void)
{
  static char name[DDS_MAX_NAME_LEN];
  
  receive_channel(from_master,
    STRING, DDS_MAX_NAME_LEN, name,
    END);
  
  return name;
}

/*-------------------------------------------------------------------------*/
/* ddsi_slave_handler:                                                     */
/* The parallel library kernel received a DDS request.                     */
/*-------------------------------------------------------------------------*/
void ddsi_slave_handler (void)
{
  DDSI_REQ req;
  
  receive_channel(from_master,
    INT, &req,
    END);
    
  switch(req)
  {
    case DDSI_REQ_CLEAR:
      ddsi_clear_database(&dds_slave_objects_first, 
        &dds_slave_objects_last);
      break;
    case DDSI_REQ_CREATE:
      ddsi_slave_create();
      break;
    case DDSI_REQ_DELETE:
      ddsi_slave_delete();
      break;
    case DDSI_REQ_COPY:
      ddsi_slave_copy();
      break;
    case DDSI_REQ_INIT:
      ddsi_slave_object_init();
      break;
    case DDSI_REQ_STORE:
      ddsi_slave_object_store();
      break;
    default:
      break;
  }

  DDSI_SEND_ACK;
}

