/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* DDS objects output.                                                     */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.03.1995  Droux       Created                                         */
/* 07.05.1995  Droux       Added dds_store_to_channel_nv()                 */
/* 08.05.1995  Droux       dds_store replaced by dds_get                   */
/*-------------------------------------------------------------------------*/
/* filename: ddsstore.c                                                    */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define DDS_IMPL
#define DDS_MGMT_IMPL
#define MKERNEL_IMPL

#include <stdio.h>
#include <stdlib.h>
#include <ppc/portable.h>
#include <ppc/parlib/mkernel.h>
#include <ppc/parlib/dds.h>
#include <ppc/parlib/ddspriv.h>
#include <ppc/csp/chan.h>
#include "ddsutil.h"

/*-------------------------------------------------------------------------*/
/* ddsi_store_print_data:                                            LOCAL */
/* Print the given data to the specified stream.                           */
/*-------------------------------------------------------------------------*/
static void ddsi_store_print_data (
  void *value,
  int type,
  int act_dim,
  int nb_dim,
  int *dim,
  int nb_idx,
  int *idx,
  FILE *stream
)
{
  int c;

  if (act_dim == nb_dim-1)
  {
    ddsi_print_vector(stream, type, dim[act_dim], value);
    return;
  }
    
  {
    void **ptr = value;
    
    if (act_dim < nb_idx)
      ddsi_store_print_data(ptr[idx[act_dim]], type, act_dim+1, 
        nb_dim, dim, nb_idx, idx, stream);
    else
    for (c = 0; c < dim[act_dim]; c++)
      ddsi_store_print_data(ptr[c], type, act_dim+1, nb_dim, dim,
        nb_idx, idx, stream);
  }
}

/*-------------------------------------------------------------------------*/
/* dds_get_to_stream:                                                      */
/* Fed the value of a DDS object into a stream.                            */
/*-------------------------------------------------------------------------*/
#ifdef PACTC
int dds_get_to_stream (
  char *name,
  FILE *stream,
  int act_idx,
  ...
)
#else
int dds_get_to_stream (va_alist)
va_dcl
#endif
{
  #ifndef PACTC
  char *name;
  FILE *stream;
  int act_idx;
  #endif
  dds_object *object;
  int idx[DDS_MAX_DIM];
  int nb_idx;
  va_list pvar;
  
  #ifdef PACTC
  va_start(pvar, act_idx);
  #else
  va_start(pvar);
  name = va_arg(pvar, char *);
  stream = va_arg(pvar, FILE *);
  act_idx = va_arg(pvar, int);
  #endif

  if ((object = ddsi_find_object(ddsi_master_get_first_object(), name)) 
      == NULL)
    return -1;                                         /* object not found */

  for (nb_idx = 0; act_idx != END; nb_idx++)                /* get indices */
  {
    idx[nb_idx] = act_idx;
    act_idx = va_arg(pvar, int);
  }

  ddsi_store_print_data(object->value, object->type, 0, 
    object->nb_dim, object->dim, nb_idx, idx, stream);
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* ddsi_store_send_data:                                             LOCAL */
/* Send the given data on the specified channel.                           */
/*-------------------------------------------------------------------------*/
static void ddsi_store_send_data (
  void *value,
  int type,
  int act_dim,
  int nb_dim,
  int *dim,
  int nb_idx,
  int *idx,
  channel chan,
  void (* before_func) (types_names, int, void *),       /* to call before */
  void *before_data,                       /* value to pass to before_func */
  void (* after_func) (types_names, int, void *),         /* to call after */
  void *after_data                          /* value to pass to after_func */
)
{
  int c;

  if (act_dim == nb_dim-1)
  {
    int the_type = type+INT_ARRAY;
    int size;
    int offset;
    
    if (nb_dim == 1 && nb_idx)
    {
      size = 1;
      offset = idx[0] * type_size[type];
    }
    else
    {
      size = dim[act_dim];
      offset = 0;
    }
    
    if (before_func != NULL)
      before_func(the_type, size, before_data);
      
    send_channel(chan,
      the_type, size, ((char *)value)+offset,
      END);

    if (after_func != NULL)
      after_func(the_type, size, after_data);

    return;
  }
    
  {
    void **ptr = value;
    
    if (act_dim < nb_idx)
      ddsi_store_send_data(ptr[idx[act_dim]], type, act_dim+1, 
        nb_dim, dim, nb_idx, idx, chan,
        before_func, before_data, after_func, after_data);
    else
      for (c = 0; c < dim[act_dim]; c++)
	ddsi_store_send_data(ptr[c], type, act_dim+1, nb_dim, dim,
	  nb_idx, idx, chan, 
	  before_func, before_data, after_func, after_data);
  }
}

/*-------------------------------------------------------------------------*/
/* dds_get_to_channel_nv:                                                  */
/* Same as dds_get_to_channel() but without variable parameters. In        */
/* addition, this version of the function allows to define functions which */
/* have to be called before and after a data transmission.                 */
/* Such functions must be defined as follows:                              */
/*   void before_or_after (                                                */
/*     types_names data_type, int size, void *custom_data);                */
/*-------------------------------------------------------------------------*/
int dds_get_to_channel_nv (
  char *name,                                               /* object name */
  channel chan,                                  /* channel to destination */
  int nb_idx,                                 /* number of elements in idx */
  int *idx,                                 /* indices to access sub-array */
  void (* before_func) (types_names, int, void *),       /* to call before */
  void *before_data,                       /* value to pass to before_func */
  void (* after_func) (types_names, int, void *),         /* to call after */
  void *after_data                          /* value to pass to after_func */
)
{
  dds_object *object;

  if ((object = ddsi_find_object(ddsi_master_get_first_object(), name)) 
      == NULL)
    return -1;                                         /* object not found */

  ddsi_store_send_data(object->value, object->type, 0, 
    object->nb_dim, object->dim, nb_idx, idx, chan,
    before_func, before_data, after_func, after_data);
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* dds_get_to_channel:                                                     */
/* Fed the value of a DDS object into a channel.                           */
/*-------------------------------------------------------------------------*/
#ifdef PACTC
int dds_get_to_channel (
  char *name,
  channel chan,
  int act_idx,
  ...
)
#else
int dds_get_to_channel (va_alist)
va_dcl
#endif
{
  #ifndef PACTC
  char *name;
  channel chan;
  int act_idx;
  #endif
  
  int idx[DDS_MAX_DIM];
  int nb_idx;
  va_list pvar;
  
  #ifdef PACTC
  va_start(pvar, act_idx);
  #else
  va_start(pvar);
  name = va_arg(pvar, char *);
  chan = va_arg(pvar, channel);
  act_idx = va_arg(pvar, int);
  #endif

  for (nb_idx = 0; act_idx != END; nb_idx++)                /* get indices */
  {
    idx[nb_idx] = act_idx;
    act_idx = va_arg(pvar, int);
  }
  
  return dds_get_to_channel_nv(name, chan, nb_idx, idx,
    NULL, NULL, NULL, NULL);
}

/*-------------------------------------------------------------------------*/
/* dds_get:                                                                */
/* Store the value of a DDS object in the given data location.             */
/*-------------------------------------------------------------------------*/
#ifdef PACTC
int dds_get (
  char *name,
  void *data,
  int act_idx,
  ...
)
#else
int dds_get (va_alist)
va_dcl
#endif
{
  #ifndef PACTC
  char *name;
  void *data;
  int act_idx;
  #endif
  dds_object *object;
  int idx[DDS_MAX_DIM];
  int nb_idx;
  va_list pvar;
  
  #ifdef PACTC
  va_start(pvar, act_idx);
  #else
  va_start(pvar);
  name = va_arg(pvar, char *);
  data = va_arg(pvar, void *);
  act_idx = va_arg(pvar, int);
  #endif

  if ((object = ddsi_find_object(ddsi_master_get_first_object(), name)) 
      == NULL)
    return -1;                                         /* object not found */

  for (nb_idx = 0; act_idx != END; nb_idx++)                /* get indices */
  {
    idx[nb_idx] = act_idx;
    act_idx = va_arg(pvar, int);
  }

  ddsi_value_copy(      /* get value from local copy (DDS_REPLICATED only) */
    object->value,
    data,
    object->type,
    0,
    object->nb_dim,
    object->dim,
    nb_idx, idx,
    DDSI_COPY_IDX_SRC);

  return 0;
}
