/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* DDS utilities used by both the master and the slaves.                   */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 05.04.1995  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: ddsutil.c                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define DDS_IMPL

#include <stdlib.h>
#include <ppc/portable.h>
#include <ppc/parlib/dds.h>
#include <ppc/parlib/ddspriv.h>
#include "ddsutil.h"

/*-------------------------------------------------------------------------*/
/* ddsi_add_object:                                                        */
/* Add given object to objects list.                                       */
/*-------------------------------------------------------------------------*/
void ddsi_add_object (
  dds_object **first,
  dds_object **last,
  dds_object *object
)
{
  object->prev = *last;
  object->next = NULL;
  if (*last == NULL)
    *first = object;
  else
    (*last)->next = object;
  *last = object;
}

/*-------------------------------------------------------------------------*/
/* ddsi_remove_object:                                                     */
/* Remove object from object list.                                         */
/*-------------------------------------------------------------------------*/
void ddsi_remove_object (
  dds_object **first,
  dds_object **last,
  dds_object *object
)
{
  if (*last == object)
    *last = object->prev;
  else
    object->next->prev = object->prev;
    
  if (*first == object)
    *first = object->next;
  else
    object->prev->next = object->next;
    
  ddsi_free_value(object->value, 0, object->nb_dim, object->dim);
  free(object);
}

/*-------------------------------------------------------------------------*/
/* ddsi_find_object:                                                       */
/* Return a pointer to the given object, or NULL if object not found.      */
/*-------------------------------------------------------------------------*/
dds_object *ddsi_find_object (
  dds_object *first,
  char *name
)
{
  dds_object *object;
  
  object = first;
  while ((object != NULL) && strcmp(object->name, name))
    object = object->next;
    
  return object;
}

/*-------------------------------------------------------------------------*/
/* ddsi_free_value:                                                        */
/* Release the memory used by an object value.                             */
/*-------------------------------------------------------------------------*/
void ddsi_free_value (
  void *value,
  int act_dim,
  int nb_dim,
  int *dim
)
{
  if (act_dim == nb_dim-1)
  {
    free(value);
    return;
  }
    
  {
    int c;
    void **ptr = value;
    
    for (c = 0; c < dim[act_dim]; c++)
      ddsi_free_value(ptr[c], act_dim+1, nb_dim, dim);
    free(value);
  }
}

/*-------------------------------------------------------------------------*/
/* ddsi_clear_database:                                                    */
/* Clear objects.                                                          */
/*-------------------------------------------------------------------------*/
void ddsi_clear_database (
  dds_object **first,
  dds_object **last
)
{
  dds_object *object;
  dds_object *next_object;
  
  object = *first;
  while (object != NULL)
  {
    next_object = object->next;
    ddsi_remove_object(first, last, object);
    object = next_object;
  }
}

/*-------------------------------------------------------------------------*/
/* ddsi_value_allocate:                                                    */
/* Allocate memory for an object value.                                    */
/*-------------------------------------------------------------------------*/
void *ddsi_value_allocate (
  types_names type,
  int act_dim,
  int nb_dim,
  int *dim
)
{
  if (act_dim == nb_dim-1)
    return malloc(dim[act_dim]*type_size[type]);
    
  {
    int c;
    void **ptr;
  
    ptr = malloc(dim[act_dim]*sizeof(void *));
    for (c = 0; c < dim[act_dim]; c++)
      ptr[c] = ddsi_value_allocate(type, act_dim+1, nb_dim, dim);
    
    return ptr;
  }
}

/*-------------------------------------------------------------------------*/
/* ddsi_value_copy:                                                        */
/* Copy the value of an object into another one                            */
/*-------------------------------------------------------------------------*/
void ddsi_value_copy (
  void *src_value,
  void *dest_value,
  int type,
  int act_dim,
  int nb_dim,
  int *dim,
  int nb_idx,
  int *idx,
  DDSI_COPY_IDX to_apply
)
{
  if (act_dim == nb_dim-1)
  {
    char *src_char = src_value;
    char *dest_char = dest_value;
    int c;
    
    for (c = 0; c < dim[act_dim]*type_size[type]; c++)
      dest_char[c] = src_char[c];
    return;
  }
    
  {
    int c;
    void **src_ptr = src_value;
    void **dest_ptr = dest_value;
    
    if (act_dim < nb_idx)
    {
      switch(to_apply)
      {
        case DDSI_COPY_IDX_DEST:
          dest_ptr = dest_ptr[idx[act_dim]];
	  break;
	case DDSI_COPY_IDX_SRC:
	  src_ptr = src_ptr[idx[act_dim]];
	  break;
	default:
	  break;
      }

      ddsi_value_copy(src_ptr, dest_ptr, type,
	act_dim+1, nb_dim, dim, nb_idx, idx, to_apply);
    }
    else
      for (c = 0; c < dim[act_dim]; c++)
	ddsi_value_copy(src_ptr[c], dest_ptr[c], type,
	  act_dim+1, nb_dim, dim, nb_idx, idx, to_apply);
  }
}

/*-------------------------------------------------------------------------*/
/* ddsi_print_vector:                                                      */
/* Print the given vector on the specified stream.                         */
/*-------------------------------------------------------------------------*/
void ddsi_print_vector (
  FILE *stream,
  int type,
  int dim,
  void *value
)
{
  int c;
  
  for (c = 0; c < dim; c++)
  {
    switch(type)
    {
      case INT:
        fprintf(stream, "%d ", ((int *)value)[c]);
        break;
      case SHORT:
        fprintf(stream, "%hd ", ((short *)value)[c]);
        break;
      case UNSIGNED_SHORT:
        fprintf(stream, "%hu ", ((unsigned short *)value)[c]);
        break;
      case LONG:
        fprintf(stream, "%ld ", ((long *)value)[c]);
        break;
      case UNSIGNED_LONG:
        fprintf(stream, "%lu ", ((unsigned long *)value)[c]);
        break;
      case FLOAT:
        fprintf(stream, "%f ", ((float *)value)[c]);
        break;
      case DOUBLE:
        fprintf(stream, "%f ", ((double *)value)[c]);
        break;
      case CHAR:
        fprintf(stream, "%c ", ((char *)value)[c]);
        break;
      default:
        break;
    }
  }
  
  fprintf(stream, "\n");
}

/*-------------------------------------------------------------------------*/
/* ddsi_scan_vector:                                                       */
/* Scan the given vector from the specified stream.                        */
/*-------------------------------------------------------------------------*/
void ddsi_scan_vector (
  FILE *stream,
  int type,
  int dim,
  void *value
)
{
  int c;
  float float_tmp;
  
  for (c = 0; c < dim; c++)
  {
    switch(type)
    {
      case INT:
        fscanf(stream, "%d ", &(((int *)value)[c]));
        break;
      case SHORT:
        fscanf(stream, "%hd ", &(((short *)value)[c]));
        break;
      case UNSIGNED_SHORT:
        fscanf(stream, "%hu ", &(((unsigned short *)value)[c]));
        break;
      case LONG:
        fscanf(stream, "%ld ", &(((long *)value)[c]));
        break;
      case UNSIGNED_LONG:
        fscanf(stream, "%lu ", &(((unsigned long *)value)[c]));
        break;
      case FLOAT:
        fscanf(stream, "%f ", &(((float *)value)[c]));
        break;
      case DOUBLE:
        fscanf(stream, "%f ", &float_tmp);
	((double *)value)[c] = float_tmp;
        break;
      case CHAR:
        fscanf(stream, "%c ", &(((char *)value)[c]));
        break;
      default:
        break;
    }
  }
}

