/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Portable Parallel Library command line utility.                         */
/* PPL configuration file parser. The function ppl_parse() parse the       */
/* .ppl file of the routine and update the properties database             */
/* accordingly.                                                            */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 13.01.1995  Droux       Created                                         */
/*-------------------------------------------------------------------------*/
/* filename: lparser.h                                                     */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define LPARSER_IMPL

#include <stdio.h>

#include <ppc/portable.h>
#include "lparser.h"
#include "lprop.h"

#define RETURN_ERROR(code) {lparser_errno = code; return -1;}

#define EO_ITEMS -1
#define EO_WORDS -1
#define EO_LINE -1
#define MAX_ITEM_LEN 100
#define MAX_LINE_LEN 1000

static FILE *ppl_file;

/*-------------------------------------------------------------------------*/
/* get_next_line:                                                          */
/* Get the next line from the configuration file.                          */
/* Filters empty lines and line beginning with a comment character.        */
/* Return EO_LINE if no more lines are available, zero otherwise.          */
/*-------------------------------------------------------------------------*/
int get_next_line (
  char *line
)
{
  static int eo_line = 0;
  static char act_line [MAX_LINE_LEN];
  static char *fgets_stat;
  static int valid;

  if (eo_line == EO_LINE)
    return EO_LINE;

  do {
    fgets_stat = fgets(act_line, MAX_LINE_LEN, ppl_file);
    lparser_act_line++;
    valid = strlen(act_line) && act_line[0] != '#';
  } while(fgets_stat != NULL && !valid);

  if (fgets_stat == NULL)
    eo_line = EO_LINE;

  strcpy(line, act_line);

  return 0;
}

/*-------------------------------------------------------------------------*/
/* get_next_word:                                                          */
/* Get the next word from the configuration file. Filters comments.        */
/* Return EO_WORDS if no more words are available, zero otherwise.         */
/*-------------------------------------------------------------------------*/
int get_next_word (
  char *word
)
{
  static int eo_words = 0;
  static int eo_line = EO_LINE;
  static char act_line[MAX_LINE_LEN];
  static char *cptr;
  int idx;

  if (eo_words == EO_WORDS)
    return EO_WORDS;
    
  if (eo_line)
  {
    eo_line = get_next_line (act_line);
    if (eo_line == EO_LINE)
    {
      eo_words = 1;
      return EO_WORDS;
    }
    cptr = act_line;
  }
  
  for (;;)
  {
    for (; *cptr == ' ' || *cptr == '\t'; cptr++);          /* skip blanks */
    for (idx = 0; 
         *cptr != ' ' && *cptr != '\t' && *cptr != '\0' && 
	 *cptr != '\n' && *cptr != '#' && idx < MAX_ITEM_LEN; 
         idx++)
    {
      word[idx] = *cptr++;
    }
    
    if (idx == 0)
    {
      switch(*cptr)
      {
        case ' ':
	  continue;
	  break;
	default:
	  eo_line = get_next_line (act_line);
	  if (eo_line == EO_LINE)
	  {
	    eo_words = 1;
	    return EO_WORDS;
	  }
	  cptr = act_line;
	  continue;
	  break;
      }
    }
    word[idx] = '\0';
    return 0;
  }
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* get_next_item:                                                          */
/* Get the next item from the configuration file and put it in the         */
/* string 'item'.                                                          */
/* Returns EO_ITEMS if no more items are available from the file,          */
/* or zero otherwise.                                                      */
/*-------------------------------------------------------------------------*/
int get_next_item (
  char *item
)
{
  static int eo_items = 0;
  static char act_item [MAX_ITEM_LEN];

  if (eo_items == EO_ITEMS)
    return EO_ITEMS;

  if (get_next_word(act_item) == EO_WORDS)
    return EO_ITEMS;
    
  strcpy(item, act_item);

  return 0;
}

/*-------------------------------------------------------------------------*/
/* ppl_parse:                                                              */
/* Parse the PPL configuration file (with extension .ppl) and update       */
/* the properties database accordingly.                                    */
/* If an error occurs then returns -1 and set the lparser_errno global     */
/* variable, returns 0 otherwise.                                          */
/*-------------------------------------------------------------------------*/
int ppl_parse (
  char *library_name,
  int verbose                            /* 1 if verbose mode, 0 otherwise */
)
{
  char full_name [100];
  char item [MAX_ITEM_LEN];
  int item_stream;
  
  sprintf(full_name, "%s.ppl", library_name);
  
  if (verbose)
    printf("starting parsing of %s\n", full_name);
  lprop_init(library_name);
  
  lparser_act_line = 0;
  
  ppl_file = fopen(full_name, "r");
  if (ppl_file == NULL)
    RETURN_ERROR(LERR_COULD_NOT_OPEN);

  item_stream = get_next_item(item);                    /* look ahead read */
  while(item_stream != EO_ITEMS)
  {                                                /* process current item */
    lprop_new_routine(item);
    item_stream = get_next_item(item);
  }
  
  if (verbose)
    printf("%s parsed successfully\n", full_name);
  
  return 0;
}
