/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Portable Parallel Library command line utility.                         */
/* Main.                                                                   */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 13.12.1994  Droux       Created                                         */
/* 20.03.1995  Droux       Ported to the Paragon                           */
/* 16.05.1995  Droux       Ported to the SP-2                              */
/*-------------------------------------------------------------------------*/
/* filename: pplcmdmain.c                                                  */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, Intel                                                   */
/*   NEXTSTEP 3.3, Intel                                                   */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <ppc/portable.h>
#include <ppc/csp/version.h>
#include "toolsutil.h"
#include "lparser.h"
#include "lprop.h"
#include "lgenerator.h"
#include "lgenunix.h"
#include "lgenpvm.h"
#include "lgenpactc.h"
#include "lgenpgon.h"
#include "lgendefs.h"
#include "genutil.h"

#define RM_NAME "/bin/rm"

#if defined(SUN4)
  extern char *sys_errlist[];
  #define strerror(x) sys_errlist[x]
#endif

#define host_arch_human(x) x              /* to be replaced by function @@ */

#define PROG_NAME "ppl"
#define ALLOWED_ACTIONS "cdhimv"

struct dir_entry
{
  char name [MAX_FILE_NAME_LEN];
  struct dir_entry *next;
};

extern int optind;                                     /* used by getopt() */
extern char *optarg;

static char ppc_home [MAX_FILE_NAME_LEN];            /* PPC home directory */
static char ppl_file [MAX_FILE_NAME_LEN];               /* PPL config file */
static int verbose;                                   /* 1 if verbose mode */
static int warnings;                      /* 1 if warnings must be printed */
static int recursive;                               /* 1 if recursive mode */
static int erase;                                       /* 1 if erase mode */
static arch_enum arch;                              /* target architecture */
static struct dir_entry *files_list = NULL;   /* list of library dir files */
static char pvm_home [MAX_FILE_NAME_LEN];            /* PVM home directory */
static char pactc_home [MAX_FILE_NAME_LEN];       /* PACT-C home directory */
static char install_directory [MAX_FILE_NAME_LEN];    /* installation path */

/*-------------------------------------------------------------------------*/
/* print_usage:                                                            */
/* Print usage and exit.                                                   */
/*-------------------------------------------------------------------------*/
void print_usage (char *msg)
{
  if (msg != NULL)
    printf("%s: %s\n", PROG_NAME, msg);
  printf("%s: usage: {%s} -[aefirvw]\n", 
    PROG_NAME, ALLOWED_ACTIONS);
  printf("%s: type '%s h' for help\n", PROG_NAME, PROG_NAME);
  exit(1);
}

/*-------------------------------------------------------------------------*/
/* print_title:                                                            */
/* Print program title.                                                    */
/*-------------------------------------------------------------------------*/
void print_title (void)
{
  printf("PPC library manager. " PPC_VERSION "\n");
  printf("Copyright (c) 1993, 1994, 1995 Engineering School of Biel-Bienne.\n");
}

/*-------------------------------------------------------------------------*/
/* print_help:                                                             */
/* Print help and exit.                                                    */
/*-------------------------------------------------------------------------*/
void print_help (void)
{
  print_title();
  printf("Usage: ppr key [options]\n");
  printf("key must be a combination of:\n");
  printf("  c         Configure library.\n");
  printf("  m         Make library.\n");
  printf("  d         Delete hardware dependent files.\n");
  printf("  i         Install library.\n");
  printf("  v         Print library configuration report.\n");
  printf("  h         Print this message.\n");
  printf("options:\n");
  printf("  -a arch   Specify target architecture. Default is unix. Allowed\n");
  printf("            architectures are unix, pvm, pactc, paragon, and sp2.\n");
  printf("  -f file   Specify library configuration file. Default is the\n");
  printf("            first .ppl file found in current directory.\n");
  printf("  -i path   Specify parallel libraries directory. Default is\n");
  printf("            $PPC_HOME/parlib.\n");
  printf("  -e        Used with 'd' key, delete all files generated by utility\n");
  printf("  -r        Perform the operation on all routines\n");
  printf("  -v        Run the utility in verbose mode.\n");
  printf("  -w        Shows all warnings during compilation.\n");
  exit(2);
}

/*-------------------------------------------------------------------------*/
/* check_key:                                                              */
/* Check if the given key is valid. If not, prints an error message,       */
/* the program usage and exit.                                             */
/*-------------------------------------------------------------------------*/
void check_key (char *key)
{
  int c;
  char err_msg [100];
  
  for (c = 0; c < strlen(key); c++)
    if (strchr(ALLOWED_ACTIONS, key[c]) == NULL)
    {
      sprintf(err_msg, "%c: unknown action", key[c]);
      print_usage(err_msg);
    }
}

/*-------------------------------------------------------------------------*/
/* process_options:                                                        */
/* Parse the specified options and update the corresponding global         */
/* variables.                                                              */
/* If an error occures, then print an error message, the program usage,    */
/* and exit.                                                               */
/*-------------------------------------------------------------------------*/
void process_options (int argc, char **argv)
{
  int option;                            /* current option being processed */

  while((option = getopt(argc-1, &argv[1], "a:f:i:ervw")) != EOF)
  {
    switch(option)
    {
      case 'a':                            /* optional target architecture */
      {                                    
        int correct = 0;
        int act_arch;
	
	if (optarg == NULL)
          print_usage("option -a must be followed by unix, pvm, pactc, paragon, or sp2");
        for (act_arch = 0; act_arch < NB_ARCH && !correct; act_arch++)
	  correct = !strcmp(optarg, arch_names[act_arch]);
	if (!correct)
	  print_usage("specified architecture incorrect (must be unix, pvm, pactc, paragon, or sp2)");
	arch = act_arch-1;
        break;
      }
      case 'f':                        /* optional configuration file name */
      {
        if (optarg == NULL)
          print_usage("option -f must be followed by a file name");
        strcpy(ppl_file, optarg);             /* validity is checked later */
	break;
      }
      case 'i':                         /* optional installation directory */
      {                                
        struct stat path_stat;

        if (optarg == NULL)
          print_usage("option -i must be followed by a path");
        if(stat(optarg, &path_stat) == -1)
	  print_usage("specified installation path does not exist or is not accessible");
        if (!(path_stat.st_mode & S_IFDIR))
	  print_usage("specified installation path is not a directory");
        strcpy(install_directory, optarg);
        break;
      }
      case 'v':                                  /* toggle to verbose mode */
        verbose = 1;
	break;
      case 'w':                   /* print all warnings during compilation */
        warnings = 1;
	break;
      case 'e':
        erase = 1;
	break;
      case 'r':
        recursive = 1;                              /* recursive operation */
	break;
      default:
        print_usage(NULL);
	break;
    }
  }
}

/*-------------------------------------------------------------------------*/
/* get_ext:                                                                */
/* Return in ext the extension of the given file name. ext is set to an    */
/* empty string if the file has no extension.                              */
/*-------------------------------------------------------------------------*/
void get_ext (
  char *name,
  char *ext
)
{
  int c;
  int i;
  int name_len = strlen(name);
  
  for (c = name_len-1, i = 0; c && (name[c] != '.'); c--, i++);
  if (c)
  {
    ext[i] = '\0';
    for (c = 0, i--; i >= 0; i--, c++)
      ext[c] = name[name_len-1-i];
  }
  else
    ext[0] = '\0';
}

/*-------------------------------------------------------------------------*/
/* check_ppl_file:                                                         */
/* Checks if the currently defined PPL file exists.                        */
/*-------------------------------------------------------------------------*/
void check_ppl_file (void)
{
  struct stat file_stat;
  char ext [100];

  if(stat(ppl_file, &file_stat) == -1)
  {                               /* <file> does not exist, try <file>.ppl */
    char file_plus_ext [MAX_FILE_NAME_LEN];
    sprintf(file_plus_ext, "%s.ppl", ppl_file);
    if(stat(file_plus_ext, &file_stat) == -1)
      print_usage("specified library configuration file does not exist");
  }

  get_ext(ppl_file, ext);             /* suppress .ppl extension if exists */
  if (!strcmp(ext, "ppl"))
    ppl_file[strlen(ppl_file)-strlen(ext)-1] = '\0';
}

/*-------------------------------------------------------------------------*/
/* get_dir_files:                                                          */
/* Initialize the list files_list with the files found in the library      */
/* directory.                                                              */
/*-------------------------------------------------------------------------*/
void get_dir_files (void)
{
  DIR *dir;
  struct direct *dirEnt;
  struct dir_entry *act_entry = files_list;
  struct dir_entry *new_entry;
  
  dir = opendir(".");
  for (dirEnt = readdir(dir); dirEnt != NULL; dirEnt = readdir(dir))
  {
    if (!strcmp(dirEnt->d_name, ".") || !strcmp(dirEnt->d_name, ".."))
      continue;
    new_entry = malloc(sizeof(struct dir_entry));
    if (files_list == NULL)
      files_list = new_entry;
    else
      act_entry->next = new_entry;
    act_entry = new_entry;
    strcpy(act_entry->name, dirEnt->d_name);
  }
}

/*-------------------------------------------------------------------------*/
/* found_ppl_file:                                                         */
/* Found the first .ppl file within the library directory and update the   */
/* ppl_file global variable accordingly.                                   */
/*-------------------------------------------------------------------------*/
void found_ppl_file (void)
{
  struct dir_entry *act_entry = files_list;
  char ext [100];
  int found = 0;
  
  while (act_entry != NULL && !found)
  {
    get_ext(act_entry->name, ext);
    if (!(found = !strcmp(ext, "ppl")))
      act_entry = act_entry->next;
  }

  if (!found)
    print_usage("no library configuration file found in current directory");
  
  strcpy(ppl_file, act_entry->name);
  ppl_file[strlen(ppl_file)-strlen(ext)-1] = '\0';
}

/*-------------------------------------------------------------------------*/
/* check_arch:                                                             */
/* Some architecture may require special environment definitions.          */
/*-------------------------------------------------------------------------*/
void check_arch (void)
{
  switch(arch)  /* special operations depending on the target architecture */
  {   
    case PVM_ARCH:
    {                                            /* get PVM home directory */
      char *pvm_home_ptr;
      
      if ((pvm_home_ptr = getenv(PVM_HOME_NAME)) == NULL)
      {
	printf("%s: %s environment variable not defined\n",
	  PROG_NAME, PVM_HOME_NAME);
	exit(4);
      }
      strcpy(pvm_home, pvm_home_ptr);               /* initialize PVM home */
      break;
    }
    case PACTC_ARCH:
    {                                         /* get PACT-C home directory */
      char *pactc_home_ptr;
      
      if ((pactc_home_ptr = getenv(PACTC_HOME_NAME)) == NULL)
      {
	printf("%s: %s environment variable not defined\n",
	  PROG_NAME, PACTC_HOME_NAME);
	exit(4);
      }
      strcpy(pactc_home, pactc_home_ptr);        /* initialize PACT-C home */
      break;
    }
    default:
      break;
  }
}

/*-------------------------------------------------------------------------*/
/* print_summary:                                                          */
/* Summarize the actual tool configuration state.                          */
/*-------------------------------------------------------------------------*/
void print_summary (void)
{
  print_title();
  printf("\n");
  printf("target architecture        : %s\n", output_arch_names[arch]);
  if (strlen(ppl_file))
    printf("library configuration file : %s\n", ppl_file);
  printf("host computer architecture : %s\n", 
   host_arch_human(HOST_ARCH));
  printf("compilation warnings       : ");
  if (warnings)
    printf("Enabled\n");
  else
    printf("Disabled\n");
  printf("recurse on routines        : ");
  if (recursive)
    printf("Yes\n");
  else
    printf("No\n");
  printf("\n");
}

/*-------------------------------------------------------------------------*/
/* parse:                                                                  */
/* Parse the configuration file.                                           */
/*-------------------------------------------------------------------------*/
void parse (void)
{
  static int parsed = 0;

  if (!parsed)
  {
    if (ppl_parse(ppl_file, verbose))
    {
      printf("%s.ppl:", ppl_file);
      if (lparser_act_line)
        printf("%d: ", lparser_act_line);
      else
        printf(" ");
      printf("%s\n", lparser_errors[lparser_errno]);
      exit(1);
    }
    
    if (!lprop_nb_routines())
    {
      printf("%s.ppl: no routine declared in library\n", ppl_file);
      exit(1);
    }
    
    parsed = 1;
  }
}

/*-------------------------------------------------------------------------*/
/* configure:                                                              */
/* Action 'c'.                                                             */
/*-------------------------------------------------------------------------*/
void configure (void)
{
  if (verbose)
    printf("Configuring %s library for %s\n", ppl_file, 
      output_arch_names[arch]);
      
  parse();
  
  if (recursive)
  {                               /* configure all routines of the library */
    char *act_routine;
    char *argv[] = 
      {NULL,                                               /* program name */
      "c",                                                       /* action */
      "-a", NULL,                                          /* architecture */
      "-d", NULL,                                     /* routine directory */
      "-v", "-V",                                        /* verbose option */
      NULL};
    char ppr_call [256];
    
    sprintf(ppr_call, "%s/lib/%s/pprcmd", ppc_home, HOST_ARCH);
    
    argv[0] = ppr_call;
    argv[3] = arch_names[arch];
    if (!verbose)
      argv[6] = argv[7] = NULL;
    
    lprop_get_routine_init();
    act_routine = lprop_get_routine_next();
    while(act_routine != NULL)
    {
      argv[5] = act_routine;                          /* routine directory */
      run_cmd(ppr_call, argv, 0);
      act_routine = lprop_get_routine_next();
    }
  }
  
  if (lgen_makefile(verbose) < 0)
    exit(1);
  
  if (lgen_lib_starter(verbose) < 0)
    exit(1);
  
  switch(arch)
  {
    case UNIX_ARCH:
      lgen_unix_files(verbose);
      break;
    case SP2_ARCH:
    case PVM_ARCH:
      lgen_pvm_files(verbose);
      break;
    case PACTC_ARCH:
      lgen_pactc_files(verbose);
      break;
    case PARAGON_ARCH:
      lgen_pgon_files(verbose);
      break;
    default:
      exit(1);
  }
  
  {
    char *argv[] = 
      {MAKE_NAME, "-f", PPL_DIR "/makefile", NULL};

    run_cmd(MAKE_NAME, argv, verbose);
  }

  if (verbose)
    printf("%s library configured for %s\n\n", ppl_file, 
      output_arch_names[arch]);
}

/*-------------------------------------------------------------------------*/
/* make:                                                                   */
/* Action 'm'.                                                             */
/*-------------------------------------------------------------------------*/
void make (void)
{
  char other_libs [100];
  char makefile_name [100];
  char warnings_opt [10];
  char bin_dir [256];

  if (verbose)
    printf("Making %s library for %s\n", ppl_file, output_arch_names[arch]);
  
  parse();
  
  if (recursive)
  {                                    /* make all routines of the library */
    char *act_routine;
    char *argv[] = 
      {NULL,                                               /* program name */
      "m",                                                       /* action */
      "-a", NULL,                                          /* architecture */
      "-d", NULL,                                     /* routine directory */
      "-v", "-V",                                        /* verbose option */
      NULL};
    char ppr_call [256];
    
    sprintf(ppr_call, "%s/lib/%s/pprcmd", ppc_home, HOST_ARCH);
    
    argv[0] = ppr_call;
    argv[3] = arch_names[arch];
    if (!verbose)
      argv[6] = argv[7] = NULL;
    
    lprop_get_routine_init();
    act_routine = lprop_get_routine_next();
    while(act_routine != NULL)
    {
      argv[5] = act_routine;                          /* routine directory */
      run_cmd(ppr_call, argv, 0);
      act_routine = lprop_get_routine_next();
    }
  }

  if (arch == PACTC_ARCH)
  {
    sprintf(bin_dir, "%s/.%s/PACTC",
      PPL_DIR, arch_names[arch]);
  }
  else
  {
    sprintf(bin_dir, "%s/.%s/%s",
      PPL_DIR, arch_names[arch], HOST_ARCH);
  }
  gen_create_dir(bin_dir, verbose);
    
  sprintf(makefile_name, "%s/.%s/makefile", PPL_DIR, arch_names[arch]);
  
  #ifdef SUN4SOL2
    if (arch != SP2_ARCH)
      strcpy(other_libs, "-lnsl -lsocket");
    else
      other_libs[0] = '\0';
  #else
    other_libs[0] = '\0';
  #endif
  
  if (warnings)
    strcpy(warnings_opt, "-Wall");
  else
    warnings_opt[0] = '\0';

  {
    char *argv[] = 
      {NULL, "-f", NULL, NULL, NULL, NULL, NULL, NULL};
    char arch_option [100];
    char host_arch [100];
    char other_libs_opt [100];
    char warning_flag [100];

    argv[0] = MAKE_NAME;
    argv[2] = makefile_name;
    sprintf(arch_option, "ARCH=%s", arch_names[arch]);
    argv[3] = arch_option;
    sprintf(host_arch, "HOST_ARCH=%s", HOST_ARCH);
    argv[4] = host_arch;
    sprintf(other_libs_opt, "OTHERLIBS=%s", other_libs);
    argv[5] = other_libs_opt;
    sprintf(warning_flag, "WARNING_FLAG=%s", warnings_opt);
    argv[6] = warning_flag;
    run_cmd(MAKE_NAME, argv, verbose);
  }

  if (verbose)
    printf("%s library made for %s\n\n", ppl_file, output_arch_names[arch]);
}

/*-------------------------------------------------------------------------*/
/* delete:                                                                 */
/* Action 'd'.                                                             */
/*-------------------------------------------------------------------------*/
void delete (void)
{
  parse();
  
  if (recursive)
  {                                    /* make all routines of the library */
    char *act_routine;
    char *argv[] = 
      {NULL,                                               /* program name */
      "d",                                                       /* action */
      "-a", NULL,                                          /* architecture */
      "-d", NULL,                                     /* routine directory */
      NULL, NULL,                                        /* verbose option */
      NULL,                                                /* erase option */
      NULL};
    char ppr_call [256];
    int act_arg;
    char verbose_option[] = "-v";
    char limited_verbose_option[] = "-V";
    char erase_option[] = "-e";
    
    sprintf(ppr_call, "%s/lib/%s/pprcmd", ppc_home, HOST_ARCH);
    
    argv[0] = ppr_call;
    argv[3] = arch_names[arch];
    act_arg = 6;
    
    if (verbose)
    {
      argv[act_arg++] = verbose_option;
      argv[act_arg++] = limited_verbose_option;
    }
    
    if (erase)
      argv[act_arg++] = erase_option;
      
    argv[act_arg] = NULL;
    
    lprop_get_routine_init();
    act_routine = lprop_get_routine_next();
    while(act_routine != NULL)
    {
      argv[5] = act_routine;                          /* routine directory */
      run_cmd(ppr_call, argv, 0);
      act_routine = lprop_get_routine_next();
    }
  }

  {
    char dir_to_delete [256];

    if (erase)
      strcpy(dir_to_delete, PPL_DIR);
    else
      sprintf(dir_to_delete, PPL_DIR "/.%s", arch_names[arch]);
  
    if (verbose)
      printf("deleting %s\n", dir_to_delete);
      
    {
      char *argv[] = 
	{NULL, "-rf", NULL, NULL};
  
      argv[0] = RM_NAME;
      argv[2] = dir_to_delete;
      run_cmd(RM_NAME, argv, verbose);
    }
  
    if (verbose)
      printf("%s deleted\n", dir_to_delete);
  }
}

/*-------------------------------------------------------------------------*/
/* view_config:                                                            */
/* Action 'v'.                                                             */
/*-------------------------------------------------------------------------*/
void view_config (void)
{
  parse();
  
  lprop_print_report();
  
  if (recursive)
  {                               /* configure all routines of the library */
    char *act_routine;
    char *argv[] = 
      {NULL,                                               /* program name */
      "v",                                                       /* action */
      "-a", NULL,                                          /* architecture */
      "-d", NULL,                                     /* routine directory */
      "-v", "-V",                                        /* verbose option */
      NULL};
    char ppr_call [256];
    
    sprintf(ppr_call, "%s/lib/%s/pprcmd", ppc_home, HOST_ARCH);
    
    argv[0] = ppr_call;
    argv[3] = arch_names[arch];
    if (!verbose)
      argv[6] = argv[7] = NULL;
    
    lprop_get_routine_init();
    act_routine = lprop_get_routine_next();
    while(act_routine != NULL)
    {
      argv[5] = act_routine;                          /* routine directory */
      run_cmd(ppr_call, argv, 0);
      act_routine = lprop_get_routine_next();
    }
  }
}

/*-------------------------------------------------------------------------*/
/* install:                                                                */
/* Install parallel library.                                               */
/*-------------------------------------------------------------------------*/
void install (void)
{
  char dir_name [256];
  char makefile_name [256];

  if (verbose)
    printf("installing %s library in %s\n", ppl_file, install_directory);
    
  {                                    /* creates parlib include directory */
    gen_create_dir(install_directory, verbose);
    sprintf(dir_name, "%s/include", install_directory);
    gen_create_dir(dir_name, verbose);
    sprintf(dir_name, "%s/include/parlib", install_directory);
    gen_create_dir(dir_name, verbose);
  }
  {                                           /* creates library directory */
    gen_create_dir(install_directory, verbose);
    sprintf(dir_name, "%s/%s", install_directory, arch_names[arch]);
    gen_create_dir(dir_name, verbose);
    sprintf(dir_name, "%s/%s/%s", 
      install_directory, arch_names[arch], HOST_ARCH);
    gen_create_dir(dir_name, verbose);
  }

  if (arch == PACTC_ARCH)
    sprintf(makefile_name, "%s/bin/config.pactc.mak", ppc_home);
  else
    sprintf(makefile_name, "%s/bin/config.mak", ppc_home);
  
  {                                         /* copy the library using make */
    char *argv[] = 
      {NULL, "-f", NULL, "parlibinstall", NULL, NULL, NULL, 
      NULL, "PPL_DIR=" PPL_DIR, NULL};
    char arch_option [100];
    char host_arch [100];
    char app_name [100];
    char install_dir_opt [MAX_FILE_NAME_LEN];

    argv[0] = MAKE_NAME;
    argv[2] = makefile_name;
    sprintf(arch_option, "ARCH=%s", arch_names[arch]);
    argv[4] = arch_option;
    sprintf(host_arch, "HOST_ARCH=%s", HOST_ARCH);
    argv[5] = host_arch;
    sprintf(app_name, "LIB_NAME=%s", ppl_file);
    argv[6] = app_name;
    sprintf(install_dir_opt, "INSTALLDIR=%s", install_directory);
    argv[7] = install_dir_opt;
    run_cmd(MAKE_NAME, argv, verbose);
  }

  if (verbose)
    printf("%s library installed\n", ppl_file);
}

/*-------------------------------------------------------------------------*/
/* process_key:                                                            */
/* Perform the actions in key.                                             */
/*-------------------------------------------------------------------------*/
void process_key (
  char *key
)
{
  int action_idx;

  for (action_idx = 0; action_idx < strlen(key); action_idx++)
  {
    switch(key[action_idx])
    {
      case 'c': configure();   break;
      case 'm': make();        break;
      case 'd': delete();      break;
      case 'v': view_config(); break;
      case 'i': install();     break;
      default:                 break;
    }
  }
}

/*-------------------------------------------------------------------------*/
/* main:                                                                   */
/* Utility entry point.                                                    */
/*-------------------------------------------------------------------------*/
void main (int argc, char *argv[])
{
  char *ppc_home_ptr;

  if (argc == 1)                            /* check if a key is specified */
    print_usage("missing key");
    
  if (strchr(argv[1], 'h'))           /* check if help action is specified */
    print_help();
    
  check_key(argv[1]);      /* check if key is combination of valid actions */
  
  if ((ppc_home_ptr = getenv(PPC_HOME_NAME)) == NULL)
  {
    printf("%s: %s environment variable not defined\n",
     PROG_NAME, PPC_HOME_NAME);
    exit(4);
  }
  strcpy(ppc_home, ppc_home_ptr);                   /* initialize PPC home */
  
  arch = UNIX_ARCH;
  strcpy(ppl_file, "");
  verbose = 0;
  warnings = 0;
  erase = 0;
  recursive = 0;
  sprintf(install_directory, "%s/parlib", ppc_home);
  
  process_options(argc, argv);

  get_dir_files();     /* get names of the files found in the library dir. */

  if (strchr(argv[1], 'c') || strchr(argv[1], 'm') || 
      strchr(argv[1], 'v') || strchr(argv[1], 'd') || 
      strchr(argv[1], 'i') || recursive)
  {
    if (strlen(ppl_file))
      check_ppl_file();                         /* check optional PPL file */
    else
      found_ppl_file();             /* found PPL file in library directory */
  }

  check_arch();                          /* check architecture environment */
  
  if (verbose)
    print_summary();

  process_key(argv[1]);                    /* perform the required actions */

  exit(0);
}
