/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Portable Parallel Routines command line utility.                        */
/* PPR configuration file parser. The function ppr_parse() parse the       */
/* .ppr file of the routine and update the properties database             */
/* accordingly.                                                            */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 01.11.1994  Droux       Created                                         */
/* 23.01.1995  Droux       Use portability header file                     */
/*-------------------------------------------------------------------------*/
/* filename: parser.c                                                      */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define PARSER_IMPL

#include <stdio.h>

#include <ppc/portable.h>
#include "parser.h"
#include "prop.h"

#define RETURN_ERROR(code) {parser_errno = code; return -1;}

#define EO_ITEMS -1
#define EO_WORDS -1
#define EO_LINE -1
#define MAX_ITEM_LEN 100
#define MAX_LINE_LEN 1000

typedef enum _parserStates
{
  PARSE_IDLE,
  PARSE_TOPOLOGY,
  PARSE_PARAMETERS,
  PARSE_SLAVE_MODULES,
  PARSE_MASTER_MODULES
} parserStates;

static FILE *ppr_file;

/*-------------------------------------------------------------------------*/
/* get_next_line:                                                          */
/* Get the next line from the configuration file.                          */
/* Filters empty lines and line beginning with a comment character.        */
/* Return EO_LINE if no more lines are available, zero otherwise.          */
/*-------------------------------------------------------------------------*/
int get_next_line (
  char *line
)
{
  static int eo_line = 0;
  static char act_line [MAX_LINE_LEN];
  static char *fgets_stat;
  static int valid;

  if (eo_line == EO_LINE)
    return EO_LINE;

  do {
    fgets_stat = fgets(act_line, MAX_LINE_LEN, ppr_file);
    parser_act_line++;
    valid = strlen(act_line) && act_line[0] != '#';
  } while(fgets_stat != NULL && !valid);

  if (fgets_stat == NULL)
    eo_line = EO_LINE;

  strcpy(line, act_line);

  return 0;
}

/*-------------------------------------------------------------------------*/
/* get_next_word:                                                          */
/* Get the next word from the configuration file. Filters comments.        */
/* Return EO_WORDS if no more words are available, zero otherwise.         */
/*-------------------------------------------------------------------------*/
int get_next_word (
  char *word
)
{
  static int eo_words = 0;
  static int eo_line = EO_LINE;
  static char act_line[MAX_LINE_LEN];
  static char *cptr;
  int idx;

  if (eo_words == EO_WORDS)
    return EO_WORDS;
    
  if (eo_line)
  {
    eo_line = get_next_line (act_line);
    if (eo_line == EO_LINE)
    {
      eo_words = 1;
      return EO_WORDS;
    }
    cptr = act_line;
  }
  
  for (;;)
  {
    for (; *cptr == ' ' || *cptr == '\t'; cptr++);          /* skip blanks */
    for (idx = 0; 
         *cptr != ' ' && *cptr != '\t' && *cptr != '\0' && 
	 *cptr != '\n' && *cptr != '#' && idx < MAX_ITEM_LEN; 
         idx++)
    {
      word[idx] = *cptr++;
    }
    
    if (idx == 0)
    {
      switch(*cptr)
      {
        case ' ':
	  continue;
	  break;
	default:
	  eo_line = get_next_line (act_line);
	  if (eo_line == EO_LINE)
	  {
	    eo_words = 1;
	    return EO_WORDS;
	  }
	  cptr = act_line;
	  continue;
	  break;
      }
    }
    word[idx] = '\0';
    return 0;
  }
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* get_next_item:                                                          */
/* Get the next item from the configuration file and put it in the         */
/* string 'item'.                                                          */
/* Returns EO_ITEMS if no more items are available from the file,          */
/* or zero otherwise.                                                      */
/*-------------------------------------------------------------------------*/
int get_next_item (
  char *item
)
{
  static int eo_items = 0;
  static char act_item [MAX_ITEM_LEN];

  if (eo_items == EO_ITEMS)
    return EO_ITEMS;

  if (get_next_word(act_item) == EO_WORDS)
    return EO_ITEMS;
    
  strcpy(item, act_item);

  return 0;
}

/*-------------------------------------------------------------------------*/
/* get_state_from_property:                                                */
/* Returns the parser state corresponding to the specified property.       */
/*-------------------------------------------------------------------------*/
parserStates get_state_from_property (
  properties_enum prop
)
{
  switch(prop)
  {                       
    case PROP_TOPOLOGY:
      return PARSE_TOPOLOGY;
    case PROP_PARAMETERS:
      return PARSE_PARAMETERS;
    case PROP_MASTER_MODULES:
      return PARSE_MASTER_MODULES;
    case PROP_SLAVE_MODULES:
      return PARSE_SLAVE_MODULES;
    default:
      return NB_PROPERTIES;
  }
}

/*-------------------------------------------------------------------------*/
/* process_item:                                                           */
/* Process the specified item and update the properties database           */
/* accordingly.                                                            */
/* Returns 0 on success, or an error code on failure.                      */
/*-------------------------------------------------------------------------*/
int process_item (
  char *item
)
{
  static parserStates parser_state = PARSE_IDLE;
  
  switch(parser_state)
  {
    case PARSE_IDLE:                         /* item should be a propoerty */
    {
      int c;
      for (c = 0; c < NB_PROPERTIES && strcmp(item, properties_str[c]); c++);
      if (c == NB_PROPERTIES)                          /* unknown property */
        return PERR_UNKNOWN_PROP;
      parser_state = get_state_from_property(c);
      break;
    }
    case PARSE_TOPOLOGY:
    {
      int c;
      for (c = 0; c < NB_TOPOLOGIES && strcmp(item, topologies_str[c]); c++);
      if (c == NB_TOPOLOGIES)                          /* unknown topology */
        return PERR_UNKNOWN_TOPO;
      prop_set_topology(c);
      parser_state = PARSE_IDLE;
      break;
    }
    case PARSE_PARAMETERS:
    {
      int c;
      for (c = 0; 
           c < NB_PARAMETER_TYPES && strcmp(item, parameter_types_str[c]); 
	   c++);
      if (c < NB_PARAMETER_TYPES)                        /* parameter type */
      {
        prop_new_parameter(c);
        break;
      }
      for (c = 0; c < NB_DIRECTIONS && strcmp(item, directions_str[c]); c++);
      if (c < NB_DIRECTIONS)
      {                                             /* parameter direction */
        if (prop_param_direction(c) == -1)
	  return PERR_SYNTAX_DIR;
	break;
      }
      for (c = 0; c < NB_PROPERTIES && strcmp(item, properties_str[c]); c++);
      if (c == NB_PROPERTIES)                          
        return PERR_UNKNOWN_PARAM;
      parser_state = get_state_from_property(c);
      break;
    }
    case PARSE_SLAVE_MODULES:
    {
      int c;
      for (c = 0; c < NB_PROPERTIES && strcmp(item, properties_str[c]); c++);
      if (c < NB_PROPERTIES)
      {
        parser_state = get_state_from_property(c);
        break;
      }
      prop_new_slave_module(item);
      break;
    }
    case PARSE_MASTER_MODULES:
    {
      int c;
      for (c = 0; c < NB_PROPERTIES && strcmp(item, properties_str[c]); c++);
      if (c < NB_PROPERTIES)
      {
        parser_state = get_state_from_property(c);
        break;
      }
      prop_new_master_module(item);
      break;
    }
    default:
      break;
  }
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* ppr_parse:                                                              */
/* Parse the PPR configuration file (with extension .ppr) and update       */
/* the properties database accordingly.                                    */
/* If an error occurs then returns -1 and set the parser_errno global      */
/* variable, returns 0 otherwise.                                          */
/*-------------------------------------------------------------------------*/
int ppr_parse (
  char *routine_name,
  int verbose                            /* 1 if verbose mode, 0 otherwise */
)
{
  char full_name [100];
  char item [MAX_ITEM_LEN];
  int item_stream;
  int process_status;
  
  sprintf(full_name, "%s.ppr", routine_name);
  
  if (verbose)
    printf("starting parsing of %s\n", full_name);
  prop_init(routine_name);
  
  parser_act_line = 0;
  
  ppr_file = fopen(full_name, "r");
  if (ppr_file == NULL)
    RETURN_ERROR(PERR_COULD_NOT_OPEN);

  item_stream = get_next_item(item);                    /* look ahead read */
  while(item_stream != EO_ITEMS)
  {                                                /* process current item */
    process_status = process_item(item);
    if (process_status)
      RETURN_ERROR(process_status);
    item_stream = get_next_item(item);
  }
  
  if (verbose)
    printf("%s parsed successfully\n", full_name);
  
  return 0;
}
