/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Portable Parallel Routines command line utility.                        */
/* Main.                                                                   */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.10.1994  Droux       Created                                         */
/* 23.01.1995  Droux       Use portability header file                     */
/* 20.03.1995  Droux       Ported to the Paragon                           */
/* 16.05.1995  Droux       Ported to the SP-2                              */
/*-------------------------------------------------------------------------*/
/* filename: pprcmd.c                                                      */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   NEXTSTEP 3.3, Intel                                                   */
/*   Solaris 2.4, SPARC                                                    */
/*   Solaris 2.4, SPARC Multiprocessor                                     */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <ppc/portable.h>
#include <ppc/csp/version.h>
#include "parser.h"
#include "prop.h"
#include "rgenerator.h"
#include "rgenunix.h"
#include "rgenpvm.h"
#include "rgenpactc.h"
#include "rgenpgon.h"
#include "genutil.h"
#include "toolsutil.h"
#include "rgendefs.h"

#ifdef NEXT
  #define MAKE_NAME "/bin/make"
#elif defined(LINUX)
  #define MAKE_NAME "/usr/bin/make"
#elif defined(SUN4SOL2)
  #define MAKE_NAME "/usr/ccs/bin/make"
#else
  #define MAKE_NAME "/bin/make"
#endif

#define RM_NAME "/bin/rm"

#define host_arch_human(x) x              /* to be replaced by function @@ */

#define PROG_NAME "ppr"
#define ALLOWED_ACTIONS "cdghmv"

struct dir_entry
{
  char name [MAX_FILE_NAME_LEN];
  struct dir_entry *next;
};

extern int optind;                                     /* used by getopt() */
extern char *optarg;

static char ppc_home [MAX_FILE_NAME_LEN];            /* PPC home directory */
static char routine_dir [MAX_FILE_NAME_LEN];          /* routine directory */
static char ppr_file [MAX_FILE_NAME_LEN];               /* PPR config file */
static int verbose;                                   /* 1 if verbose mode */
static int warnings;                      /* 1 if warnings must be printed */
static arch_enum arch;                              /* target architecture */
static struct dir_entry *files_list = NULL;   /* list of routine dir files */
static char pvm_home [MAX_FILE_NAME_LEN];            /* PVM home directory */
static char pactc_home [MAX_FILE_NAME_LEN];       /* PACT-C home directory */
static int gen_master_only = 0;              /* generates master file only */
static int gen_slave_only = 0;                /* generates slave file only */
static int gen_file_name_opt = 0;              /* 1 if -g option specified */
static char gen_file_name [MAX_FILE_NAME_LEN];          /* option -g value */
static int gen_stdout = 0;                          /* generates to stdout */
static int erase = 0;                                         /* option -e */
static int limited_verbose = 0;     /* when tool is called by another tool */

/*-------------------------------------------------------------------------*/
/* print_usage:                                                            */
/* Print usage and exit.                                                   */
/*-------------------------------------------------------------------------*/
void print_usage (char *msg)
{
  if (msg != NULL)
    printf("%s: %s\n", PROG_NAME, msg);
  printf("%s: usage: {%s} -[adefgmosvw]\n", 
    PROG_NAME, ALLOWED_ACTIONS);
  printf("%s: type '%s h' for help\n", PROG_NAME, PROG_NAME);
  exit(1);
}

/*-------------------------------------------------------------------------*/
/* print_title:                                                            */
/* Print program title.                                                    */
/*-------------------------------------------------------------------------*/
void print_title (void)
{
  printf("PPC routines manager. " PPC_VERSION "\n");
  printf("Copyright (c) 1993, 1994, 1995 Engineering School of Biel-Bienne.\n");
}

/*-------------------------------------------------------------------------*/
/* print_help:                                                             */
/* Print help and exit.                                                    */
/*-------------------------------------------------------------------------*/
void print_help (void)
{
  print_title();
  printf("Usage: ppr key [options]\n");
  printf("key must be a combination of:\n");
  printf("  c         Configure routine.\n");
  printf("  m         Make routine.\n");
  printf("  g         Generates master and slave skeletons.\n");
  printf("  d         Delete hardware dependent files.\n");
  printf("  v         Print routine configuration report.\n");
  printf("  h         Print this message.\n");
  printf("options:\n");
  printf("  -a arch   Specify target architecture. Default is unix. Allowed\n");
  printf("            architectures are unix, pvm, pactc, paragon, and sp2.\n");
  printf("  -f file   Specify routine configuration file. Default is the\n");
  printf("            first .ppr file found in routine directory.\n");
  printf("  -d file   Specify routine directory. Default is the working directory.\n");
  printf("  -m        Used with 'g' key, generates only the master file.\n");
  printf("  -s        Used with 'g' key, generates only the slave file.\n");
  printf("  -g file   Used with 'g' key and '-m' or '-s' options, specify the\n");
  printf("            name of the file to generate.\n");
  printf("  -o        Used with 'g' key and '-m' or '-s' options, redirect to stdout.\n");
  printf("  -e        Used with 'd' key, delete all files generated by ppr\n");
  printf("  -v        Run the utility in verbose mode.\n");
  printf("  -w        Shows all warnings during routine compilation.\n");
  exit(2);
}

/*-------------------------------------------------------------------------*/
/* check_key:                                                              */
/* Check if the given key is valid. If not, prints an error message,       */
/* the program usage and exit.                                             */
/*-------------------------------------------------------------------------*/
void check_key (char *key)
{
  int c;
  char err_msg [100];
  
  for (c = 0; c < strlen(key); c++)
    if (strchr(ALLOWED_ACTIONS, key[c]) == NULL)
    {
      sprintf(err_msg, "%c: unknown action", key[c]);
      print_usage(err_msg);
    }
}

/*-------------------------------------------------------------------------*/
/* process_options:                                                        */
/* Parse the specified options and update the corresponding global         */
/* variables.                                                              */
/* If an error occures, then print an error message, the program usage,    */
/* and exit.                                                               */
/*-------------------------------------------------------------------------*/
void process_options (int argc, char **argv)
{
  int option;                            /* current option being processed */

  while((option = getopt(argc-1, &argv[1], "a:f:d:msg:ovweV")) != EOF)
  {
    switch(option)
    {
      case 'a':                            /* optional target architecture */
      {                                    
        int correct = 0;
        int act_arch;
	
	if (optarg == NULL)
          print_usage("option -a must be followed by unix, pvm, pactc, or paragon");
        for (act_arch = 0; act_arch < NB_ARCH && !correct; act_arch++)
	  correct = !strcmp(optarg, arch_names[act_arch]);
	if (!correct)
	  print_usage("specified architecture incorrect (must be unix, pvm, pactc, paragon, or sp2)");
	arch = act_arch-1;
        break;
      }
      case 'f':                        /* optional configuration file name */
      {
        if (optarg == NULL)
          print_usage("option -f must be followed by a file name");
        strcpy(ppr_file, optarg);             /* validity is checked later */
	break;
      }
      case 'd':                              /* optional routine directory */
      {
        struct stat pathStat;
      
        if (optarg == NULL)
          print_usage("option -d must be followed by a directory");
        if(stat(optarg, &pathStat) == -1)
          print_usage("specified routine directory does not exist or is not accessible");
	if (!(pathStat.st_mode & S_IFDIR))
	  print_usage("specified routine path is not a directory");
	strcpy(routine_dir, optarg);
	if (chdir(routine_dir))
	  print_usage("cannot go to specified routine directory");
        break;
      }
      case 'm':                              /* generates master file only */
        gen_master_only = 1;
	break;
      case 's':                               /* generates slave file only */
        gen_slave_only = 1;
	break;
      case 'g':            /* name of the slave or master file to generate */
        if (gen_stdout)
	  print_usage("options -g and -o are mutually exclusive");
        gen_file_name_opt = 1;
        if (optarg == NULL)
          print_usage("option -g must be followed by a file name");
	strcpy(gen_file_name, optarg);
	break;
      case 'o':                           /* redirect generation to stdout */
        if (gen_file_name_opt)
	  print_usage("options -g and -o are mutually exclusive");
	gen_stdout = 1;
	break;
      case 'v':                                  /* toggle to verbose mode */
        verbose = 1;
	break;
      case 'w':
        warnings = 1;
	break;
      case 'e':
        erase = 1;
	break;
      case 'V':
        limited_verbose = 1;
	break;
      default:
        print_usage(NULL);
	break;
    }
  }
}

/*-------------------------------------------------------------------------*/
/* get_ext:                                                                */
/* Return in ext the extension of the given file name. ext is set to an    */
/* empty string if the file has no extension.                              */
/*-------------------------------------------------------------------------*/
void get_ext (
  char *name,
  char *ext
)
{
  int c;
  int i;
  int name_len = strlen(name);
  
  for (c = name_len-1, i = 0; c && (name[c] != '.'); c--, i++);
  if (c)
  {
    ext[i] = '\0';
    for (c = 0, i--; i >= 0; i--, c++)
      ext[c] = name[name_len-1-i];
  }
  else
    ext[0] = '\0';
}

/*-------------------------------------------------------------------------*/
/* check_ppr_file:                                                         */
/* Checks if the currently defined PPR file exists.                        */
/*-------------------------------------------------------------------------*/
void check_ppr_file (void)
{
  struct stat file_stat;
  char ext [100];

  if(stat(ppr_file, &file_stat) == -1)
  {                               /* <file> does not exist, try <file>.ppr */
    char file_plus_ext [MAX_FILE_NAME_LEN];
    sprintf(file_plus_ext, "%s.ppr", ppr_file);
    if(stat(file_plus_ext, &file_stat) == -1)
      print_usage("specified routine configuration file does not exist");
  }

  get_ext(ppr_file, ext);             /* suppress .ppr extension if exists */
  if (!strcmp(ext, "ppr"))
    ppr_file[strlen(ppr_file)-strlen(ext)-1] = '\0';
}

/*-------------------------------------------------------------------------*/
/* get_dir_files:                                                          */
/* Initialize the list files_list with the files found in the routine      */
/* directory.                                                              */
/*-------------------------------------------------------------------------*/
void get_dir_files (void)
{
  DIR *dir;
  struct direct *dirEnt;
  struct dir_entry *act_entry = files_list;
  struct dir_entry *new_entry;
  
  dir = opendir(".");
  for (dirEnt = readdir(dir); dirEnt != NULL; dirEnt = readdir(dir))
  {
    if (!strcmp(dirEnt->d_name, ".") || !strcmp(dirEnt->d_name, ".."))
      continue;
    new_entry = malloc(sizeof(struct dir_entry));
    if (files_list == NULL)
      files_list = new_entry;
    else
      act_entry->next = new_entry;
    act_entry = new_entry;
    strcpy(act_entry->name, dirEnt->d_name);
  }
}

/*-------------------------------------------------------------------------*/
/* found_ppr_file:                                                         */
/* Found the first .ppr file within the routine directory and update the   */
/* ppr_file global variable accordingly.                                   */
/*-------------------------------------------------------------------------*/
void found_ppr_file (void)
{
  struct dir_entry *act_entry = files_list;
  char ext [100];
  int found = 0;
  
  while (act_entry != NULL && !found)
  {
    get_ext(act_entry->name, ext);
    if (!(found = !strcmp(ext, "ppr")))
      act_entry = act_entry->next;
  }

  if (!found)
    print_usage("no routine configuration file found in active directory");
  
  strcpy(ppr_file, act_entry->name);
  ppr_file[strlen(ppr_file)-strlen(ext)-1] = '\0';
}

/*-------------------------------------------------------------------------*/
/* check_arch:                                                             */
/* Some architecture may require special environment definitions.          */
/*-------------------------------------------------------------------------*/
void check_arch (void)
{
  switch(arch)  /* special operations depending on the target architecture */
  {   
    case PVM_ARCH:
    {                                            /* get PVM home directory */
      char *pvm_home_ptr;
      
      if ((pvm_home_ptr = getenv(PVM_HOME_NAME)) == NULL)
      {
	printf("%s: %s environment variable not defined\n",
	  PROG_NAME, PVM_HOME_NAME);
	exit(4);
      }
      strcpy(pvm_home, pvm_home_ptr);               /* initialize PVM home */
      break;
    }
    case PACTC_ARCH:
    {                                         /* get PACT-C home directory */
      char *pactc_home_ptr;
      
      if ((pactc_home_ptr = getenv(PACTC_HOME_NAME)) == NULL)
      {
	printf("%s: %s environment variable not defined\n",
	  PROG_NAME, PACTC_HOME_NAME);
	exit(4);
      }
      strcpy(pactc_home, pactc_home_ptr);        /* initialize PACT-C home */
      break;
    }
    default:
      break;
  }
}

/*-------------------------------------------------------------------------*/
/* print_summary:                                                          */
/* Summarize the actual tool configuration state.                          */
/*-------------------------------------------------------------------------*/
void print_summary (void)
{
  print_title();
  printf("\n");
  printf("target architecture        : %s\n", output_arch_names[arch]);
  if (strlen(ppr_file))
    printf("routine configuration file : %s\n", ppr_file);
  printf("routine directory          : %s\n", routine_dir);
  printf("host computer architecture : %s\n", 
   host_arch_human(HOST_ARCH));
  printf("compilation warnings       : ");
  if (warnings)
    printf("Enabled\n");
  else
    printf("Disabled\n");
  printf("\n");
}

/*-------------------------------------------------------------------------*/
/* parse:                                                                  */
/* Parse the configuration file.                                           */
/*-------------------------------------------------------------------------*/
void parse (void)
{
  static int parsed = 0;

  if (!parsed)
  {
    if (ppr_parse(ppr_file, verbose))
    {
      printf("%s.ppr:", ppr_file);
      if (parser_act_line)
        printf("%d: ", parser_act_line);
      else
        printf(" ");
      printf("%s\n", parser_errors[parser_errno]);
      exit(1);
    }
    parsed = 1;
  }
}

/*-------------------------------------------------------------------------*/
/* configure:                                                              */
/* Action 'c'.                                                             */
/*-------------------------------------------------------------------------*/
void configure (void)
{
  if (verbose)
    printf("Configuring %s routine for %s\n", ppr_file, 
      output_arch_names[arch]);

  parse();
  if (rgen_user_interface(verbose) < 0)
    exit(1);
  if (rgen_master_starter(verbose) < 0)
    exit(1);
  if (rgen_master_header(verbose) < 0)
    exit(1);
  if (rgen_slave_starter(verbose) < 0)
    exit(1);
  if (rgen_slave_header(verbose) < 0)
    exit(1);
  if (rgen_object_list(verbose) < 0)
    exit(1);
  
  switch(arch)
  {
    case UNIX_ARCH:
      rgen_unix_files(verbose);
      break;
    case SP2_ARCH:
    case PVM_ARCH:
      rgen_pvm_files(verbose);
      break;
    case PACTC_ARCH:
      rgen_pactc_files(verbose);
      break;
    case PARAGON_ARCH:
      rgen_pgon_files(verbose);
      break;
    default:
      exit(1);
  }
  
  if (verbose)
  {
    printf("%s routine configured for %s\n", ppr_file, 
      output_arch_names[arch]);
    if (!limited_verbose)
      printf("\n");
  }
}

/*-------------------------------------------------------------------------*/
/* make:                                                                   */
/* Action 'm'.                                                             */
/*-------------------------------------------------------------------------*/
void make (void)
{
  char other_libs [100];
  char makefile_name [100];
  char warnings_opt [10];
  char bin_dir [256];

  if (verbose)
    printf("Making %s routine for %s\n", ppr_file, output_arch_names[arch]);
  
  if (arch == PACTC_ARCH)
  {
    sprintf(bin_dir, "%s/.%s/PACTC",
      PPR_DIR, arch_names[arch]);
  }
  else
  {
    sprintf(bin_dir, "%s/.%s/%s",
      PPR_DIR, arch_names[arch], HOST_ARCH);
  }
  gen_create_dir(bin_dir, verbose);
    
  sprintf(makefile_name, "%s/.%s/makefile", PPR_DIR, arch_names[arch]);
  
  #ifdef SUN4SOL2
    if (arch != SP2_ARCH)
      strcpy(other_libs, "-lnsl -lsocket");
    else
      other_libs[0] = '\0';
  #else
    other_libs[0] = '\0';
  #endif
  
  if (warnings)
    strcpy(warnings_opt, "-Wall");
  else
    warnings_opt[0] = '\0';

  {
    char *argv[] = 
      {NULL, "-f", NULL, NULL, NULL, NULL, NULL, NULL};
    char arch_option [100];
    char host_arch [100];
    char other_libs_opt [100];
    char warning_flag [100];

    argv[0] = MAKE_NAME;
    argv[2] = makefile_name;
    sprintf(arch_option, "ARCH=%s", arch_names[arch]);
    argv[3] = arch_option;
    sprintf(host_arch, "HOST_ARCH=%s", HOST_ARCH);
    argv[4] = host_arch;
    sprintf(other_libs_opt, "OTHERLIBS=%s", other_libs);
    argv[5] = other_libs_opt;
    sprintf(warning_flag, "WARNING_FLAG=%s", warnings_opt);
    argv[6] = warning_flag;
    run_cmd(MAKE_NAME, argv, verbose);
  }

  if (verbose)
  {
    printf("%s routine made for %s\n", ppr_file, output_arch_names[arch]);
    if (!limited_verbose)
      printf("\n");
  }
}

/*-------------------------------------------------------------------------*/
/* delete:                                                                 */
/* Action 'd'.                                                             */
/*-------------------------------------------------------------------------*/
void delete (void)
{
  char dir_to_delete [256];
  
  if (erase)
    strcpy(dir_to_delete, PPR_DIR);
  else
    sprintf(dir_to_delete, PPR_DIR "/.%s", arch_names[arch]);
    
  if (verbose)
    printf("deleting %s\n", dir_to_delete);
    
  {
    char *argv[] = 
      {NULL, "-rf", NULL, NULL};

    argv[0] = RM_NAME;
    argv[2] = dir_to_delete;
    run_cmd(RM_NAME, argv, verbose);
  }

  if (verbose)
    printf("%s deleted\n", dir_to_delete);
}

/*-------------------------------------------------------------------------*/
/* view_config:                                                            */
/* Action 'v'.                                                             */
/*-------------------------------------------------------------------------*/
void view_config (void)
{
  parse();
  prop_print_report();
}

/*-------------------------------------------------------------------------*/
/* generate:                                                               */
/* Action 'g'.                                                             */
/*-------------------------------------------------------------------------*/
void generate (void)
{
  gen_method method;
  char *optional_name = NULL;
  
  parse();

  if (gen_stdout)
    method = GEN_STDOUT;
  else if (gen_file_name_opt)
  {
    method = GEN_USE_NAME;
    optional_name = gen_file_name;
  }
  else
    method = GEN_DEFAULT;
  
  if (!gen_slave_only)
  {
    if (rgen_master(method, optional_name, verbose) < 0)
    {
      fprintf(stderr, "%s: error generating master skeleton\n", PROG_NAME);
      exit(1);
    }
  }

  if (!gen_master_only)
  {
    if (rgen_slave(method, optional_name, verbose) < 0)
    {
      fprintf(stderr, "%s: error generating slave skeleton\n", PROG_NAME);
      exit(1);
    }
  }
}

/*-------------------------------------------------------------------------*/
/* process_key:                                                            */
/* Perform the actions in key.                                             */
/*-------------------------------------------------------------------------*/
void process_key (
  char *key
)
{
  int action_idx;

  for (action_idx = 0; action_idx < strlen(key); action_idx++)
  {
    switch(key[action_idx])
    {
      case 'c': configure();   break;
      case 'm': make();        break;
      case 'd': delete();      break;
      case 'v': view_config(); break;
      case 'g': generate();    break;
      default:                 break;
    }
  }
}

/*-------------------------------------------------------------------------*/
/* main:                                                                   */
/* Utility entry point.                                                    */
/*-------------------------------------------------------------------------*/
void main (int argc, char *argv[])
{
  char *ppc_home_ptr;

  if (argc == 1)                            /* check if a key is specified */
    print_usage("missing key");
    
  if (strchr(argv[1], 'h'))           /* check if help action is specified */
    print_help();
    
  check_key(argv[1]);      /* check if key is combination of valid actions */
  
  if ((ppc_home_ptr = getenv(PPC_HOME_NAME)) == NULL)
  {
    printf("%s: %s environment variable not defined\n",
     PROG_NAME, PPC_HOME_NAME);
    exit(4);
  }
  strcpy(ppc_home, ppc_home_ptr);                   /* initialize PPC home */
  
  strcpy(routine_dir, "./");                               /* set defaults */
  arch = UNIX_ARCH;
  strcpy(ppr_file, "");
  verbose = 0;
  warnings = 0;
  
  process_options(argc, argv);

  get_dir_files();     /* get names of the files found in the routine dir. */

  if (strchr(argv[1], 'c') || strchr(argv[1], 'm') || 
      strchr(argv[1], 'v') || strchr(argv[1], 'g'))
  {
    if (strlen(ppr_file))
      check_ppr_file();                         /* check optional PPR file */
    else
      found_ppr_file();             /* found PPR file in routine directory */
  }

  check_arch();                          /* check architecture environment */
  
  if (verbose && !limited_verbose)
    print_summary();

  process_key(argv[1]);                    /* perform the required actions */

  exit(0);
}
