/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Portable Parallel Routines command line utility.                        */
/* Routine properties manager. This is the repository of the properties    */
/* of a parallel routine.                                                  */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 01.11.1994  Droux       Created                                         */
/* 05.04.1995  Droux       Added STRING routine parameter type             */
/* 01.05.1995  Droux       Added complex type                              */
/* 03.05.1995  Droux       Added boolean type                              */
/*-------------------------------------------------------------------------*/
/* filename: prop.c                                                        */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#define PROP_IMPL

#include <stdlib.h>
#include <stdio.h>
#include "prop.h"

#ifdef SUN4
  extern int printf(char *, ...);
#endif

static char *parameter_types_c[] =
{
  "int", 
  "short", 
  "unsigned short", 
  "long", 
  "unsigned long", 
  "float", 
  "double", 
  "char",
  "ppc_complex",
  "boolean",
  
  "int *", 
  "short *", 
  "unsigned short *", 
  "long *",
  "unsigned long *", 
  "float *", 
  "double *", 
  "char *",
  "ppc_complex *",
  "boolean *",

  "int **", 
  "short **", 
  "unsigned short **", 
  "long **",
  "unsigned long **", 
  "float **", 
  "double **", 
  "char **",
  "ppc_complex **",
  "boolean **",
  
  "char *",
};

/*-------------------------------------------------------------------------*/
/* new types:                                                              */
/*-------------------------------------------------------------------------*/

/* The parameters of a routine are stored in a linked list. Each parameter */
/* is characterized by its type and its direction.                         */
typedef struct parameter_struct
{
  parameter_types_enum type;
  direction_type direction;
  struct parameter_struct *next;
} parameter_types;

/* The master and slave modules are stored in NULL terminated linked       */
/* lists of the type below.                                                */
typedef struct modules_struct
{
  char *name;
  struct modules_struct *next;
} module_names;

/* The properties of a parallel routine are described in a unique variable */
/* local to this module.                                                   */
typedef struct
{
  char name[100];
  topology_type topology;
  parameter_types *first_parameter;
  module_names *first_slave_module;
  module_names *first_master_module;
} routine_properties_type;

/*-------------------------------------------------------------------------*/
/* local variables:                                                        */
/*-------------------------------------------------------------------------*/

static routine_properties_type prop;                 /* routine properties */
static parameter_types *last_param = NULL;         /* last parameter added */
static module_names *last_master = NULL;       /* last master module added */
static module_names *last_slave = NULL;         /* last slave module added */
static parameter_types *last_param_read;            /* last parameter read */
static parameter_types *next_param_read;         /* next parameter to read */
static module_names *next_master_read;       /* next master module to read */
static module_names *next_slave_read;         /* next slave module to read */

/*-------------------------------------------------------------------------*/
/* prop_init:                                                              */
/* Initializes the properties database.                                    */
/*-------------------------------------------------------------------------*/
void prop_init (
  char *routine_name
)
{
  strcpy(prop.name, routine_name);
  prop.topology = TOPO_UNKNOWN;
  prop.first_parameter = NULL;
  prop.first_slave_module = NULL;
  prop.first_master_module = NULL;
}

/*-------------------------------------------------------------------------*/
/* prop_set_topology:                                                      */
/* Set the routine topology.                                               */
/*-------------------------------------------------------------------------*/
void prop_set_topology (
  topology_type topo
)
{
  prop.topology = topo;
}

/*-------------------------------------------------------------------------*/
/* prop_new_parameter:                                                     */
/* Add the specified parameter to the parameters list of the routine.      */
/*-------------------------------------------------------------------------*/
void prop_new_parameter (
  parameter_types_enum param
)
{
  parameter_types *ptr;
  
  if (last_param == NULL)
    ptr = prop.first_parameter = malloc(sizeof(parameter_types));
  else
    ptr = last_param->next = malloc(sizeof(parameter_types));

  ptr->type = param;
  ptr->direction = IN;
  ptr->next = NULL;

  last_param = ptr;
}

/*-------------------------------------------------------------------------*/
/* prop_param_direction:                                                   */
/* Set the direction of the last declared parameter.                       */
/* Returns -1 if no parameter has been previously defined.                 */
/*-------------------------------------------------------------------------*/
int prop_param_direction (
  direction_type dir
)
{
  if (last_param == NULL)
    return -1;
  
  last_param->direction = dir;
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* prop_new_master_module:                                                 */
/* Add the specified master module to the routine.                         */
/*-------------------------------------------------------------------------*/
void prop_new_master_module (
  char *module_name
)
{
  module_names *ptr;
  
  if (last_master == NULL)
    ptr = prop.first_master_module = malloc(sizeof(module_names));
  else
    ptr = last_master->next = malloc(sizeof(module_names));

  ptr->name = malloc(strlen(module_name)+1);
  strcpy(ptr->name, module_name);
  ptr->next = NULL;

  last_master = ptr;
}

/*-------------------------------------------------------------------------*/
/* prop_new_slave_module:                                                  */
/* Add the specified slave module to the routine.                          */
/*-------------------------------------------------------------------------*/
void prop_new_slave_module (
  char *module_name
)
{
  module_names *ptr;
  
  if (last_slave == NULL)
    ptr = prop.first_slave_module = malloc(sizeof(module_names));
  else
    ptr = last_slave->next = malloc(sizeof(module_names));

  ptr->name = malloc(strlen(module_name)+1);
  strcpy(ptr->name, module_name);
  ptr->next = NULL;

  last_slave = ptr;
}

/*-------------------------------------------------------------------------*/
/* print_spaces:                                                           */
/*-------------------------------------------------------------------------*/
void print_spaces (
  int n
)
{
  int c;
  
  for (c = 0; c < n; c++)
    printf(" ");
}

/*-------------------------------------------------------------------------*/
/* prop_print_report:                                                      */
/* Print routine configuration report.                                     */
/*-------------------------------------------------------------------------*/
void prop_print_report (void)
{
  parameter_types *param_ptr;
  module_names *module_ptr;
  
  printf("Parallel routine configuration:\n");
  printf("  Name:           \"%s\"\n", prop.name);
  printf("  Topology:       %s\n", topologies_str[prop.topology]);
  
  printf("  Parameters:     ");
  param_ptr = prop.first_parameter; 
  while(param_ptr != NULL)
  {
    if (param_ptr != prop.first_parameter)
      printf("                  ");
    printf(parameter_types_str[param_ptr->type]);
    print_spaces(25-strlen(parameter_types_str[param_ptr->type]));
    printf(directions_str[param_ptr->direction]);
    printf("\n");
    param_ptr = param_ptr->next;
  }
  
  printf("  Master modules: ");
  module_ptr = prop.first_master_module;
  while(module_ptr != NULL)
  {
    if (module_ptr != prop.first_master_module)
      printf(", ");
    printf("%s", module_ptr->name);
    module_ptr = module_ptr->next;
  }
  printf("\n");
  
  printf("  Slave modules:  ");
  module_ptr = prop.first_slave_module;
  while(module_ptr != NULL)
  {
    if (module_ptr != prop.first_slave_module)
      printf(", ");
    printf("%s", module_ptr->name);
    module_ptr = module_ptr->next;
  }
  printf("\n");
}

/*-------------------------------------------------------------------------*/
/* prop_get_master_name:                                                   */
/* Returns the first name of the master files appearing                    */
/* in the list of master files.                                            */
/*-------------------------------------------------------------------------*/
char *prop_get_master_name (void)
{
  return prop.first_master_module->name;
}

/*-------------------------------------------------------------------------*/
/* prop_get_slave_name:                                                    */
/* Returns the first name of the slave files appearing                     */
/* in the list of slave files.                                             */
/*-------------------------------------------------------------------------*/
char *prop_get_slave_name (void)
{
  return prop.first_slave_module->name;
}

/*-------------------------------------------------------------------------*/
/* prop_get_routine_name:                                                  */
/* Returns the name of the routine.                                        */
/*-------------------------------------------------------------------------*/
char *prop_get_routine_name (void)
{
  return prop.name;
}

/*-------------------------------------------------------------------------*/
/* prop_get_topology:                                                      */
/* Returns the topology of the routine.                                    */
/*-------------------------------------------------------------------------*/
topology_type prop_get_topology (void)
{
  return prop.topology;
}

/*-------------------------------------------------------------------------*/
/* prop_get_param_init:                                                    */
/* Must be called before any call to prop_get_param:                       */
/*-------------------------------------------------------------------------*/
void prop_get_param_init (void)
{
  next_param_read = prop.first_parameter;
}

/*-------------------------------------------------------------------------*/
/* prop_get_param:                                                         */
/* Returns the ANSI-C type of the next routine parameter.                  */
/* Returns NULL if no more parameters are available.                       */
/*-------------------------------------------------------------------------*/
char *prop_get_param (void)
{
  if (next_param_read == NULL)
    return NULL;
  last_param_read = next_param_read;
  next_param_read = next_param_read->next;
  return parameter_types_c[last_param_read->type];
}

/*-------------------------------------------------------------------------*/
/* prop_get_param_is_out:                                                  */
/* Returns 1 if the parameter returned by prop_get_param is an             */
/* output parameter.                                                       */ 
/*-------------------------------------------------------------------------*/
int prop_get_param_is_out (void)
{
  return last_param_read->direction == OUT;
}

/*-------------------------------------------------------------------------*/
/* prop_get_master_init:                                                   */
/* Must be called before any call to prop_get_master_next.                 */
/*-------------------------------------------------------------------------*/
void prop_get_master_init (void)
{
  next_master_read = prop.first_master_module;
}

/*-------------------------------------------------------------------------*/
/* prop_get_master_next:                                                   */
/* Return the next master module name, or NULL.                            */
/*-------------------------------------------------------------------------*/
char *prop_get_master_next (void)
{
  module_names *tmp;
   
  if (next_master_read == NULL)
    return NULL;
  
  tmp = next_master_read;
  next_master_read = next_master_read->next;

  return tmp->name;
}

/*-------------------------------------------------------------------------*/
/* prop_get_slave_init:                                                    */
/* Must be called before any call to prop_get_slave_next.                  */
/*-------------------------------------------------------------------------*/
void prop_get_slave_init (void)
{
  next_slave_read = prop.first_slave_module;
}

/*-------------------------------------------------------------------------*/
/* prop_get_slave_next:                                                    */
/* Return the next slave module name, or NULL.                             */
/*-------------------------------------------------------------------------*/
char *prop_get_slave_next (void)
{
  module_names *tmp;
   
  if (next_slave_read == NULL)
    return NULL;
  
  tmp = next_slave_read;
  next_slave_read = next_slave_read->next;

  return tmp->name;
}

