/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Portable Parallel Routines command line utility.                        */
/* Routines generator. This module can generate the files needed to        */
/* build a parallel routine.                                               */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 06.01.1995  Droux       Created                                         */
/* 07.03.1995  Droux       Include portable.h                              */
/* 01.04.1995  Droux       Fixed grid topology connections                 */
/* 01.05.1995  Droux       Local functions are declared as static          */
/*-------------------------------------------------------------------------*/
/* filename: rgenerator.c                                                  */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>

#include <ppc/portable.h>
#include <ppc/csp/version.h>
#include "rgenerator.h"
#include "genutil.h"
#include "prop.h"
#include "rgendefs.h"

#define USER_INTERFACE_NAME "user.h"             /* routine user interface */
#define MASTER_PROTO_NAME "master.h"                   /* master prototype */
#define SLAVE_PROTO_NAME "slave.h"                      /* slave prototype */

typedef enum
{
  SCALAR,
  ARRAY
} array_spec;

typedef struct
{
  char *type;
  char *name;
  array_spec is_array;
} param_type;

typedef param_type param_descr[];

static param_descr master_farm_param = 
{
  {"channel", "from_host", SCALAR},
  {"channel", "to_host", SCALAR},
  {"channel", "from_slaves", ARRAY},
  {"channel", "to_slaves", ARRAY},
  {"int", "nb_row", SCALAR},
  {"int", "nb_col", SCALAR},
  {NULL, NULL, 0}
};

static param_descr slave_farm_param = 
{
  {"channel", "from_master", SCALAR},
  {"channel", "to_master", SCALAR},
  {"int", "nb_slaves", SCALAR},
  {"int", "my_nb", SCALAR},
  {NULL, NULL, 0}
};

static param_descr slave_ring_param =
{
  {"channel", "from_master", SCALAR},
  {"channel", "to_master", SCALAR},
  {"channel", "from_next", SCALAR},
  {"channel", "to_next", SCALAR},
  {"channel", "from_prev", SCALAR},
  {"channel", "to_prev", SCALAR},
  {"int", "nb_slaves", SCALAR},
  {"int", "my_nb", SCALAR},
  {NULL, NULL, 0}
};

static param_descr slave_grid_param =
{
  {"channel", "from_master", SCALAR},
  {"channel", "to_master", SCALAR},
  {"channel", "from_north", SCALAR},
  {"channel", "to_north", SCALAR},
  {"channel", "from_south", SCALAR},
  {"channel", "to_south", SCALAR},
  {"channel", "from_west", SCALAR},
  {"channel", "to_west", SCALAR},
  {"channel", "from_east", SCALAR},
  {"channel", "to_east", SCALAR},
  {"int", "nb_row", SCALAR},
  {"int", "nb_col", SCALAR},
  {"int", "my_row", SCALAR},
  {"int", "my_col", SCALAR},
  {NULL, NULL, 0}
};

static param_descr *master_param[] = 
{
  &master_farm_param,
  &master_farm_param,
  &master_farm_param,
  NULL
};

static param_descr *slave_param[] = 
{
  &slave_farm_param,
  &slave_ring_param,
  &slave_grid_param,
  NULL
};

/*-------------------------------------------------------------------------*/
/* rgen_get_stream:                                                  LOCAL */
/* Returns a stream according to the specified generation method and       */
/* optional name.                                                          */
/* Returns NULL if an error occured.                                       */
/*-------------------------------------------------------------------------*/
static FILE *rgen_get_stream (
  gen_method method,
  char *default_name,
  char *optional_name,
  int verbose
)
{
  switch(method)
  {
    case GEN_DEFAULT:
    {
      if (verbose)
        printf("%s\n", default_name);
      return fopen(default_name, "w");
      break;
    }
    case GEN_STDOUT:
      if (verbose)
        printf("standard output\n");
      return stdout;
      break;
    case GEN_USE_NAME:
      if (verbose)
        printf("%s\n", optional_name);
      return fopen(optional_name, "w");
      break;
    default:
      break;
  }
  
  return NULL;
}

/*-------------------------------------------------------------------------*/
/* rgen_master:                                                            */
/* Generates the routine main master file.                                 */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
int rgen_master (
  gen_method method,
  char *optional_name,
  int verbose
)
{
  char default_name [100];
  FILE *stream;
  
  if (verbose)
    printf("generating master skeleton to ");
    
  strcpy(default_name, prop_get_master_name());
  strcat(default_name, ".c");
  stream = rgen_get_stream(method, default_name, optional_name, verbose);
  
  fprintf(stream, "\n/* %s parallel routine master */\n\n", 
    prop_get_routine_name());
  fprintf(stream, "#include <ppc/csp.h>\n");
  fprintf(stream, "#include <ppc/parlib.h>\n\n");
  fprintf(stream, "int ppr_main (\n");
  
  {                        /* standard parameters for the routine topology */
    int c;
    topology_type topo = prop_get_topology();
    
    for(c = 0; (*(master_param[topo]))[c].type != NULL; c++)
    {
      if (c > 0)
        fprintf(stream, ",\n");
      fprintf(stream, "  %s %s", 
        (*(master_param[topo]))[c].type,
	(*(master_param[topo]))[c].name);
      if ((*(master_param[topo]))[c].is_array == ARRAY)
        fprintf(stream, "[]");
    }
  }
  {                                   /* custom parameters for the routine */
    char *act_type;
    int c;
    
    prop_get_param_init();
    act_type = prop_get_param();
    for(c = 0; act_type != NULL; c++)
    {
      fprintf(stream, ",\n");
      fprintf(stream, "  %s ", act_type);
      if (!strchr(act_type, '*') && prop_get_param_is_out())
        fprintf(stream, "*");
      fprintf(stream, "p%d", c);
      act_type = prop_get_param();
    }
  }
  
  fprintf(stream, "\n)\n");
  fprintf(stream, "{\n");
  fprintf(stream, "  /* insert master code here */\n\n");
  fprintf(stream, "  return PPR_OK;\n");
  fprintf(stream, "}\n\n");
  
  if (stream != stdout)
    fclose(stream);
  
  if (verbose)
    printf("master skeleton generated\n");
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* rgen_slave:                                                             */
/* Generates the routine main slave file.                                  */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
int rgen_slave (
  gen_method method,
  char *optional_name,
  int verbose
)
{
  char default_name [100];
  FILE *stream;
  
  if (verbose)
    printf("generating slave skeleton to ");
    
  strcpy(default_name, prop_get_slave_name());
  strcat(default_name, ".c");
  stream = rgen_get_stream(method, default_name, optional_name, verbose);
  
  fprintf(stream, "\n/* %s parallel routine slave */\n\n", 
    prop_get_routine_name());
  fprintf(stream, "#include <ppc/csp.h>\n\n");

  fprintf(stream, "void ppr_main (\n");
  
  {                        /* standard parameters for the routine topology */
    int c;
    topology_type topo = prop_get_topology();
    
    for(c = 0; (*(slave_param[topo]))[c].type != NULL; c++)
    {
      if (c > 0)
        fprintf(stream, ",\n");
      fprintf(stream, "  %s %s", 
        (*(slave_param[topo]))[c].type,
	(*(slave_param[topo]))[c].name);
      if ((*(slave_param[topo]))[c].is_array == ARRAY)
        fprintf(stream, "[]");
    }
    fprintf(stream, "\n");
  }
  
  fprintf(stream, ")\n");
  fprintf(stream, "{\n");
  fprintf(stream, "  /* insert slave code here */\n");
  fprintf(stream, "}\n\n");
  
  if (stream != stdout)
    fclose(stream);
  
  if (verbose)
    printf("slave skeleton generated\n");
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* rgen_user_interface:                                                    */
/* Generates the routine user interface.                                   */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
int rgen_user_interface (
  int verbose
)
{
  FILE *out;
  char filename[256];
  
  if (verbose)
    printf("generating routine user interface\n");
  
  if (gen_create_dir(PPR_DIR, verbose) == -1)
    return -1;
    
  sprintf(filename, "%s/%s", PPR_DIR, USER_INTERFACE_NAME);
  out = fopen(filename, "w");
  if (out == NULL)
    return -1;
    
  fprintf(out, "\n");
  fprintf(out, "/* This file has been generated by the PPC ");
  fprintf(out, "routines manager */\n");
  fprintf(out, "/* " PPC_VERSION " */\n");
  fprintf(out, "/* NEVER EDIT IT ! */\n");
  fprintf(out, "/* Created on %s by %s */\n", 
    gen_get_date(), gen_get_user_name());
  fprintf(out, "/* %s routine user interface */\n",
    prop_get_routine_name());
  fprintf(out, "\n");
  
  fprintf(out, "int %s (\n", prop_get_routine_name());
  {                                   /* custom parameters for the routine */
    char *act_type;
    int c;
    
    prop_get_param_init();
    act_type = prop_get_param();
    for(c = 0; act_type != NULL; c++)
    {
      if (c > 0)
        fprintf(out, ",\n");
      fprintf(out, "  %s ", act_type);
      if (!strchr(act_type, '*') && prop_get_param_is_out())
        fprintf(out, "*");
      fprintf(out, "p%d", c);
      act_type = prop_get_param();
    }
  }
  fprintf(out, "\n);\n\n");

  fclose(out);
    
  if (verbose)
    printf("routine user interface generated\n");
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* rgen_master_starter:                                                    */
/* Generates the master starter.                                           */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
int rgen_master_starter (
  int verbose
)
{
  FILE *out;
  char filename[256];
  
  if (verbose)
    printf("generating master starter\n");
  
  if (gen_create_dir(PPR_DIR, verbose) == -1)
    return -1;
    
  sprintf(filename, "%s/%s.c", PPR_DIR, MASTER_STARTER_NAME);
  out = fopen(filename, "w");
  if (out == NULL)
    return -1;
    
  fprintf(out, "\n");
  fprintf(out, "/* This file has been generated by the PPC ");
  fprintf(out, "routines manager */\n");
  fprintf(out, "/* " PPC_VERSION " */\n");
  fprintf(out, "/* NEVER EDIT IT ! */\n");
  fprintf(out, "/* Created on %s by %s */\n", 
    gen_get_date(), gen_get_user_name());
  fprintf(out, "/* %s master starter */\n",
    prop_get_routine_name());
  fprintf(out, "\n");
  
  fprintf(out, "#include <ppc/parlib/mkernel.h>\n");
  fprintf(out, "#include \"" MASTER_PROTO_NAME "\"\n");
  fprintf(out, "\n");
  
  fprintf(out, "int %s (\n", prop_get_routine_name());
  {                                   /* custom parameters for the routine */
    char *act_type;
    int c;
    
    prop_get_param_init();
    act_type = prop_get_param();
    for(c = 0; act_type != NULL; c++)
    {
      if (c > 0)
        fprintf(out, ",\n");
      fprintf(out, "  %s ", act_type);
      if (!strchr(act_type, '*') && prop_get_param_is_out())
        fprintf(out, "*");
      fprintf(out, "p%d", c);
      act_type = prop_get_param();
    }
  }
  fprintf(out, "\n)\n");
  fprintf(out, "{\n");
  fprintf(out, "  int master_status;\n\n");
  fprintf(out, "  mkern_start_routine(\"%s\");\n",
    prop_get_routine_name());
  fprintf(out, "  master_status = %s_master_main(\n", 
    prop_get_routine_name());
  {                        /* standard parameters for the routine topology */
    int c;
    topology_type topo = prop_get_topology();
    
    for(c = 0; (*(master_param[topo]))[c].type != NULL; c++)
    {
      if (c > 0)
        fprintf(out, ",\n");
      fprintf(out, "    %s", 
	(*(master_param[topo]))[c].name);
    }
  }
  {                                   /* custom parameters for the routine */
    char *act_type;
    int c;
    
    prop_get_param_init();
    act_type = prop_get_param();
    for(c = 0; act_type != NULL; c++)
    {
      fprintf(out, ",\n");
      fprintf(out, "    p%d", c);
      act_type = prop_get_param();
    }
  }
  fprintf(out, ");\n\n");
  
  fprintf(out, "  { /* wait for slaves to terminate */\n");
  fprintf(out, "    int c;\n");
  fprintf(out, "    int ack;\n\n");
  fprintf(out, "    for (c = 0; c < nb_slaves; c++)\n");
  fprintf(out, "    {\n");
  fprintf(out, "      receive_channel(from_slaves[c],\n");
  fprintf(out, "        INT, &ack,\n");
  fprintf(out, "        END);\n");
  fprintf(out, "    }\n");
  fprintf(out, "  }\n\n");

  fprintf(out, "  return master_status;\n");
  fprintf(out, "}\n");

  fclose(out);
    
  if (verbose)
    printf("master starter generated\n");
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* rgen_slave_starter:                                                     */
/* Generates the slave starter.                                            */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
int rgen_slave_starter (
  int verbose
)
{
  FILE *out;
  char filename[256];
  
  if (verbose)
    printf("generating slave starter\n");
  
  if (gen_create_dir(PPR_DIR, verbose) == -1)
    return -1;
    
  sprintf(filename, "%s/%s.c", PPR_DIR, SLAVE_STARTER_NAME);
  out = fopen(filename, "w");
  if (out == NULL)
    return -1;
    
  fprintf(out, "\n");
  fprintf(out, "/* This file has been generated by the PPC ");
  fprintf(out, "routines manager */\n");
  fprintf(out, "/* " PPC_VERSION " */\n");
  fprintf(out, "/* NEVER EDIT IT ! */\n");
  fprintf(out, "/* Created on %s by %s */\n", 
    gen_get_date(), gen_get_user_name());
  fprintf(out, "/* %s slave starter */\n",
    prop_get_routine_name());
  fprintf(out, "\n");
  
  fprintf(out, "#include <ppc/parlib/skernel.h>\n");
  fprintf(out, "#include \"" SLAVE_PROTO_NAME "\"\n");
  fprintf(out, "\n");
  
  fprintf(out, "int %s_slave_starter (\n", prop_get_routine_name());
  fprintf(out, "  char *name\n");
  fprintf(out, ")\n");
  fprintf(out, "{\n");
  fprintf(out, "  if (!strcmp(name, \"%s\"))\n", prop_get_routine_name());
  fprintf(out, "  {\n");
  fprintf(out, "    %s_slave_main(\n", prop_get_routine_name());

  {                        /* standard parameters for the routine topology */
    int c;
    topology_type topo = prop_get_topology();
    
    for(c = 0; (*(slave_param[topo]))[c].type != NULL; c++)
    {
      if (c > 0)
        fprintf(out, ",\n");
      fprintf(out, "      %s", 
	(*(slave_param[topo]))[c].name);
    }
    fprintf(out, ");\n\n");
  }
  
  fprintf(out, "    return 0;\n");
  fprintf(out, "  }\n\n");
  fprintf(out, "  return -1;\n");
  fprintf(out, "}\n\n");

  fclose(out);
    
  if (verbose)
    printf("slave starter generated\n");
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* rgen_master_header:                                                     */
/* Generates the master header file.                                       */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
int rgen_master_header (
  int verbose
)
{
  FILE *out;
  char filename[256];
  
  if (verbose)
    printf("generating master prototype\n");
  
  if (gen_create_dir(PPR_DIR, verbose) == -1)
    return -1;
    
  sprintf(filename, "%s/%s", PPR_DIR, MASTER_PROTO_NAME);
  out = fopen(filename, "w");
  if (out == NULL)
    return -1;
      
  fprintf(out, "\n");
  fprintf(out, "/* This file has been generated by the PPC ");
  fprintf(out, "routines manager */\n");
  fprintf(out, "/* " PPC_VERSION " */\n");
  fprintf(out, "/* NEVER EDIT IT ! */\n");
  fprintf(out, "/* Created on %s by %s */\n", 
    gen_get_date(), gen_get_user_name());
  fprintf(out, "/* %s master prototype */\n",
    prop_get_routine_name());
  fprintf(out, "\n");
  fprintf(out, "#include <ppc/csp.h>\n");

  fprintf(out, "int %s_master_main(\n", prop_get_routine_name());
  
  {                        /* standard parameters for the routine topology */
    int c;
    topology_type topo = prop_get_topology();
    
    for(c = 0; (*(master_param[topo]))[c].type != NULL; c++)
    {
      if (c > 0)
        fprintf(out, ",\n");
      fprintf(out, "  %s %s", 
        (*(master_param[topo]))[c].type,
	(*(master_param[topo]))[c].name);
      if ((*(master_param[topo]))[c].is_array == ARRAY)
        fprintf(out, "[]");
    }
  }
  {                                   /* custom parameters for the routine */
    char *act_type;
    int c;
    
    prop_get_param_init();
    act_type = prop_get_param();
    for(c = 0; act_type != NULL; c++)
    {
      fprintf(out, ",\n");
      fprintf(out, "  %s ", act_type);
      if (!strchr(act_type, '*') && prop_get_param_is_out())
        fprintf(out, "*");
      fprintf(out, "p%d", c);
      act_type = prop_get_param();
    }
  }
  
  fprintf(out, "\n);\n");
  fclose(out);
    
  if (verbose)
    printf("master prototype generated\n");
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* rgen_slave_header:                                                      */
/* Generates the slave header file.                                        */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
int rgen_slave_header (
  int verbose
)
{
  FILE *out;
  char filename[256];
  
  if (verbose)
    printf("generating slave prototype\n");
  
  if (gen_create_dir(PPR_DIR, verbose) == -1)
    return -1;
    
  sprintf(filename, "%s/%s", PPR_DIR, SLAVE_PROTO_NAME);
  out = fopen(filename, "w");
  if (out == NULL)
    return -1;
      
  fprintf(out, "\n");
  fprintf(out, "/* This file has been generated by the PPC ");
  fprintf(out, "routines manager */\n");
  fprintf(out, "/* " PPC_VERSION " */\n");
  fprintf(out, "/* NEVER EDIT IT ! */\n");
  fprintf(out, "/* Created on %s by %s */\n", 
    gen_get_date(), gen_get_user_name());
  fprintf(out, "/* %s slave prototype */\n",
    prop_get_routine_name());
  fprintf(out, "\n");
  fprintf(out, "#include <ppc/csp.h>\n");

  fprintf(out, "int %s_slave_main(\n", prop_get_routine_name());
  
  {                        /* standard parameters for the routine topology */
    int c;
    topology_type topo = prop_get_topology();
    
    for(c = 0; (*(slave_param[topo]))[c].type != NULL; c++)
    {
      if (c > 0)
        fprintf(out, ",\n");
      fprintf(out, "  %s %s", 
        (*(slave_param[topo]))[c].type,
	(*(slave_param[topo]))[c].name);
      if ((*(slave_param[topo]))[c].is_array == ARRAY)
        fprintf(out, "[]");
    }
  }
  
  fprintf(out, "\n);\n");
  fclose(out);
    
  if (verbose)
    printf("slave prototype generated\n");
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* rgen_object_list:                                                       */
/* Generates in a special file the list of all object files that are part  */
/* of the routine and must be included in the library.                     */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
int rgen_object_list (
  int verbose
)
{
  FILE *out;
  char filename[256];
  char *act_module;
  int c;
  
  if (verbose)
    printf("generating object files list\n");
  
  if (gen_create_dir(PPR_DIR, verbose) == -1)
    return -1;
    
  sprintf(filename, "%s/%s", PPR_DIR, GEN_OBJECTS_FILE);
  out = fopen(filename, "w");
  if (out == NULL)
    return -1;
    
  fprintf(out, MASTER_STARTER_NAME "\n");
  fprintf(out, SLAVE_STARTER_NAME "\n");
  
  prop_get_master_init();
  act_module = prop_get_master_next();
  for(c = 0; act_module != NULL; c++)
  {
    fprintf(out, "%s\n", act_module);
    act_module = prop_get_master_next();
  }
      
  prop_get_slave_init();
  act_module = prop_get_slave_next();
  for(c = 0; act_module != NULL; c++)
  {
    fprintf(out, "%s\n", act_module);
    act_module = prop_get_slave_next();
  }
  
  fprintf(out, "\n");

  fclose(out);
  
  if (verbose)
    printf("object files list generated\n");

  return 0;
}
