/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering School of Biel, Switzerland.                                */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* Portable Parallel Routines command line utility.                        */
/* Routines generator. This module can generate the files needed to        */
/* build a parallel routine for UNIX.                                      */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 10.01.1995  Droux       Created                                         */
/* 05.04.1995  Droux       Define dds_object_ptr for master and slaves     */
/* 01.05.1995  Droux       Local functions are declared as static          */
/*-------------------------------------------------------------------------*/
/* filename: rgenunix.c                                                    */
/* Tested on :                                                             */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1995 Engineering School of Biel-Bienne, Switzerland.      */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#include <stdio.h>

#include <ppc/portable.h>
#include <ppc/csp/version.h>
#include "rgenunix.h"
#include "genutil.h"
#include "prop.h"
#include "toolsutil.h"
#include "rgendefs.h"

/*-------------------------------------------------------------------------*/
/* rgen_unix_makefile:                                               LOCAL */
/* Generates UNIX makefile.                                                */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
static int rgen_unix_makefile (
  int verbose
)
{
  FILE *out;
  char filename[256];
  char *act_module;
  int c;
  
  if (verbose)
    printf("generating UNIX makefile\n");
  
  if (gen_create_dir(PPR_DIR, verbose) == -1)
    return -1;
    
  sprintf(filename, "%s/.%s/makefile", PPR_DIR, arch_names[UNIX_ARCH]);
  out = fopen(filename, "w");
  if (out == NULL)
    return -1;
      
  fprintf(out, "\n");
  fprintf(out, "# This file has been generated by the PPC ");
  fprintf(out, "routines manager\n");
  fprintf(out, "# " PPC_VERSION "\n");
  fprintf(out, "# NEVER EDIT IT !\n");
  fprintf(out, "# Created on %s by %s\n", 
    gen_get_date(), gen_get_user_name());
  fprintf(out, "# %s UNIX makefile\n",
    prop_get_routine_name());
  fprintf(out, "\n");

  fprintf(out, "CFLAGS = -I$(PPC_HOME)/include -D_$(ARCH) ");
  fprintf(out, "-D$(HOST_ARCH) -O $(OTHER_CFLAGS) \\\n");
  fprintf(out, "  $(WARNING_FLAG)\n\n");
  
  fprintf(out, "OBJ = " PPR_DIR "/.%s/$(HOST_ARCH)/" MASTER_STARTER_NAME ".o \\\n",
    arch_names[UNIX_ARCH]);
  fprintf(out, "  " PPR_DIR "/.%s/$(HOST_ARCH)/" SLAVE_STARTER_NAME ".o ",
    arch_names[UNIX_ARCH]);
  
  prop_get_master_init();
  act_module = prop_get_master_next();
  for(c = 0; act_module != NULL; c++)
  {
    fprintf(out, "\\\n");
    fprintf(out, "  " PPR_DIR "/.%s/$(HOST_ARCH)/%s.o ",
      arch_names[UNIX_ARCH], act_module);
    act_module = prop_get_master_next();
  }
  
  prop_get_slave_init();
  act_module = prop_get_slave_next();
  for(c = 0; act_module != NULL; c++)
  {
    fprintf(out, "\\\n");
    fprintf(out, "  " PPR_DIR "/.%s/$(HOST_ARCH)/%s.o ",
      arch_names[UNIX_ARCH], act_module);
    act_module = prop_get_slave_next();
  }
  
  fprintf(out, "\n\n");
  fprintf(out, "all: $(OBJ)\n");
  fprintf(out, "\n");
  
  fprintf(out, PPR_DIR "/.%s/$(HOST_ARCH)/" MASTER_STARTER_NAME ".o: " PPR_DIR "/" MASTER_STARTER_NAME ".c\n",
    arch_names[UNIX_ARCH]);
  fprintf(out, "\t$(CC) $(CFLAGS) -o $@ -c $?\n\n");

  fprintf(out, PPR_DIR "/.%s/$(HOST_ARCH)/" SLAVE_STARTER_NAME ".o: " PPR_DIR "/" SLAVE_STARTER_NAME ".c\n",
    arch_names[UNIX_ARCH]);
  fprintf(out, "\t$(CC) $(CFLAGS) -o $@ -c $?\n\n");

  prop_get_master_init();
  act_module = prop_get_master_next();
  for(c = 0; act_module != NULL; c++)
  {
    fprintf(out, PPR_DIR "/.%s/$(HOST_ARCH)/%s.o: %s.c\n",
      arch_names[UNIX_ARCH], act_module, act_module);
    fprintf(out, "\t$(CC) $(CFLAGS) -Dppr_main=%s_master_main -Ddds_object_ptr=dds_master_object_ptr -o $@ -c $?\n\n", 
      prop_get_routine_name());
    act_module = prop_get_master_next();
  }
  
  prop_get_slave_init();
  act_module = prop_get_slave_next();
  for(c = 0; act_module != NULL; c++)
  {
    fprintf(out, PPR_DIR "/.%s/$(HOST_ARCH)/%s.o: %s.c\n",
      arch_names[UNIX_ARCH], act_module, act_module);
    fprintf(out, "\t$(CC) $(CFLAGS) -Dppr_main=%s_slave_main -Ddds_object_ptr=dds_slave_object_ptr -o $@ -c $?\n\n", 
      prop_get_routine_name());
    act_module = prop_get_slave_next();
  }
  
  fclose(out);
  
  if (verbose)
    printf("UNIX makefile generated\n");
  
  return 0;
}

/*-------------------------------------------------------------------------*/
/* rgen_unix_files:                                                        */
/* Generates files for UNIX.                                               */
/* Returns 0 on success, -1 otherwise.                                     */
/*-------------------------------------------------------------------------*/
int rgen_unix_files (
  int verbose
)
{
  char dir[256];
  
  if (verbose)
    printf("generating UNIX files\n");
  
  if (gen_create_dir(PPR_DIR, verbose) == -1)
    return -1;
    
  sprintf(dir, "%s/.%s", PPR_DIR, arch_names[UNIX_ARCH]);
  if (gen_create_dir(dir, verbose) == -1)
    return -1;
  
  if (rgen_unix_makefile(verbose) < 0)
    return -1;
    
  if (verbose)
    printf("UNIX files generated\n");
  
  return 0;
}

