/*-------------------------------------------------------------------------*/
/* Mathematics on Parallel and Distributed Systems                         */
/* Engineering College of Biel, Switzerland.                               */
/* Computer Science Dpt.                                                   */
/* Swiss National Science Foundation project 5003-034331.                  */
/* Pierre-Andre Chevalier, Walter Businger, Nicolas Droux, Werner Hett.    */
/*-------------------------------------------------------------------------*/
/* Description:                                                            */
/* PPC: Portable Parallel C                                                */
/* NEXTSTEP PPC Control Center.                                            */
/* ProjectController class implementation.                                 */
/* One object of this class is instanciated for each opened project.       */
/* It controls the project objects.                                        */
/*-------------------------------------------------------------------------*/
/* History:                                                                */
/* Date        By          Description                                     */
/* ----------  ----------- ----------------------------------------------- */
/* 28.07.1994  Droux       Created                                         */
/* 27.10.1994  Droux       Added support for warnings switch               */
/* 31.12.1994  Droux       Exec path defined for the user                  */
/* 02.01.1995  Droux       verbose and warning defined for the user        */
/*-------------------------------------------------------------------------*/
/* filename: ProjectController.m                                           */
/* Tested on :                                                             */
/*   NEXTSTEP 3.2, NeXT Computers                                          */
/*   NEXTSTEP 3.2, Intel Computers                                         */
/*   NEXTSTEP 3.3, Intel Computers                                         */
/*-------------------------------------------------------------------------*/
/* Copyright (c) 1994 Engineering College of Biel-Bienne, Switzerland.     */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by the   */
/* Free Software Foundation; either version 2 of the License, or (at your  */
/* option) any later version.                                              */
/*                                                                         */
/* This program is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    */
/* See the GNU General Public License for more details.                    */
/*-------------------------------------------------------------------------*/

#import "ProjectController.h"
#import "CenterController.h"
#import "BrowserController.h"
#import "ProcCoordinator.h"
#import <syslog.h>

#define MINI_ICON_NAME "vmc"                      /* mini icon to be shown */
#define CONFIG_FILE_NAME ".ppccenter"
#define SWITCH_OFF 0                         /* architecture switch states */
#define SWITCH_ON  1

#define OPT_VMCFILE        0x0001                       /* optModif fields */
#define OPT_HWFILES        0x0002

@implementation ProjectController

/*-------------------------------------------------------------------------*/
/* setUp:                                                                  */
/* Set up a new project. Initialize the project, shows a new window and    */
/* initializes its contents.                                               */
/* In addition to the name of the project and offset of the window, this   */
/* method takes as argument the menu options which can be eventually       */
/* modified depending on the state of the project.                         */
/*-------------------------------------------------------------------------*/
- setUp: sender
  projectName: (const char *)name 
  windowOffset: (float)offset
  : menuFilesUpdate
  : menuFilesOpenInWorkspace
  : menuFilesSelectInWorkspace
  : menuFilesAddExisting
  : menuFilesCreateNew
  : menuFilesDelete
{
  NXRect frame;
  
  centerController = sender;

  [projectWindow getFrame: &frame];
  NX_X(&frame) += offset;
  NX_Y(&frame) -= offset;
    
  [projectWindow setTitleAsFilename: name];
  [projectWindow placeWindowAndDisplay: &frame];
  [projectWindow makeKeyAndOrderFront: nil];
      
  strcpy(actProjectName, name);
  [self splitProjectName];
  sprintf(configFileName, "%s" CONFIG_FILE_NAME, projectPath);

  [browserController initForProject: projectName
    path: projectPath
    : menuFilesUpdate
    : menuFilesOpenInWorkspace
    : menuFilesSelectInWorkspace
    : menuFilesAddExisting
    : menuFilesCreateNew
    : menuFilesDelete];
    
  [self loadConfig];

  [[statusTextField cell] setStringValue: "Idle"];
  
  procCoordinator = [[ProcCoordinator alloc] 
    init: centerController 
      scrollView: scrollView
      procStringTable: subProcessStringTable
      projectController: self
      status: statusTextField];
  
  optModif = 0;
      
  return self;
}

/*-------------------------------------------------------------------------*/
/* selectedArch:                                                           */
/* Returns the selected architecture.                                      */
/*-------------------------------------------------------------------------*/
- (archEnum) selectedArch
{
  return (archEnum)selectedArch;
}

/*-------------------------------------------------------------------------*/
/* unselectUnsupportedArch:                                                */
/* Uselect the selected switches which are not enabled.                    */
/*-------------------------------------------------------------------------*/
- unselectUnsupportedArch
{
  int actArch;
  
  for (actArch = 0; actArch < NB_ARCH; actArch++)
  {
    if (![[archSwitches cellAt: actArch: 0] isEnabled] &&
	  [[archSwitches cellAt: actArch: 0] intValue])
    {                /* architecture is selected but disabled. diselect it */
      [[archSwitches cellAt: actArch: 0] setEnabled: YES];
      [[archSwitches cellAt: actArch: 0] performClick: self];
      [[archSwitches cellAt: actArch: 0] setEnabled: NO];
    }
  }
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* clearBuffer:                                                            */
/* Sent when the clear buffer menu option is selected.                     */
/*-------------------------------------------------------------------------*/
- clearBuffer: sender
{
  [procCoordinator clearBuffer: sender];
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* vmcOptionalFile:                                                        */
/* Returns the VMC file if defined, or an empty string if it is not the    */
/* case.                                                                   */
/*-------------------------------------------------------------------------*/
- (char *)vmcOptionalFile
{
  return vmcFile;
}

/*-------------------------------------------------------------------------*/
/* hwOptionalFile:                                                         */
/* Returns the hardware configuration file for the given architecture if   */
/* defined, or an empty string if it is not the case.                      */
/*-------------------------------------------------------------------------*/
- (char *)hwOptionalFile: (archEnum)arch
{
  return hwFiles[arch];
}

/*-------------------------------------------------------------------------*/
/* vmcFileButtonMatrixClicked:                                             */
/* Sent when one of the virtual machine configuration file buttons is      */
/* clicked.                                                                */
/*-------------------------------------------------------------------------*/
- vmcFileButtonMatrixClicked: sender
{
  if ([[sender selectedCell] tag] % 2)
  {
    strcpy(tmpOptions.vmcFile, "");                     /* restore default */
  }
  else
  {                                                        /* set new file */
    char *types[2] = {"vmc", 0}; /* allowed file extensions for open panel */
  
    if ([[OpenPanel new] 
         runModalForDirectory: projectPath file: NULL types: types]
	!= NX_OKTAG)
    {
      return nil;
    }
    else
    {
      char newFile[MAX_NAME_LEN];
      char newDir[MAX_NAME_LEN];
      
      [self splitPath: [[OpenPanel new] filename]
        directory: newDir
	fileName: newFile];
      
      if (strcmp(newDir, projectPath))
      {
        NXRunAlertPanel("Set Optional VMC File",
	  "An optional virtual machine configuration file must be stored in the PPC program directory.", NULL, NULL, NULL);
	return nil;
      }
      else
      {
        strcpy(tmpOptions.vmcFile, newFile);
      }
    }
  }
  
  [[vmcFileFormCellMatrix cellAt: 0: 0] setStringValue: tmpOptions.vmcFile];
  optModif |= OPT_VMCFILE;
  
  [optionsSetButton setEnabled: YES];
  [optionsRevertButton setEnabled: YES];

  return self;
}

/*-------------------------------------------------------------------------*/
/* hwFilesButtonMatrixClicked:                                             */
/* Sent when one of the hardware configuration files buttons is            */
/* clicked.                                                                */
/*-------------------------------------------------------------------------*/
- hwFilesButtonMatrixClicked: sender
{
  int arch = [[sender selectedCell] tag] / 2;
  
  if ([[sender selectedCell] tag] % 2)
  {
    strcpy(tmpOptions.hwFiles[arch], "");               /* restore default */
  }
  else
  {                                                        /* set new file */
    char *types[2] = {"c", 0};   /* allowed file extensions for open panel */
  
    if ([[OpenPanel new] 
         runModalForDirectory: projectPath file: NULL types: types]
	!= NX_OKTAG)
    {
      return nil;
    }
    else
    {
      char newFile[MAX_NAME_LEN];
      char newDir[MAX_NAME_LEN];
      
      [self splitPath: [[OpenPanel new] filename]
        directory: newDir
	fileName: newFile];
	
      if (strcmp(newDir, projectPath))
      {
        NXRunAlertPanel("Set Optional Hardware Config File",
	  "An optional hardware configuration file must be stored in the PPC program directory.", NULL, NULL, NULL);
	return nil;
      }
      else
      {
        strcpy(tmpOptions.hwFiles[arch], newFile);
      }
    }
  }
  
  [[hwFilesFormCellMatrix cellAt: arch: 0] 
    setStringValue: tmpOptions.hwFiles[arch]];
  optModif |= (OPT_HWFILES << arch);
  
  [optionsSetButton setEnabled: YES];
  [optionsRevertButton setEnabled: YES];

  return self;
}

/*-------------------------------------------------------------------------*/
/* filesMenuSelected:                                                      */
/* Message sent by the center controller when one of the Files menu option */
/* is selected.                                                            */
/*-------------------------------------------------------------------------*/
- filesMenuSelected: sender;
{
  [browserController filesMenuSelected: sender];

  return self;
}

/*-------------------------------------------------------------------------*/
/* actionsMenuSelected:                                                    */
/* Message sent by the menu when one of the Actions options is selected.   */
/*-------------------------------------------------------------------------*/
- actionsMenuSelected: sender
{
  [procCoordinator actionsMenuSelected: sender];

  return self;
}

/*-------------------------------------------------------------------------*/
/* updateArchitectures:                                                    */
/* Update all architectures switches and field.                            */
/*-------------------------------------------------------------------------*/
- updateArchitectures
{
  int actArch;

  for (actArch = 0; actArch < NB_ARCH; actArch++)
  {                                           /* disable all architectures */
    [[archSwitches cellAt: actArch: 0] setEnabled: NO];
    [[hwFilesFormCellMatrix cellAt: actArch: 0] setEnabled: NO];
    [[hwFilesButtonMatrix cellAt: actArch: 0] setEnabled: NO];
    [[hwFilesButtonMatrix cellAt: actArch: 1] setEnabled: NO];
  }
  
  [browserController updateField: HARDWARE_FIELD];
  [self unselectUnsupportedArch];

  return self;
}

/*-------------------------------------------------------------------------*/
/* showOptions:                                                            */
/* Shows the options panel. This message is sent when the "Options..."     */
/* button of the project window is clicked.                                */
/*-------------------------------------------------------------------------*/
- showOptions: sender
{
  NXRect optionsButtonFrame;
  NXRect optionsPanelFrame;
  NXRect buttonWindowFrame;
  char panelTitle [100];
  
  /* shows the panel, make it the key window (but not the main window),    */
  /* and place it just in front of the options button.                     */
  
  [optionsPanel getFrame: &optionsPanelFrame];
  [optionsButton getFrame: &optionsButtonFrame];
  [[optionsButton window] getFrame: &buttonWindowFrame];

  NX_X(&optionsPanelFrame) = 
    NX_X(&optionsButtonFrame) + NX_X(&buttonWindowFrame);
  NX_Y(&optionsPanelFrame) = 
    NX_Y(&optionsButtonFrame) + NX_Y(&buttonWindowFrame)
    - NX_HEIGHT(&optionsPanelFrame)
    + NX_HEIGHT(&optionsButtonFrame) + 10.0;

  [optionsPanel placeWindowAndDisplay: &optionsPanelFrame];
  
  sprintf(panelTitle, "%s Project Options", projectName);
  [optionsPanel setTitle: panelTitle];

  [[optionsPanel orderFront: self] makeKeyWindow];

  return self;
}

/*-------------------------------------------------------------------------*/
/* archStateChanged:                                                       */
/* Message sent by the architecture matrix when the state of a switch is   */
/* changed.                                                                */
/*-------------------------------------------------------------------------*/
- archStateChanged: sender
{
  int newArch;
  
  newArch = [[sender selectedCell] tag];
  if (newArch != selectedArch)
  {
    [[archSwitches cellAt: selectedArch: 0] setIntValue: SWITCH_OFF];
    [projectWindow setDocEdited: YES];
  }
  selectedArch = newArch;
  [[archSwitches cellAt: selectedArch: 0] setIntValue: SWITCH_ON];

  return self;
}

/*-------------------------------------------------------------------------*/
/* projectName:                                                            */
/* Return the name of the project.                                         */
/*-------------------------------------------------------------------------*/
- (const char *)projectName
{
  return actProjectName;
}

/*-------------------------------------------------------------------------*/
/* projectNameWithoutPath:                                                 */
/* Return the name of the project but without its path.                    */
/*-------------------------------------------------------------------------*/
- (const char *)projectNameWithoutPath
{
  return projectName;
}

/*-------------------------------------------------------------------------*/
/* projectPath:                                                            */
/* Return the path of the project.                                         */
/*-------------------------------------------------------------------------*/
- (const char *)projectPath
{
  return projectPath;
}

/*-------------------------------------------------------------------------*/
/* makeActive:                                                             */
/* Make the project active by making the project window the key and main   */
/* window of the application.                                              */
/*-------------------------------------------------------------------------*/
- makeActive
{
  [projectWindow makeKeyAndOrderFront: nil];
  return self;
}

/*-------------------------------------------------------------------------*/
/* isActive:                                                               */
/* Return YES if the project is active. Only one project may be active at  */
/* a time.                                                                 */
/*-------------------------------------------------------------------------*/
- (BOOL) isActive
{
  return [projectWindow isMainWindow];
}

/*-------------------------------------------------------------------------*/
/* splitPath:                                                              */
/* Split a path into a directory and a file name.                          */
/*-------------------------------------------------------------------------*/
- splitPath: (const char *)path
  directory: (char *)dir
  fileName: (char *)file
{
  int last_slash = 0;
  int idx;
  int last_point = 0;
  char newPath [MAX_NAME_LEN];
  char newName [MAX_NAME_LEN];
    
  for (idx = 0; path[idx] != '\0'; idx++)                  /* extract path */
    if (path[idx] == '/')
      last_slash = idx;
      
  strcpy(newPath, path);
  newPath[last_slash+1] = '\0';
  
  for (idx = last_slash+1; path[idx] != '\0'; idx++)       /* extract name */
    newName[idx-(last_slash+1)] = path[idx];
  newName[idx-(last_slash+1)] = '\0';
  
  for (idx = 0; newName[idx] != '\0'; idx++)       /* check file extension */
    if (newName[idx] == '.')
      last_point = idx;
  newName[last_point] = '\0';

  strcpy(file, newName);
  strcpy(dir, newPath);

  return self;
}

/*-------------------------------------------------------------------------*/
/* splitProjectName:                                                       */
/* This extract the path and the file name components from the project     */
/* full name and store the corresponding values in the instance variables. */
/*-------------------------------------------------------------------------*/
- splitProjectName
{
  [self splitPath: actProjectName
    directory: projectPath
    fileName: projectName];

  return self;
}

/*-------------------------------------------------------------------------*/
/* setDefaults:                                                            */
/* Set default configuration parameters.                                   */
/*-------------------------------------------------------------------------*/
- setDefaults
{
  /* The defaults are default search path and UNIX architecture            */
  
  [[archSwitches cellAt: ARCH_UNIX: 0] setEnabled: YES];
  [[archSwitches cellAt: ARCH_UNIX: 0] setIntValue: SWITCH_ON];
  strcpy(vmcFile, "");
  {
    int actArch;
    for (actArch = 0; actArch < NB_ARCH; actArch++)
      strcpy(hwFiles[actArch], "");
  }
  
  [projectWindow setDocEdited: YES];
  canBeRevertedToSaved = NO;
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* projectCorrupted:                                                       */
/* Must be sent when a project configuration file is corrupted.            */
/*-------------------------------------------------------------------------*/
- projectCorrupted: (FILE *) configFile
  status: (char *) msg
{
  NXRunAlertPanel("Load",
    "Cannot load project, configuration file is corrupted (%s.) %s",
    NULL, NULL, NULL,
    msg,
    "Setting defaults.");
    
  [self setDefaults];
  fclose(configFile);
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* loadConfig:                                                             */
/* Load the configuration informations of the given project.               */
/* If no configuration is defined, then define defaults and set the        */
/* project to edited.                                                      */
/*-------------------------------------------------------------------------*/
- loadConfig
{
  FILE *configFile = fopen(configFileName, "r");
  BOOL isEdited = NO;
  
  {                                           /* disable all architectures */
    int actArch;
    for (actArch = 0; actArch < NB_ARCH; actArch++)
    {
      [[archSwitches cellAt: actArch: 0] setEnabled: NO];
      [[archSwitches cellAt: actArch: 0] setIntValue: SWITCH_OFF];
      [[hwFilesFormCellMatrix cellAt: actArch: 0] setEnabled: NO];
      [[hwFilesButtonMatrix cellAt: actArch: 0] setEnabled: NO];
      [[hwFilesButtonMatrix cellAt: actArch: 1] setEnabled: NO];
    }
  }

  if (configFile == NULL)
  {                            /* config file does not exist, set defaults */
    [self setDefaults];                         
    [browserController updateAllFields];
    [self unselectUnsupportedArch];
    return self;
  }

  { /* - - - - - - - - - - - - - - - - - - - - - - - - - - executable path */
    /* The default executable path is no more defined for a project but    */
    /* for the user of PPCCenter. It must be read in order to support      */
    /* backward compatibility with configuration files generated by        */
    /* older versions of PPCCenter.                                        */
    
    unsigned short pathLength;
    unsigned short pathLengthBig;
    int defaultExecPath;
                                            /* read executable path length */
    if (fread(&pathLengthBig, sizeof(pathLengthBig), 1, configFile) < 1)
    {                                                /* error reading file */
      [self projectCorrupted: configFile 
        status: "cannot read executable path length"];
      return self;
    }
    
    pathLength = NXSwapBigShortToHost(pathLengthBig);
    if (pathLength >= MAX_NAME_LEN)
    {                                          /* path length is not valid */
      [self projectCorrupted: configFile
        status: "invalid path length"];
      return self;
    }
    defaultExecPath = pathLength == 0;
    
    if (!defaultExecPath)
    {                                              /* read executable path */
      char executablePath[MAX_NAME_LEN];          
      if (fread(executablePath, sizeof(char), pathLength, configFile)
          < pathLength)
      {                                              /* error reading file */
	[self projectCorrupted: configFile
	  status: "cannot read executable path"];
	return self;
      }
      executablePath[pathLength] = '\0';
    }
  }
  { /* - - - - - - - - - - - - - - - - - - - - - - - - target architecture */      
    int colCount;
    int rowCount;
    int maxTargetArch;
    char selectedArchChar;
      
    [archSwitches getNumRows: &rowCount 
      numCols: &colCount];   /* get maximum allowed target architectures */
    maxTargetArch = rowCount-1;

    fread(&selectedArchChar, 1, 1, configFile);
    selectedArch = (int)selectedArchChar;
    
    if (selectedArch < 0 || selectedArch > maxTargetArch)
    {
      [self projectCorrupted: configFile 
	status: "invalid target architecture index"];
      return self;
    }
    
    [[archSwitches cellAt: selectedArch: 0] setIntValue: SWITCH_ON];
    [[archSwitches cellAt: selectedArch: 0] performClick: self];
  }
  { /* - - - - - - - - - - - - - - - - - - - - - - - - - - -  verbose mode */
    char verboseChar;
      
    fread(&verboseChar, 1, 1, configFile);
  }
  { /* - - - - - - - - - - - - - - - - - - - - - - - - - optional VMC file */
    unsigned short nameLength;
    unsigned short nameLengthBig;
    
    if (fread(&nameLengthBig, sizeof(nameLengthBig), 1, configFile) < 1)
    {                                                /* error reading file */
      [self projectCorrupted: configFile 
        status: "cannot read VMC optional path length"];
      return self;
    }
    nameLength = NXSwapBigShortToHost(nameLengthBig);
    
    if (nameLength >= MAX_NAME_LEN)
    {                                          /* name length is not valid */
      [self projectCorrupted: configFile
        status: "invalid name length"];
      return self;
    }
    
    if (nameLength)
    {
      if (fread(vmcFile, sizeof(char), nameLength, configFile)
          < nameLength)
      {                                              /* error reading file */
	[self projectCorrupted: configFile
	  status: "cannot read optional VMC file name"];
	return self;
      }
      vmcFile[nameLength] = '\0';
    }
    else
      strcpy(vmcFile, "");
      
    [[vmcFileFormCellMatrix cellAt: 0: 0] setStringValue: vmcFile];
  }
  { /* - - - - - - - - - - - - - - -  optional hardware configuration file */
    unsigned short nameLength;
    unsigned short nameLengthBig;
    int actArch;
    
    for (actArch = 0; actArch < NB_ARCH; actArch++)
    {
      if (fread(&nameLengthBig, sizeof(nameLengthBig), 1, configFile) < 1)
      {                                              /* error reading file */
	[self projectCorrupted: configFile 
	  status: "cannot read hardware configuration optional path length"];
	return self;
      }
      nameLength = NXSwapBigShortToHost(nameLengthBig);
      
      if (nameLength >= MAX_NAME_LEN)
      {                                        /* name length is not valid */
	[self projectCorrupted: configFile
	  status: "invalid name length"];
	return self;
      }
      
      if (nameLength)
      {
	if (fread(hwFiles[actArch], sizeof(char), nameLength, configFile)
	    < nameLength)
	{                                            /* error reading file */
	  [self projectCorrupted: configFile
	    status: "cannot read optional hardware configuration file name"];
	  return self;
	}
	hwFiles[actArch][nameLength] = '\0';
      }
      else
	strcpy(hwFiles[actArch], "");
	
      [[hwFilesFormCellMatrix cellAt: actArch: 0] 
        setStringValue: hwFiles[actArch]];
    }
  }
  { /* - - - - - - - - - - - - - - - - - - - - - - - - - - warnings switch */
    char warningsChar;
      
    fread(&warningsChar, 1, 1, configFile);
  }
    
  [projectWindow setDocEdited: isEdited];
  canBeRevertedToSaved = YES;
  [browserController updateAllFields];
  [self unselectUnsupportedArch];
  fclose(configFile);

  return self;
}

/*-------------------------------------------------------------------------*/
/* saveConfig:                                                             */
/* Save the project configuration.                                         */
/*-------------------------------------------------------------------------*/
- saveConfig
{
  FILE *configFile = fopen(configFileName, "w");

  if (configFile == NULL)
  {                               /* error creating the configuration file */
    int panelResult;
    
    panelResult = NXRunAlertPanel("Save",
      "Cannot create configuration file for %s",
      "Don't Save", "Cancel", NULL,
      projectName);
      
    switch(panelResult)
    {
      case 1:                                                /* Don't Save */
        return self;
      case 0:                                                    /* Cancel */
        [centerController saveCanceled];
        break;
      default:                                /* should normally not occur */
        return self;
    }
      
    return nil;                                                /* canceled */
  }
  
  { /* - - - - - - - - - - - - - - - - - - - - - - - - - - executable path */
    unsigned short pathLength;
    unsigned short pathLengthBig;
    int defaultExecPath = 1;                 /* for backward compatibility */
    char execPath[] = "";
                             
    if (defaultExecPath)         /* length of the optional path if defined */
      pathLength = 0;
    else
      pathLength = strlen(execPath);
    pathLengthBig = NXSwapHostShortToBig(pathLength);
    fwrite(&pathLengthBig, sizeof(pathLengthBig), 1, configFile);
                                             /* executable path if defined */
    fwrite(execPath, sizeof(char), pathLength, configFile);
  }
  { /* - - - - - - - - - - - - - - - - - - -  selected target architecture */
      char selectedArchByte = (char)selectedArch;
      fwrite(&selectedArchByte, 1, 1, configFile);
  }
  { /* - - - - - - - - - - - - - - - - - - - - - - - - verbose mode or not */
      char verboseChar = (char)0;
      fwrite(&verboseChar, 1, 1, configFile);
  }
  { /* - - - - - - - - - - - - - - - - - - - - - - - - - optional VMC file */
    unsigned short nameLength;
    unsigned short nameLengthBig;
    
    nameLength = strlen(vmcFile);
    nameLengthBig = NXSwapHostShortToBig(nameLength);
    fwrite(&nameLengthBig, sizeof(nameLengthBig), 1, configFile);
    fwrite(vmcFile, sizeof(char), nameLength, configFile);
  }
  { /* - - - - - - - - - - - - - - - optional hardware configuration files */
    unsigned short nameLength;
    unsigned short nameLengthBig;
    int actArch;
    
    for (actArch = 0; actArch < NB_ARCH; actArch++)
    {
      nameLength = strlen(hwFiles[actArch]);
      nameLengthBig = NXSwapHostShortToBig(nameLength);
      fwrite(&nameLengthBig, sizeof(nameLengthBig), 1, configFile);
      fwrite(hwFiles[actArch], sizeof(char), nameLength, configFile);
    }
  }
  { /* - - - - - - - - - - - - - - - - - - - - - - - - - - warnings switch */
      char warningsChar = (char)0;
      fwrite(&warningsChar, 1, 1, configFile);
  }
  
  [projectWindow setDocEdited: NO];
  canBeRevertedToSaved = YES;
  fclose(configFile);
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* revertToSaved:                                                          */
/* Revert project to its saved state.                                      */
/*-------------------------------------------------------------------------*/
- revertToSaved
{
  int panelResult;

  if ([projectWindow isDocEdited])
  {
    panelResult = NXRunAlertPanel("Revert",
      "Do you want to revert changes to %s?",
      "Revert", "Cancel", NULL,
      projectName);
      
    switch(panelResult)
    {
      case 1:                                                    /* Revert */
        break;
      case 0:                                                    /* Cancel */
        return self;
      default:                                /* should normally not occur */
	return self;
    }
  }

  [self loadConfig];

  return self;
}

/*-------------------------------------------------------------------------*/
/* isEdited:                                                               */
/* Returns YES if the project needs to be saved, NO otherwise.             */
/*-------------------------------------------------------------------------*/
- (BOOL) isEdited
{
  return [projectWindow isDocEdited];
}

/*-------------------------------------------------------------------------*/
/* setNotEdited:                                                           */
/* Set the project to be not edited.                                       */
/*-------------------------------------------------------------------------*/
- setNotEdited
{
  [projectWindow setDocEdited: NO];
  return self;
}

/*-------------------------------------------------------------------------*/
/* close:                                                                  */
/* Simulates a close operation on the project window. This message is      */
/* sent to the controller when closing a project.                          */
/*-------------------------------------------------------------------------*/
- close
{
  [projectWindow performClose: self];
  return self;
}

/*-------------------------------------------------------------------------*/
/* enableArch:                                                             */
/* Enable or disable the specified architecture. This method is called by  */
/* the browser controller when it updates the list of the available        */
/* hardware configuration file names.                                      */
/*-------------------------------------------------------------------------*/
- enableArch: (archEnum) arch: (BOOL)flag
{
  [[archSwitches cellAt: arch: 0] setEnabled: flag];
  [[hwFilesFormCellMatrix cellAt: arch: 0] setEnabled: flag];
  [[hwFilesButtonMatrix cellAt: arch: 0] setEnabled: flag];
  [[hwFilesButtonMatrix cellAt: arch: 1] setEnabled: flag];
  
  return self;
}

/*-------------------------------------------------------------------------*/
/* canBeRevertedToSaved:                                                   */
/* Return YES if the project can be reverted to saved.                     */
/*-------------------------------------------------------------------------*/
- (BOOL) canBeRevertedToSaved
{
  return canBeRevertedToSaved;
}

/*-------------------------------------------------------------------------*/
/* optionsSet:                                                             */
/* The set button of the options panel has been clicked.                   */
/*-------------------------------------------------------------------------*/
- optionsSet: sender
{
  int actArch;
  
  if (optModif & OPT_VMCFILE)
    strcpy(vmcFile, tmpOptions.vmcFile);
  
  for (actArch = 0; actArch < NB_ARCH; actArch++)
  {
    if (optModif & (OPT_HWFILES << actArch))
      strcpy(hwFiles[actArch], tmpOptions.hwFiles[actArch]);
  }

  optModif = 0;
  [projectWindow setDocEdited: YES];

  [optionsSetButton setEnabled: NO];
  [optionsRevertButton setEnabled: NO];

  return self;
}

/*-------------------------------------------------------------------------*/
/* optionsRevert:                                                          */
/* The revert button of the options panel has been clicked.                */
/*-------------------------------------------------------------------------*/
- optionsRevert: sender
{
  int actArch;
  
  if (optModif & OPT_VMCFILE)
  {
    [[vmcFileFormCellMatrix cellAt: 0: 0] 
      setStringValue: vmcFile];
  }
  
  for (actArch = 0; actArch < NB_ARCH; actArch++)
  {
    if (optModif & (OPT_HWFILES << actArch))
      [[hwFilesFormCellMatrix cellAt: actArch: 0] 
        setStringValue: hwFiles[actArch]];
  }

  optModif = 0;

  [optionsSetButton setEnabled: NO];
  [optionsRevertButton setEnabled: NO];

  return self;
}

/*-------------------------------------------------------------------------*/
/* windowWillClose:                                                        */
/* Project window delegate window. Message sent when the window is closed  */
/* manually by the user or by the close: method.                           */
/*-------------------------------------------------------------------------*/
- windowWillClose: sender
{
  if ([projectWindow isDocEdited])
  {         /* project configuration has changed, save if required by user */
    int alertResult;
    
    [projectWindow makeKeyAndOrderFront: self];
    
    alertResult = NXRunAlertPanel("Save",
      "Save changes to %s ?",
      "Save", "Don't Save", "Cancel",
      projectName);
    
    switch(alertResult)
    { 
      case 1:                                                      /* save */
        if ([self saveConfig] == nil)
	{
          [centerController saveCanceled];
          return nil;
	}
        break;
      case 0:                                                /* don't save */
        break;
      case -1:                                                   /* cancel */
        [centerController saveCanceled];
        return nil;
      default:                                /* should normally not occur */
        return nil;
    }
  }
  
  [optionsPanel performClose: self];
  [sender setDelegate: nil];
  [centerController removeProject: self];
  [self free];
  return self;
}

/*-------------------------------------------------------------------------*/
/* windowWillMiniaturize:                                                  */
/* Message sent when the project window is miniaturized. Allows the        */
/* definition of the miniaturized window image.                            */
/*-------------------------------------------------------------------------*/
- windowWillMiniaturize: sender 
  toMiniwindow: miniwindow
{
  [sender setMiniwindowIcon: MINI_ICON_NAME];
  return self;
}

/*-------------------------------------------------------------------------*/
/* windowDidBecomeMain:                                                    */
/* Message sent when the project window become the main window.            */
/*-------------------------------------------------------------------------*/
- windowDidBecomeMain:sender
{
  [centerController setActiveProject: self];
  [browserController clicked: self];               /* to update Files menu */
  return self;
}

/*-------------------------------------------------------------------------*/
/* free:                                                                   */
/* Release dynamically allocated instance variable.                        */
/*-------------------------------------------------------------------------*/
- free
{
  [browserController free];
  return [super free];
}

@end
